package com.nordicpeak.flowengine.queries.basetablequery;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Key;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.standardutils.xml.XMLValidationUtils;

import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListQuery;
import com.nordicpeak.flowengine.interfaces.ImmutableAlternative;
import com.nordicpeak.flowengine.queries.basequery.BaseQuery;
import com.nordicpeak.flowengine.queries.basetablequery.interfaces.TableQuery;

@XMLElement
public abstract class BaseTableQuery<DropAlternative extends BaseDropDownAlternative, Query extends BaseTableQuery<DropAlternative, Query, Column>, Column extends BaseTableColumn<DropAlternative, Column, Query>> extends BaseQuery implements TableQuery, AlternativeListQuery {

	private static final long serialVersionUID = 5055207847886310491L;

	@DAOManaged
	@Key
	@XMLElement
	private Integer queryID;

	private Map<Integer, Integer> columnConversionMap;
	private Map<Integer, Integer> dropDownAlternativeConversionMap;

	@Override
	public Integer getQueryID() {

		return queryID;
	}

	public void setQueryID(int queryID) {

		this.queryID = queryID;
	}

	@Override
	public String toString() {

		if (this.queryDescriptor != null) {

			return queryDescriptor.getName() + " (queryID: " + queryID + ")";
		}

		return this.getClass().getSimpleName() + " (queryID: " + queryID + ")";
	}

	@Override
	public String getXSDTypeName() {

		return this.getClass().getSimpleName() + queryID;
	}

	@Override
	public void toXSD(Document doc) {

		Element complexTypeElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:complexType");
		complexTypeElement.setAttribute("name", getXSDTypeName());

		Element complexContentElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:complexContent");
		complexTypeElement.appendChild(complexContentElement);

		Element extensionElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:extension");
		extensionElement.setAttribute("base", "Query");
		complexContentElement.appendChild(extensionElement);

		Element sequenceElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:sequence");
		extensionElement.appendChild(sequenceElement);

		addElementType(doc, sequenceElement, "Name", "xs:string", queryDescriptor.getName());

		Element rowElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:element");
		rowElement.setAttribute("name", "Row");
		rowElement.setAttribute("type", getXSDTypeName() + "Row");
		rowElement.setAttribute("minOccurs", "" + getMinRowOccurance());
		rowElement.setAttribute("maxOccurs", "" + getMaxRowOccurance());

		sequenceElement.appendChild(rowElement);

		doc.getDocumentElement().appendChild(complexTypeElement);

		appendXSDTypes(doc);
	}

	protected abstract int getMinRowOccurance();

	protected abstract int getMaxRowOccurance();

	public void appendXSDTypes(Document doc) {

		// Row type
		Element rowComplexTypeElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:complexType");
		rowComplexTypeElement.setAttribute("name", getXSDTypeName() + "Row");

		Element rowSequenceElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:sequence");
		rowComplexTypeElement.appendChild(rowSequenceElement);

		addElementType(doc, rowSequenceElement, "ID", "xs:positiveInteger");
		addElementType(doc, rowSequenceElement, "Name", "xs:string", false);

		if (!CollectionUtils.isEmpty(getColumns())) {

			ArrayList<String> columnElementNames = new ArrayList<>(getColumns().size());

			// Columns
			for (Column column : getColumns()) {

				String columnTypeName = getXSDTypeName() + "Column" + column.getColumnID();

				Element columnSimpleTypeElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:complexType");
				columnSimpleTypeElement.setAttribute("name", columnTypeName);

				Element columnSequenceElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:sequence");
				columnSimpleTypeElement.appendChild(columnSequenceElement);

				addElementType(doc, columnSequenceElement, "ID", "xs:positiveInteger", column.getColumnID().toString());
				addElementType(doc, columnSequenceElement, "Name", "xs:string", column.getLabel());

				if (column.getColumnType() == ColumnType.CHECKBOX) {

					addElementType(doc, columnSequenceElement, "Value", "xs:boolean");

				} else if (column.getColumnType() == ColumnType.STRING || column.getColumnType() == ColumnType.TEXTAREA) {

					addElementType(doc, columnSequenceElement, "Value", "xs:string");

				} else if (column.getColumnType() == ColumnType.DROPDOWN) {

					String typeName = getXSDTypeName() + "Alternative" + column.getColumnID();

					addElementType(doc, columnSequenceElement, "Value", typeName);

					Element alternativesSimpleTypeElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:simpleType");
					alternativesSimpleTypeElement.setAttribute("name", typeName);

					Element alternativesRestrictionElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:restriction");
					alternativesRestrictionElement.setAttribute("base", "xs:string");
					alternativesSimpleTypeElement.appendChild(alternativesRestrictionElement);

					if (column.getDropDownAlternatives() != null) {

						for (ImmutableAlternative alternative : column.getDropDownAlternatives()) {

							Element alternativesEnumerationElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:enumeration");
							alternativesEnumerationElement.setAttribute("value", alternative.getName());
							alternativesRestrictionElement.appendChild(alternativesEnumerationElement);
						}
					}

					doc.getDocumentElement().appendChild(alternativesSimpleTypeElement);
				}

				doc.getDocumentElement().appendChild(columnSimpleTypeElement);

				addElementType(doc, rowSequenceElement, generateElementName("Column", columnElementNames), columnTypeName, false);
			}
		}

		doc.getDocumentElement().appendChild(rowComplexTypeElement);
	}

	public void addElementType(Document doc, Element sequenceElement, String name, String type) {

		addElementType(doc, sequenceElement, name, type, null, true);
	}

	public void addElementType(Document doc, Element sequenceElement, String name, String type, boolean required) {

		addElementType(doc, sequenceElement, name, type, null, required);
	}

	public void addElementType(Document doc, Element sequenceElement, String name, String type, String fixed) {

		addElementType(doc, sequenceElement, name, type, fixed, true);
	}

	public void addElementType(Document doc, Element sequenceElement, String name, String type, String fixed, boolean required) {

		Element element = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:element");
		element.setAttribute("name", name);
		element.setAttribute("type", type);
		element.setAttribute("minOccurs", required ? "1" : "0");
		element.setAttribute("maxOccurs", "1");

		if (!StringUtils.isEmpty(fixed)) {
			element.setAttribute("fixed", fixed);
		}

		sequenceElement.appendChild(element);
	}

	public static String generateElementName(String label, ArrayList<String> fieldElementNames) {

		String elementName = XMLUtils.toValidElementName(label);

		if (fieldElementNames.contains(elementName)) {

			int counter = 1;

			while (fieldElementNames.contains(elementName + counter)) {

				counter++;
			}

			elementName += counter;
		}

		fieldElementNames.add(elementName);

		return elementName;
	}

	@Override
	public void populate(XMLParser xmlParser) throws ValidationException {

		List<ValidationError> errors = new ArrayList<>();

		description = XMLValidationUtils.validateParameter("description", xmlParser, false, 1, 65535, StringPopulator.getPopulator(), errors);
		helpText = XMLValidationUtils.validateParameter("helpText", xmlParser, false, 1, 65535, StringPopulator.getPopulator(), errors);

		populateColumns(xmlParser, errors);

		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}

	}

	protected abstract void populateColumns(XMLParser xmlParser, List<ValidationError> errors) throws ValidationException;

	public void setQueryID(Integer queryID) {

		this.queryID = queryID;
	}

	public abstract boolean hasRequiredColumn();

	public boolean hasRequiredColumn(List<Column> columns) {

		if (!CollectionUtils.isEmpty(columns)) {

			for (Column column : columns) {

				if (column.isRequired()) {

					return true;
				}
			}
		}

		return false;
	}

	@Override
	public abstract List<Column> getColumns();

	public abstract void setColumns(List<Column> columns);

	public Map<Integer, Integer> getColumnConversionMap() {

		return columnConversionMap;
	}

	public void setColumnConversionMap(Map<Integer, Integer> columnConversionMap) {

		this.columnConversionMap = columnConversionMap;
	}

	public Map<Integer, Integer> getDropDownAlternativeConversionMap() {

		return dropDownAlternativeConversionMap;
	}

	public void setDropDownAlternativeConversionMap(Map<Integer, Integer> dropDownAlternativeConversionMap) {

		this.dropDownAlternativeConversionMap = dropDownAlternativeConversionMap;
	}

	@Override
	public String getAlternativeListName() {

		List<String> listNames = CollectionUtils.map(getColumns(), Column::getAlternativeListName);

		return listNames != null ? String.join(", ", listNames) : null;
	}

	@Override
	public String getExtraUsageInfo() {

		List<String> columnLabels = CollectionUtils.map(getColumns(), Column::getLabel);

		return columnLabels != null ? String.join(", ", columnLabels) : null;
	}

	public abstract Integer getDecimalsToShow();

	public Column getColumnByID(Integer columnID) {

		return CollectionUtils.find(getColumns(), column -> column.getColumnID().equals(columnID));
	}

}
