package com.nordicpeak.flowengine.booking;

import java.sql.Date;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.handlers.UserHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.utils.UserUtils;
import se.unlogic.hierarchy.foregroundmodules.usersessionadmin.UserNameComparator;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.querys.BooleanQuery;
import se.unlogic.standardutils.dao.querys.ObjectQuery;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.string.AnnotatedBeanTagSourceFactory;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.xml.XMLUtils;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Contact;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.booking.beans.Booking;
import com.nordicpeak.flowengine.booking.beans.BookingCalendar;
import com.nordicpeak.flowengine.booking.beans.BookingSlot;
import com.nordicpeak.flowengine.booking.beans.ManualContact;
import com.nordicpeak.flowengine.booking.beans.PrimaryContact;
import com.nordicpeak.flowengine.booking.enums.UserType;

public class BookingUtils {

	BookingUtils() {}

	private static final AnnotatedBeanTagSourceFactory<BookingSlot> BOOKING_SLOT_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(BookingSlot.class, "$bookingSlot.");

	public static boolean isActiveMemberOfBookingSlot(DataSource dataSource, BookingSlot bookingSlot, String citizenIdentifier) throws SQLException {

		//@formatter:off
		String queryString = 
			"SELECT x.bookingSlotID FROM (" +
				"SELECT DISTINCT b.bookingSlotID " +
				"FROM bookings_contacts c " +
				"INNER JOIN bookings b ON c.bookingID = b.bookingID " +
				"WHERE c.citizenIdentifier = ? " +
				"AND b.archived = 0" +
			") x " +
			"INNER JOIN booking_slots s ON s.bookingSlotID = x.bookingSlotID " +
			"WHERE s.bookingSlotID = ?";
		//@formatter:on

		BooleanQuery query = new BooleanQuery(dataSource, queryString);

		query.setString(1, citizenIdentifier);
		query.setInt(2, bookingSlot.getBookingSlotID());

		return query.executeQuery();
	}

	public static boolean hasActiveBookingInCalendar(DataSource dataSource, BookingCalendar calendar, String citizenIdentifier) throws SQLException {

		//@formatter:off
		String queryString =
			"SELECT bs.calendarID FROM booking_slots bs " +
			"INNER JOIN bookings b ON b.bookingSlotID = bs.bookingSlotID " +
			"INNER JOIN bookings_contacts co ON b.bookingID = co.bookingID " +
			"WHERE b.archived = 0 AND co.citizenIdentifier = ? AND bs.calendarID = ? AND bs.start >= ? LIMIT 1";
		//@formatter:on

		BooleanQuery query = new BooleanQuery(dataSource, queryString);

		query.setString(1, citizenIdentifier);
		query.setInt(2, calendar.getCalendarID());
		query.setDate(3, new Date(System.currentTimeMillis()));

		return query.executeQuery();
	}

	public static UserType getUserType(Booking booking, UserHandler userHandler, boolean includeSMS, boolean overrideContactBySMS) {

		User poster = booking.getPoster(userHandler);

		if (poster == null) {

			return UserType.ANALOG;
		}

		if (isUserContactable(booking, includeSMS, overrideContactBySMS)) {

			return UserType.DIGITAL;
		}

		return UserType.DIGITAL_NOT_CONTACTABLE;
	}

	private static boolean isUserContactable(Booking booking, boolean includeSMS, boolean overrideContactBySMS) {

		Contact contact = getContact(booking, null);

		if (contact == null) {

			return false;
		}

		if (StringUtils.isEmpty(contact.getEmail())) {

			if (!includeSMS) {

				return false;
			}

			if (!contact.isContactBySMS() && !overrideContactBySMS) {

				return false;
			}

			if (contact.getMobilePhone() == null) {

				return false;
			}
		}

		return true;
	}

	public static void setCalendarsTransientData(Collection<BookingCalendar> calendars, AnnotatedDAO<BookingCalendar> calendarDAO) throws SQLException {

		if (!CollectionUtils.isEmpty(calendars)) {

			for (BookingCalendar calendar : calendars) {

				BookingUtils.setCalendarTransientData(calendar, calendarDAO);
			}
		}
	}

	public static void setCalendarTransientData(BookingCalendar calendar, AnnotatedDAO<BookingCalendar> calendarDAO) throws SQLException {

		ObjectQuery<Integer> bookingSlotsQuery = new ObjectQuery<>(calendarDAO.getDataSource(), "SELECT COUNT(*) FROM booking_slots WHERE calendarID = ?", IntegerPopulator.getPopulator());

		bookingSlotsQuery.setInt(1, calendar.getCalendarID());

		calendar.setBookingSlotCount(bookingSlotsQuery.executeQuery());

		ObjectQuery<Integer> bookingsQuery = new ObjectQuery<>(calendarDAO.getDataSource(), "SELECT COUNT(b.bookingID) FROM bookings b INNER JOIN booking_slots s ON b.bookingSlotID = s.bookingSlotID WHERE s.calendarID = ? AND s.start >= ? AND b.archived = 0", IntegerPopulator.getPopulator());

		bookingsQuery.setInt(1, calendar.getCalendarID());
		bookingsQuery.setTimestamp(2, new Timestamp(System.currentTimeMillis()));

		calendar.setActiveBookings(bookingsQuery.executeQuery());
	}

	public static void setBookingSlotsTransientData(Collection<BookingSlot> bookingSlots, AnnotatedDAO<BookingSlot> bookingSlotDAO) throws SQLException {

		if (!CollectionUtils.isEmpty(bookingSlots)) {

			for (BookingSlot bookingSlot : bookingSlots) {

				BookingUtils.setBookingSlotTransientData(bookingSlot, bookingSlotDAO);
			}
		}
	}

	public static void setBookingSlotTransientData(BookingSlot bookingSlot, AnnotatedDAO<BookingSlot> bookingSlotDAO) throws SQLException {

		ObjectQuery<Integer> bookingSlotsQuery = new ObjectQuery<>(bookingSlotDAO.getDataSource(), "SELECT SUM(size) FROM bookings WHERE bookingSlotID = ? and archived = 0", IntegerPopulator.getPopulator());

		bookingSlotsQuery.setInt(1, bookingSlot.getBookingSlotID());

		bookingSlot.setBookingsCount(bookingSlotsQuery.executeQuery());
	}

	public static JsonNode getBookingSlotsJSON(BookingCalendar calendar, boolean admin, String preliminaryBookingText, List<BookingSlot> bookingSlots) {

		JsonObject jsonObject = new JsonObject();

		if (bookingSlots == null) {

			return jsonObject;
		}

		boolean appendManagers = admin || calendar.isShowManagers();
		boolean appendBookings = admin && calendar.isShowBookings();
		boolean showFreeSlotsAsBoolean = calendar.isShowFreeSlotsAsBoolean();

		JsonArray dayArray = null;
		BookingSlot prevSlot = null;

		for (Iterator<BookingSlot> it = bookingSlots.iterator(); it.hasNext();) {

			BookingSlot bookingSlot = it.next();

			if (!admin && !calendar.isShowNotBookableSlots() && !bookingSlot.isBookable()) {

				if (!it.hasNext() && prevSlot != null) {

					jsonObject.putField(DateUtils.DATE_FORMATTER.format(prevSlot.getStart()), dayArray);
				}

				continue;
			}

			if (prevSlot == null) {

				prevSlot = bookingSlot;
				dayArray = new JsonArray();

			} else if (!prevSlot.getStartDate().equals(bookingSlot.getStartDate())) {

				jsonObject.putField(DateUtils.DATE_FORMATTER.format(prevSlot.getStart()), dayArray);

				prevSlot = bookingSlot;
				dayArray = new JsonArray();
			}

			dayArray.addNode(bookingSlot.toJSON(appendManagers, appendBookings, preliminaryBookingText, showFreeSlotsAsBoolean));

			if (!it.hasNext()) {

				jsonObject.putField(DateUtils.DATE_FORMATTER.format(prevSlot.getStart()), dayArray);
			}
		}

		return jsonObject;
	}

	public static JsonArray getBookingSlotsTitleJSON(List<BookingSlot> bookingSlots, boolean admin) {

		return getBookingSlotsTitleJSON(bookingSlots, admin, null);
	}

	public static JsonArray getBookingSlotsTitleJSON(List<BookingSlot> bookingSlots, boolean admin, BookingCalendar calendar) {

		JsonArray jsonArray = new JsonArray();

		CollectionUtils.forEach(bookingSlots, slot -> appendBookingSlotTitleJSON(jsonArray, slot, slot.getFreeSlots(), calendar, admin));

		return jsonArray;
	}

	private static void appendBookingSlotTitleJSON(JsonArray jsonArray, BookingSlot bookingSlot, int freeSlots, BookingCalendar calendar, boolean admin) {

		if (calendar == null) {

			calendar = bookingSlot.getCalendar();
		}

		if (!admin && !calendar.isShowNotBookableSlots() && !bookingSlot.isBookable()) {

			return;
		}

		JsonObject json = new JsonObject();

		json.putField("id", bookingSlot.getBookingSlotID());
		json.putField("start", DateUtils.DATE_TIME_FORMATTER.format(bookingSlot.getStart()));
		json.putField("startTime", TimeUtils.TIME_FORMATTER.format(bookingSlot.getStart()));
		json.putField("end", DateUtils.DATE_TIME_FORMATTER.format(bookingSlot.getEnd()));
		json.putField("endTime", TimeUtils.TIME_FORMATTER.format(bookingSlot.getEnd()));
		json.putField("duration", (bookingSlot.getEnd().getTime() - bookingSlot.getStart().getTime()) / MillisecondTimeUnits.MINUTE);
		json.putField("freeSlots", freeSlots);
		json.putField("bookable", bookingSlot.isBookable());
		json.putField("location", bookingSlot.getLocation());

		json.putField("showFreeSlotsAsBoolean", calendar.isShowFreeSlotsAsBoolean());

		if (bookingSlot.hasPassed()) {

			json.putField("passed", 1);
		}

		jsonArray.addNode(json);
	}

	public static void appendCalendarManagers(Document doc, Element targetElement, List<BookingSlot> bookingSlots, boolean onlyEnabledManagers) {

		Element managersElement = XMLUtils.appendNewElement(doc, targetElement, "CalendarManagers");

		Collection<User> managers = CollectionUtils.flatMap(bookingSlots, BookingSlot::getManagers);

		if (!CollectionUtils.isEmpty(managers)) {

			Collection<User> sortedUniqueManagers = new TreeSet<>(UserNameComparator.getInstance());
			
			sortedUniqueManagers.addAll(managers);

			for (User manager : sortedUniqueManagers) {
				
				if (onlyEnabledManagers && !manager.isEnabled()) {
					
					continue;
				}

				Element managerElement = manager.toXML(doc);

				AttributeHandler managerAttributeHandler = manager.getAttributeHandler();

				if (managerAttributeHandler != null && !managerAttributeHandler.isEmpty() && managerAttributeHandler.isSet(BookingConstants.MANAGER_NAME_ATTRIBUTE)) {

					XMLUtils.appendNewElement(doc, managerElement, "CustomName", managerAttributeHandler.getString(BookingConstants.MANAGER_NAME_ATTRIBUTE));
				}

				managersElement.appendChild(managerElement);
			}
		}
	}

	public static void setUserElementValue(Document doc, Element userElement, String elementName, String elementValue) {

		Element existingElement = null;
		NodeList children = userElement.getElementsByTagName(elementName);

		if (children != null && children.getLength() > 0) {

			existingElement = (Element) children.item(0);
		}

		if (existingElement != null) {

			existingElement.setTextContent(elementValue);

		} else {

			XMLUtils.appendNewElement(doc, userElement, elementName, elementValue);
		}
	}

	public static Contact getContact(Booking booking, UserHandler userHandler) {

		PrimaryContact primaryContact = booking.getPrimaryContact();
		User poster = booking.getPoster(userHandler);

		if (poster != null) {

			return getContactFromPosterAndPrimaryContact(booking, poster, primaryContact);
		}

		ManualContact manualContact = booking.getManualContact();

		if (manualContact != null) {

			return getContactFromManualAndPrimaryContact(manualContact, primaryContact);
		}

		if (primaryContact != null) {

			return getContactFromPrimaryContact(primaryContact);
		}

		return null;
	}

	private static Contact getContactFromPosterAndPrimaryContact(Booking booking, User poster, PrimaryContact primaryContact) {

		Contact contact = new Contact();

		if (primaryContact != null && !StringUtils.isEmpty(primaryContact.getCitizenIdentifier())) {

			contact.setCitizenIdentifier(primaryContact.getCitizenIdentifier());

		} else if (booking.getBookingSlot().getCalendar().isUseCalendarForInternalUsers()) {

			contact.setCitizenIdentifier(UserUtils.getAttribute(Constants.MANAGER_CITIZEN_IDENTIFIER_ATTRIBUTE, poster));

		} else {

			contact.setCitizenIdentifier(UserUtils.getAttribute(Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE, poster));
		}

		contact.setFirstname(poster.getFirstname());
		contact.setLastname(poster.getLastname());
		contact.setEmail(poster.getEmail());
		contact.setMobilePhone(UserUtils.getAttribute("mobilePhone", poster));
		contact.setAddress(UserUtils.getAttribute("address", poster));
		contact.setPostalAddress(UserUtils.getAttribute("postalAddress", poster));
		contact.setZipCode(UserUtils.getAttribute("zipCode", poster));

		contact.setContactBySMS(UserUtils.getBooleanAttribute("contactBySMS", poster));
		contact.setContactByEmail(true);

		if (primaryContact != null) {

			if (StringUtils.isEmpty(contact.getEmail())) {

				contact.setEmail(primaryContact.getEmail());
			}

			if (StringUtils.isEmpty(contact.getMobilePhone())) {

				contact.setMobilePhone(primaryContact.getMobilePhone());
			}

			if (StringUtils.isEmpty(contact.getAddress())) {

				contact.setAddress(primaryContact.getAddress());
				contact.setPostalAddress(primaryContact.getPostalAddress());
				contact.setZipCode(primaryContact.getZipCode());
			}
		}

		return contact;
	}

	private static Contact getContactFromManualAndPrimaryContact(ManualContact manualContact, PrimaryContact primaryContact) {

		Contact contact = new Contact();

		if (primaryContact != null) {

			contact.setCitizenIdentifier(primaryContact.getCitizenIdentifier());
		}

		contact.setFirstname(manualContact.getFirstname());
		contact.setLastname(manualContact.getLastname());
		contact.setEmail(manualContact.getEmail());
		contact.setMobilePhone(manualContact.getMobilePhone());
		contact.setAddress(manualContact.getAddress());
		contact.setPostalAddress(manualContact.getPostalAddress());
		contact.setZipCode(manualContact.getZipCode());

		if (primaryContact != null) {

			if (StringUtils.isEmpty(contact.getFirstname())) {

				contact.setFirstname(primaryContact.getFirstname());
			}

			if (StringUtils.isEmpty(contact.getLastname())) {

				contact.setLastname(primaryContact.getLastname());
			}

			if (StringUtils.isEmpty(contact.getEmail())) {

				contact.setEmail(primaryContact.getEmail());
			}

			if (StringUtils.isEmpty(contact.getMobilePhone())) {

				contact.setMobilePhone(primaryContact.getMobilePhone());
			}

			if (StringUtils.isEmpty(contact.getAddress())) {

				contact.setAddress(primaryContact.getAddress());
			}

			if (StringUtils.isEmpty(contact.getPostalAddress())) {

				contact.setPostalAddress(primaryContact.getPostalAddress());
			}

			if (StringUtils.isEmpty(contact.getZipCode())) {

				contact.setZipCode(primaryContact.getZipCode());
			}
		}

		contact.setContactByEmail(true);

		if (!StringUtils.isEmpty(contact.getMobilePhone())) {

			contact.setContactBySMS(true);
		}

		return contact;
	}

	private static Contact getContactFromPrimaryContact(PrimaryContact primaryContact) {

		Contact contact = new Contact();

		contact.setCitizenIdentifier(primaryContact.getCitizenIdentifier());
		contact.setFirstname(primaryContact.getFirstname());
		contact.setLastname(primaryContact.getLastname());
		contact.setEmail(primaryContact.getEmail());
		contact.setMobilePhone(primaryContact.getMobilePhone());
		contact.setAddress(primaryContact.getAddress());
		contact.setPostalAddress(primaryContact.getPostalAddress());
		contact.setZipCode(primaryContact.getZipCode());

		contact.setContactByEmail(true);

		if (!StringUtils.isEmpty(contact.getMobilePhone())) {

			contact.setContactBySMS(true);
		}

		return contact;
	}

	public static Map<String, String> getBookingFlowInstanceAttributes(Booking booking, BookingCalendar calendar, FlowAdminModule flowAdminModule) throws SQLException {

		Integer flowInstanceID = booking.getFlowInstanceID();

		if (flowInstanceID == null) {

			return null;
		}

		Map<String, String> attributeMap = new HashMap<>();

		FlowInstance flowInstance = flowAdminModule.getFlowInstance(flowInstanceID);

		if (flowInstance.getAttributeHandler() != null && !CollectionUtils.isEmpty(calendar.getExportAttributes())) {

			List<String> exportAttributes = calendar.getExportAttributes();

			for (String exportAttribute : exportAttributes) {

				if (flowInstance.getAttributeHandler().getString(exportAttribute) != null) {

					attributeMap.put(exportAttribute, flowInstance.getAttributeHandler().getString(exportAttribute));
				}
			}
		}

		return attributeMap;
	}

	public static void setBookingSlotBookingUsers(BookingSlot bookingSlot, UserHandler userHandler) {

		setBookingSlotsBookingUsers(bookingSlot.getCalendar(), Collections.singletonList(bookingSlot), userHandler);
	}

	public static void setBookingSlotsBookingUsers(BookingCalendar calendar, List<BookingSlot> bookingSlots, UserHandler userHandler) {

		if (bookingSlots == null) {

			return;
		}

		List<Booking> bookings = CollectionUtils.flatMap(bookingSlots, BookingSlot::getBookings);

		bookings = CollectionUtils.filter(bookings, booking -> !StringUtils.isEmpty(booking.getPrimaryContactCitizenIdentifier()));

		Collection<String> citizenIdentifiers = CollectionUtils.map(bookings, Booking::getPrimaryContactCitizenIdentifier);

		if (citizenIdentifiers == null) {

			return;
		}

		citizenIdentifiers = new HashSet<>(citizenIdentifiers);

		if (citizenIdentifiers.isEmpty()) {

			return;
		}

		List<User> externalUsers = userHandler.getUsersByAttribute(Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE, citizenIdentifiers, false, true);
		List<User> internalUsers = userHandler.getUsersByAttribute(Constants.MANAGER_CITIZEN_IDENTIFIER_ATTRIBUTE, citizenIdentifiers, false, true);

		if (CollectionUtils.isEmpty(externalUsers) && CollectionUtils.isEmpty(internalUsers)) {

			return;
		}

		Map<String, User> externalUserMap = UserUtils.getUserAttributeMap(externalUsers, Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE);
		Map<String, User> internalUserMap = UserUtils.getUserAttributeMap(internalUsers, Constants.MANAGER_CITIZEN_IDENTIFIER_ATTRIBUTE);

		for (Booking booking : bookings) {

			String citizenIdentifier = booking.getPrimaryContactCitizenIdentifier();

			if (calendar.isUseCalendarForInternalUsers() && !CollectionUtils.isEmpty(internalUserMap)) {

				// For internal calendars, fetch citizen identifier from internalUserMap
				booking.setPoster(internalUserMap.get(citizenIdentifier));

			} else if (!CollectionUtils.isEmpty(externalUserMap) && externalUserMap.get(citizenIdentifier) != null) {

				booking.setPoster(externalUserMap.get(citizenIdentifier));

			} else if (!CollectionUtils.isEmpty(internalUserMap) && internalUserMap.get(citizenIdentifier) != null) {

				booking.setPoster(internalUserMap.get(citizenIdentifier));
			}
		}
	}

	public static String getManagerFirstNames(List<User> managers) {

		return getManagerNames(managers, true);
	}

	public static String getManagerNames(List<User> managers) {

		return getManagerNames(managers, false);
	}

	public static String getManagerNames(List<User> managers, boolean onlyFirstNames) {

		if (managers == null) {

			return "";
		}

		StringBuilder builder = new StringBuilder();

		for (Iterator<User> it = managers.iterator(); it.hasNext();) {

			User manager = it.next();

			if (manager.getAttributeHandler() != null && manager.getAttributeHandler().isSet(BookingConstants.MANAGER_NAME_ATTRIBUTE)) {

				builder.append(manager.getAttributeHandler().getString(BookingConstants.MANAGER_NAME_ATTRIBUTE));

			} else {

				builder.append(manager.getFirstname());

				if (!onlyFirstNames) {

					builder.append(" ");
					builder.append(manager.getLastname());
				}
			}

			if (it.hasNext()) {

				builder.append(", ");
			}
		}

		return builder.toString();
	}

	public static List<User> getManagers(BookingSlot slot, BookingCalendar calendar, UserHandler userHandler, boolean onlySlotManagers) {

		if (slot == null) {

			return null;
		}

		List<User> managers = null;

		if (!CollectionUtils.isEmpty(slot.getManagers())) {

			managers = new ArrayList<>(slot.getManagers());

		} else if (!onlySlotManagers && calendar != null && userHandler != null) {

			managers = new ArrayList<>();

			if (!CollectionUtils.isEmpty(calendar.getManagerUserIDs())) {

				List<User> users = userHandler.getUsers(calendar.getManagerUserIDs(), false, true);

				if (users != null) {

					managers.addAll(users);
				}
			}

			if (!CollectionUtils.isEmpty(calendar.getManagerGroupIDs())) {

				List<User> users = userHandler.getUsersByGroups(calendar.getManagerGroupIDs(), true);

				if (users != null) {

					managers.addAll(users);
				}
			}
		}

		if (CollectionUtils.isEmpty(managers)) {

			return null;
		}

		return managers;
	}

	public static void setNotBookableText(List<BookingSlot> bookingSlots, String bookingSlotNotBookableTextBetween, String bookingSlotNotBookableTextOpens, String bookingSlotNotBookableTextClosed) {

		CollectionUtils.forEach(bookingSlots, slot -> setNotBookableText(slot, bookingSlotNotBookableTextBetween, bookingSlotNotBookableTextOpens, bookingSlotNotBookableTextClosed));
	}

	public static void setNotBookableText(BookingSlot bookingSlot, String bookingSlotNotBookableTextBetween, String bookingSlotNotBookableTextOpens, String bookingSlotNotBookableTextClosed) {

		if (bookingSlot.isBookable()) {

			return;
		}

		TagReplacer tagReplacer = new TagReplacer();
		tagReplacer.addTagSource(BOOKING_SLOT_TAG_SOURCE_FACTORY.getTagSource(bookingSlot));

		if (bookingSlot.getBookableStart() != null && bookingSlot.getBookableEnd() != null) {

			bookingSlot.setNotBookableText(tagReplacer.replace(bookingSlotNotBookableTextBetween));

		} else if (bookingSlot.getBookableStart() != null) {

			bookingSlot.setNotBookableText(tagReplacer.replace(bookingSlotNotBookableTextOpens));

		} else if (bookingSlot.getBookableEnd() != null) {

			bookingSlot.setNotBookableText(tagReplacer.replace(bookingSlotNotBookableTextClosed));
		}
	}

	public static boolean isNewExternalMessagesAllowed(Booking booking) {

		if (booking == null) {

			return false;
		}

		if (booking.getBookingSlot().getCalendar().isHideExternalMessages()) {

			return false;
		}

		if (booking.getBookingSlot().getCalendar().getNewExternalMessagesAllowedDays() != null) {

			LocalDateTime bookingSlotStartDate = booking.getBookingSlot().getStart().toLocalDateTime();
			LocalDateTime lastAllowedDate = bookingSlotStartDate.plusDays(booking.getBookingSlot().getCalendar().getNewExternalMessagesAllowedDays());

			if (LocalDateTime.now().isAfter(lastAllowedDate)) {

				return false;
			}
		}

		return true;
	}

}
