package com.nordicpeak.flowengine.booking.beans;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.handlers.SourceAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeSource;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Key;
import se.unlogic.standardutils.dao.annotations.ManyToOne;
import se.unlogic.standardutils.dao.annotations.OneToMany;
import se.unlogic.standardutils.dao.annotations.OrderBy;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.xml.GeneratedElementable;
import se.unlogic.standardutils.xml.XMLElement;

import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.booking.enums.BookingEventType;
import com.nordicpeak.flowengine.enums.EventType;

@Table(name = "bookings_events")
@XMLElement
public class BookingEvent extends GeneratedElementable implements Serializable, AttributeSource {

	private static final long serialVersionUID = -7272910298485758740L;

	public static final Field QUEUE_SLOT_RELATION = ReflectionUtils.getField(BookingEvent.class, "booking");
	public static final Field ATTRIBUTES_RELATION = ReflectionUtils.getField(BookingEvent.class, "attributes");

	@DAOManaged(autoGenerated = true)
	@Key
	@XMLElement
	private Integer eventID;

	@DAOManaged(columnName = "bookingID")
	@ManyToOne
	private Booking booking;

	@DAOManaged
	@OrderBy
	@XMLElement
	protected Timestamp added;

	@DAOManaged(dontUpdateIfNull = true)
	@XMLElement(name = "poster")
	protected User poster;

	@DAOManaged
	@XMLElement
	private BookingEventType eventType;

	@DAOManaged
	@XMLElement
	protected String details;

	@DAOManaged
	@OneToMany
	@XMLElement(fixCase = true)
	private List<BookingEventAttribute> attributes;

	private SourceAttributeHandler attributeHandler;

	public BookingEvent() {}

	public BookingEvent(Timestamp added, User poster, BookingEventType eventType, String details, Map<String, String> attributes) {

		this.added = added;
		this.poster = poster;
		this.details = details;
		this.eventType = eventType;

		if (!CollectionUtils.isEmpty(attributes)) {

			for (Entry<String, String> attribute : attributes.entrySet()) {

				getAttributeHandler().setAttribute(attribute.getKey(), attribute.getValue());
			}
		}
	}

	public BookingEvent(FlowInstanceEvent flowInstanceEvent) {

		added = flowInstanceEvent.getAdded();
		poster = flowInstanceEvent.getPoster();

		if (flowInstanceEvent.getEventType() == EventType.PAYED) {

			eventType = BookingEventType.PAYED;

		} else {

			eventType = BookingEventType.OTHER;
		}

		details = flowInstanceEvent.getDetails();
		
		if (!CollectionUtils.isEmpty(flowInstanceEvent.getAttributes())) {

			flowInstanceEvent.getAttributeHandler().getAttributeMap().forEach((key, value) -> getAttributeHandler().setAttribute(key, value));
		}
	}

	public Integer getEventID() {

		return eventID;
	}

	public void setEventID(Integer eventID) {

		this.eventID = eventID;
	}

	public Timestamp getAdded() {

		return added;
	}

	public void setAdded(Timestamp added) {

		this.added = added;
	}

	public User getPoster() {

		return poster;
	}

	public void setPoster(User poster) {

		this.poster = poster;
	}

	public String getDetails() {

		return details;
	}

	public void setDetails(String details) {

		this.details = details;
	}

	public Booking getBookingSlot() {

		return booking;
	}

	public void setBookingSlot(Booking booking) {

		this.booking = booking;
	}

	public BookingEventType getEventType() {

		return eventType;
	}

	public void setEventType(BookingEventType eventType) {

		this.eventType = eventType;
	}

	@Override
	public String toString() {

		return "(ID=" + eventID + ", booking=" + booking + ", added=" + added + ", poster=" + poster + ", details=" + details + ")";
	}

	@Override
	public List<BookingEventAttribute> getAttributes() {

		return attributes;
	}

	public void setAttributes(List<BookingEventAttribute> attributes) {

		this.attributes = attributes;
	}

	@Override
	public void addAttribute(String name, String value) {

		if (this.attributes == null) {

			attributes = new ArrayList<>();
		}

		attributes.add(new BookingEventAttribute(name, value));
	}

	public synchronized MutableAttributeHandler getAttributeHandler() {

		if (attributeHandler == null) {

			this.attributeHandler = new SourceAttributeHandler(this, 255, 65536);
		}

		return attributeHandler;
	}

}
