package com.nordicpeak.flowengine.booking.beans;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Objects;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.standardutils.annotations.NoDuplicates;
import se.unlogic.standardutils.annotations.WebPopulate;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Key;
import se.unlogic.standardutils.dao.annotations.ManyToOne;
import se.unlogic.standardutils.dao.annotations.OneToMany;
import se.unlogic.standardutils.dao.annotations.OrderBy;
import se.unlogic.standardutils.dao.annotations.SimplifiedRelation;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.date.DateTimeStringyfier;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.string.StringTag;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.xml.GeneratedElementable;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLUtils;

import com.nordicpeak.flowengine.beans.Contact;
import com.nordicpeak.flowengine.booking.BookingConstants;
import com.nordicpeak.flowengine.booking.BookingUtils;
import com.nordicpeak.flowengine.booking.queries.BookingQueryInstance;
import com.nordicpeak.flowengine.interfaces.ImmutableAlternative;

@Table(name = "booking_slots")
@XMLElement
public class BookingSlot extends GeneratedElementable implements Serializable, ImmutableAlternative {

	private static final long serialVersionUID = 9041413820791856640L;

	private static final DateTimeFormatter DATE_FORMATTER = DateTimeFormatter.ofPattern("dd MMMM yyyy").withLocale(new Locale("sv"));

	public static final Field CALENDAR_RELATION = ReflectionUtils.getField(BookingSlot.class, "calendar");
	public static final Field BOOKINGS_RELATION = ReflectionUtils.getField(BookingSlot.class, "bookings");
	public static final Field REPEATING_RELATION = ReflectionUtils.getField(BookingSlot.class, "repeating");
	public static final Field MANAGERS_RELATION = ReflectionUtils.getField(BookingSlot.class, "managers");
	public static final Field FIELD_POSTER = ReflectionUtils.getField(BookingSlot.class, "poster");
	public static final Field FIELD_UPDATER = ReflectionUtils.getField(BookingSlot.class, "updater");
	public static final Field FIELD_DESCRIPTION = ReflectionUtils.getField(BookingSlot.class, "description");
	public static final Field FIELD_LOCATION = ReflectionUtils.getField(BookingSlot.class, "location");

	@DAOManaged(autoGenerated = true)
	@Key
	@XMLElement
	private Integer bookingSlotID;

	@DAOManaged(columnName = "calendarID")
	@ManyToOne
	@XMLElement
	private BookingCalendar calendar;

	@DAOManaged
	@StringTag(valueFormatter = DateTimeStringyfier.class)
	@OrderBy(priority = 0)
	@XMLElement
	private Timestamp start;

	@DAOManaged
	@StringTag(valueFormatter = DateTimeStringyfier.class)
	@OrderBy(priority = 1)
	@XMLElement
	private Timestamp end;

	@DAOManaged
	@WebPopulate(required = true)
	@XMLElement
	private int size;

	@DAOManaged
	@OrderBy(priority = 2)
	@XMLElement
	private Timestamp added;

	@DAOManaged
	@XMLElement
	private Timestamp updated;

	@DAOManaged
	private User poster;

	@DAOManaged
	private User updater;

	@DAOManaged
	@WebPopulate(maxLength = 65535)
	@XMLElement
	private String description;

	@DAOManaged
	@WebPopulate(maxLength = 255)
	@StringTag
	@XMLElement
	private String location;

	@DAOManaged
	@WebPopulate(maxLength = 255)
	@XMLElement
	private String title;

	@DAOManaged(columnName = "repeatID")
	@ManyToOne
	@XMLElement
	private Repeating repeating;

	@DAOManaged(dontUpdateIfNull = true)
	@OneToMany
	@SimplifiedRelation(table = "booking_slot_managers", remoteValueColumnName = "userID")
	@NoDuplicates
	private List<User> managers;

	@DAOManaged
	@OneToMany
	@XMLElement(fixCase = true)
	private List<Booking> bookings;

	@DAOManaged
	@OneToMany
	@XMLElement(fixCase = true)
	private List<BookingQueryInstance> bookingQueryInstances;

	@DAOManaged
	@XMLElement
	private boolean muteContactNotifications;

	@XMLElement
	private Integer bookingsCount;

	@DAOManaged
	@WebPopulate
	@XMLElement
	private boolean bookableRestricted;

	@DAOManaged
	@StringTag(valueFormatter = DateTimeStringyfier.class)
	@XMLElement
	private Timestamp bookableStart;

	@DAOManaged
	@StringTag(valueFormatter = DateTimeStringyfier.class)
	@XMLElement
	private Timestamp bookableEnd;

	@XMLElement
	private String notBookableText;

	public BookingSlot() {}

	public BookingSlot(BookingSlot bookingSlot) {

		calendar = bookingSlot.getCalendar();
		start = bookingSlot.getStart();
		end = bookingSlot.getEnd();
		size = bookingSlot.getSize();
		added = bookingSlot.getAdded();
		repeating = bookingSlot.getRepeating();
		description = bookingSlot.getDescription();
		managers = bookingSlot.getManagers();
		poster = bookingSlot.getPoster();
		location = bookingSlot.getLocation();
		title = bookingSlot.getTitle();
		muteContactNotifications = bookingSlot.isMuteContactNotifications();
		bookableRestricted = bookingSlot.isBookableRestricted();
		bookableStart = bookingSlot.getBookableStart();
		bookableEnd = bookingSlot.getBookableEnd();
	}

	public int getSize() {

		return size;
	}

	public void setSize(int size) {

		this.size = size;
	}

	public List<Booking> getBookings() {

		return bookings;
	}

	public void setBookings(List<Booking> bookings) {

		this.bookings = bookings;
	}

	public Repeating getRepeating() {

		return repeating;
	}

	public void setRepeating(Repeating repeating) {

		this.repeating = repeating;
	}

	public Integer getBookingSlotID() {

		return bookingSlotID;
	}

	public void setBookingSlotID(Integer bookingSlotID) {

		this.bookingSlotID = bookingSlotID;
	}

	public BookingCalendar getCalendar() {

		return calendar;
	}

	public void setCalendar(BookingCalendar calendar) {

		this.calendar = calendar;
	}

	public Timestamp getAdded() {

		return added;
	}

	public void setAdded(Timestamp added) {

		this.added = added;
	}

	public Timestamp getUpdated() {

		return updated;
	}

	public void setUpdated(Timestamp updated) {

		this.updated = updated;
	}

	public Timestamp getStart() {

		return start;
	}

	public Date getStartDate() {

		java.util.Calendar startDateCalendar = java.util.Calendar.getInstance();

		startDateCalendar.setTime(start);
		startDateCalendar.set(java.util.Calendar.HOUR_OF_DAY, 0);
		startDateCalendar.set(java.util.Calendar.MINUTE, 0);
		startDateCalendar.set(java.util.Calendar.SECOND, 0);
		startDateCalendar.set(java.util.Calendar.MILLISECOND, 0);

		return new Date(startDateCalendar.getTimeInMillis());
	}

	public String getTimeFormatted(BookingCalendar calendar) {

		StringBuilder timeStringBuilder = new StringBuilder();

		timeStringBuilder.append(TimeUtils.TIME_FORMATTER.format(start));

		if (calendar.isShowDuration()) {

			timeStringBuilder.append(" ");
			timeStringBuilder.append(getDurationInMinutes() + "min");

		} else {

			timeStringBuilder.append(" - ");
			timeStringBuilder.append(TimeUtils.TIME_FORMATTER.format(end));
		}

		return timeStringBuilder.toString();
	}

	public void setStart(Timestamp timestamp) {

		this.start = timestamp;
	}

	public Timestamp getEnd() {

		return end;
	}

	public void setEnd(Timestamp end) {

		this.end = end;
	}

	public User getPoster() {

		return poster;
	}

	public void setPoster(User poster) {

		this.poster = poster;
	}

	public User getUpdater() {

		return updater;
	}

	public void setUpdater(User updater) {

		this.updater = updater;
	}

	@Override
	public String toString() {

		if (calendar != null) {

			return "(ID: " + bookingSlotID + ")" + " in " + calendar;
		}

		if (bookingSlotID != null) {

			return "(ID: " + bookingSlotID + ")";
		}

		return null;
	}

	public int getDurationInMinutes() {

		return (int) ((end.getTime() - start.getTime()) / MillisecondTimeUnits.MINUTE);
	}

	@Override
	public Element toXML(Document doc) {

		Element element = super.toXML(doc);

		if (start != null) {

			XMLUtils.appendNewElement(doc, element, "startDate", DateUtils.DATE_FORMATTER.format(start));
			XMLUtils.appendNewElement(doc, element, "localizedStartDate", DATE_FORMATTER.format(start.toLocalDateTime()));
			XMLUtils.appendNewElement(doc, element, "startTime", TimeUtils.TIME_FORMATTER.format(start));

			XMLUtils.appendNewElement(doc, element, "endDate", DateUtils.DATE_FORMATTER.format(end));
			XMLUtils.appendNewElement(doc, element, "localizedEndDate", DATE_FORMATTER.format(end.toLocalDateTime()));
			XMLUtils.appendNewElement(doc, element, "endTime", TimeUtils.TIME_FORMATTER.format(end));
			XMLUtils.appendNewElement(doc, element, "duration", getDurationInMinutes());
		}

		if (poster != null) {

			Element posterElement = XMLUtils.appendNewElement(doc, element, "Poster");

			posterElement.appendChild(poster.toXML(doc));
		}

		if (updater != null) {

			Element updaterElement = XMLUtils.appendNewElement(doc, element, "Updater");

			updaterElement.appendChild(updater.toXML(doc));
		}

		if (hasPassed()) {

			XMLUtils.appendNewElement(doc, element, "HasPassed");
		}

		if (!CollectionUtils.isEmpty(managers)) {

			Element managersElement = XMLUtils.appendNewElement(doc, element, "Managers");

			for (User manager : managers) {

				Element userElement = manager.toXML(doc);

				AttributeHandler attributeHandler = manager.getAttributeHandler();

				if (attributeHandler != null && !attributeHandler.isEmpty()) {

					userElement.appendChild(attributeHandler.toXML(doc));

					BookingUtils.setUserElementValue(doc, userElement, "email", attributeHandler.getString(BookingConstants.MANAGER_NAME_ATTRIBUTE));

				}

				managersElement.appendChild(userElement);
			}
		}

		if (bookableStart != null) {

			XMLUtils.appendNewElement(doc, element, "bookableStartDate", DateUtils.DATE_FORMATTER.format(bookableStart));
			XMLUtils.appendNewElement(doc, element, "bookableStartTime", TimeUtils.TIME_FORMATTER.format(bookableStart));
		}

		if (bookableEnd != null) {

			XMLUtils.appendNewElement(doc, element, "bookableEndDate", DateUtils.DATE_FORMATTER.format(bookableEnd));
			XMLUtils.appendNewElement(doc, element, "bookableEndTime", TimeUtils.TIME_FORMATTER.format(bookableEnd));
		}

		XMLUtils.appendNewElement(doc, element, "bookable", isBookable());
		XMLUtils.appendNewElement(doc, element, "freeSlots", getFreeSlots());

		return element;
	}

	public JsonObject toJSON(boolean appendManagers, boolean appendBookings, String preliminaryBookingText, boolean showFreeSlotsAsBoolean) {

		JsonObject json = new JsonObject();

		json.putField("id", bookingSlotID);
		json.putField("description", description);
		json.putField("location", location);

		json.putField("start", DateUtils.DATE_TIME_FORMATTER.format(start));
		json.putField("startDate", DateUtils.DATE_FORMATTER.format(start));
		json.putField("startTime", TimeUtils.TIME_FORMATTER.format(start));

		json.putField("end", DateUtils.DATE_TIME_FORMATTER.format(end));
		json.putField("endDate", DateUtils.DATE_FORMATTER.format(end));
		json.putField("endTime", TimeUtils.TIME_FORMATTER.format(end));

		json.putField("duration", (end.getTime() - start.getTime()) / MillisecondTimeUnits.MINUTE);
		json.putField("freeSlots", getFreeSlots());
		json.putField("showFreeSlotsAsBoolean", showFreeSlotsAsBoolean);

		json.putField("bookable", isBookable());
		json.putField("notBookableText", notBookableText);

		if (appendManagers && !CollectionUtils.isEmpty(managers)) {

			JsonArray managersArray = new JsonArray(managers.size());

			for (User manager : managers) {

				JsonObject mangerJson = new JsonObject();

				AttributeHandler attributeHandler = manager.getAttributeHandler();

				if (attributeHandler != null && attributeHandler.isSet(BookingConstants.MANAGER_NAME_ATTRIBUTE)) {

					mangerJson.putField("name", attributeHandler.getString(BookingConstants.MANAGER_NAME_ATTRIBUTE));

				} else {

					mangerJson.putField("name", manager.getFirstname() + " " + manager.getLastname());
				}

				managersArray.addNode(mangerJson);
			}

			json.putField("managers", managersArray);
		}

		if (hasPassed()) {

			json.putField("passed", 1);
		}

		if (appendBookings && !CollectionUtils.isEmpty(bookings)) {

			JsonArray bookingsArray = new JsonArray(bookings.size());

			for (Booking booking : bookings) {

				if (booking.isArchived()) {

					continue;
				}

				if (!booking.isPreliminary()) {

					Contact contact = BookingUtils.getContact(booking, null);

					if (contact != null) {

						JsonObject bookingObject = new JsonObject();

						bookingObject.putField("firstname", contact.getFirstname());
						bookingObject.putField("lastname", contact.getLastname());

						bookingsArray.addNode(bookingObject);
					}

				} else {

					JsonObject bookingObject = new JsonObject();

					bookingObject.putField("firstname", preliminaryBookingText);
					bookingObject.putField("lastname", "");

					bookingsArray.addNode(bookingObject);
				}
			}

			if (bookingsArray.size() > 0) {

				json.putField("bookings", bookingsArray);
			}
		}

		if (!StringUtils.isEmpty(title)) {

			json.putField("title", title);
		}

		return json;
	}

	public boolean hasPassed() {

		return new Date().after(start);
	}

	@Override
	public Integer getAlternativeID() {

		return getBookingSlotID();
	}

	@Override
	public String getName() {

		return TimeUtils.TIME_FORMATTER.format(start);
	}

	@Override
	public String getExportXMLValue() {

		return null;
	}

	public Integer getBookingsCount() {

		return bookingsCount;
	}

	public void setBookingsCount(Integer bookingsCount) {

		this.bookingsCount = bookingsCount;
	}

	public String getDescription() {

		return description;
	}

	public void setDescription(String description) {

		this.description = description;
	}

	public int getFreeSlots() {
		
		if (bookingsCount == null) {
			
			return size;
		}

		return size - bookingsCount;
	}

	public List<User> getManagers() {

		return managers;
	}

	public void setManagers(List<User> managers) {

		this.managers = managers;
	}

	@Override
	public String getAttributeValue() {

		return null;
	}

	public String getLocation() {

		return location;
	}

	public void setLocation(String location) {

		this.location = location;
	}

	public String getTitle() {

		return title;
	}

	public void setTitle(String title) {

		this.title = title;
	}

	public boolean isMuteContactNotifications() {

		return muteContactNotifications;
	}

	public void setMuteContactNotifications(boolean muteContactNotifications) {

		this.muteContactNotifications = muteContactNotifications;
	}

	public boolean isBookableRestricted() {

		return bookableRestricted;
	}

	public void setBookableRestricted(boolean bookableRestricted) {

		this.bookableRestricted = bookableRestricted;
	}

	public Timestamp getBookableStart() {

		return bookableStart;
	}

	public void setBookableStart(Timestamp bookableStart) {

		this.bookableStart = bookableStart;
	}

	public Timestamp getBookableEnd() {

		return bookableEnd;
	}

	public void setBookableEnd(Timestamp bookableEnd) {

		this.bookableEnd = bookableEnd;
	}

	public boolean isBookable() {

		if (!bookableRestricted) {

			return true;
		}

		LocalDateTime now = LocalDateTime.now();

		return (bookableStart == null || bookableStart.toLocalDateTime().isBefore(now)) && (bookableEnd == null || bookableEnd.toLocalDateTime().isAfter(now));
	}

	public String getNotBookableText() {

		return notBookableText;
	}

	public void setNotBookableText(String notBookableText) {

		this.notBookableText = notBookableText;
	}

	@Override
	public int hashCode() {

		return Objects.hash(bookingSlotID);
	}

	@Override
	public boolean equals(Object obj) {

		if (this == obj) {
			return true;
		}
		if (!(obj instanceof BookingSlot)) {
			return false;
		}
		BookingSlot other = (BookingSlot) obj;
		return Objects.equals(bookingSlotID, other.bookingSlotID);
	}

}
