package com.nordicpeak.flowengine.integration.castor.client;

import java.io.File;
import java.lang.reflect.Field;
import java.nio.file.Files;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.List;

import javax.net.ssl.SSLContext;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.log4j.Logger;

import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.jaxws.WSUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.ssl.SSLUtils;

import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationRequestException;
import com.nordicpeak.flowengine.integration.castor.beans.AttachmentTypeMapping;
import com.nordicpeak.flowengine.integration.castor.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.castor.exceptions.IncompleteCaseDataException;
import com.nordicpeak.flowengine.integration.castor.exceptions.InvalidCaseIDException;
import com.nordicpeak.flowengine.integration.castor.exceptions.InvalidPartnerIDException;
import com.nordicpeak.flowengine.integration.castor.proxy.ArrayOfCastorDocumentData;
import com.nordicpeak.flowengine.integration.castor.proxy.CaseResult;
import com.nordicpeak.flowengine.integration.castor.proxy.CastorCaseTemplate;
import com.nordicpeak.flowengine.integration.castor.proxy.CastorDocumentData;
import com.nordicpeak.flowengine.integration.castor.proxy.CastorEventTypeEnum;
import com.nordicpeak.flowengine.integration.castor.proxy.CastorPersonData;
import com.nordicpeak.flowengine.integration.castor.proxy.CastorPersonTypeEnum;
import com.nordicpeak.flowengine.integration.castor.proxy.CcmService;
import com.nordicpeak.flowengine.integration.castor.proxy.CcmServiceSoap;
import com.nordicpeak.flowengine.utils.PDFByteAttachment;

public class CastorClient {

	protected final Logger log = Logger.getLogger(this.getClass());

	private final CcmServiceSoap castorProxy;
	private final IntegrationSettings settings;
	private final boolean logTemplate;

	public CastorClient(IntegrationSettings settings, Integer connectionTimeout, Integer requestTimeout, boolean logTemplate, boolean disableCertificateValidation) {

		this(settings, connectionTimeout, requestTimeout, logTemplate, disableCertificateValidation, new CcmService().getCcmServiceSoap());
	}

	public CastorClient(IntegrationSettings settings, Integer connectionTimeout, Integer requestTimeout, boolean logTemplate, boolean disableCertificateValidation, CcmServiceSoap castorProxy) {

		this.settings = settings;

		this.castorProxy = castorProxy;

		BindingProvider bindingProvider = (BindingProvider) this.castorProxy;

		WSUtils.setEndpoint(bindingProvider, settings.getEndpoint().getAddress());

		WSUtils.setTimeouts(bindingProvider, connectionTimeout, requestTimeout);

		if (disableCertificateValidation) {

			try {
				SSLContext sslContext = SSLContext.getInstance("SSL");
				sslContext.init(null, SSLUtils.INSECURE_TRUST_MANGERS, new java.security.SecureRandom());
				WSUtils.setSSLSocketFactory((BindingProvider) castorProxy, sslContext.getSocketFactory());

			} catch (KeyManagementException | NoSuchAlgorithmException e) {

				throw new RuntimeException(e);
			}
		}

		this.logTemplate = logTemplate;
	}

	public CaseResult createCastorCase(CaseData caseData) throws IntegrationRequestException {

		validateCaseData(caseData);

		CastorCaseTemplate template = getCaseTemplate();

		setPersonDatas(template, caseData);
		setPDFFile(template, caseData.getPdfFile());

		if (caseData.getPdfAttachments() != null) {

			setAttachmentFiles(template, caseData.getPdfAttachments());
		}

		if (logTemplate) {

			log.info("CastorCaseTemplate: " + template.getTemplateDescription() + " (" + template.getTemplateid() + ")");
			logObject(template.getPersondata());

			if (template.getPersons().getCastorPersonData() != null) {

				template.getPersons().getCastorPersonData().forEach(this::logObject);
			}

			if (template.getFiles() != null) {

				logObject(template.getFiles().getCastorDocumentData());
			}
		}

		return createCaseFromTemplate(template);
	}

	private void validateCaseData(CaseData caseData) throws IncompleteCaseDataException {

		if (caseData == null || !caseData.isPopulated()) {

			throw new IncompleteCaseDataException("Case data not fully populated (" + caseData + ")");
		}
	}

	protected CastorCaseTemplate getCaseTemplate() throws IntegrationRequestException {

		CastorCaseTemplate template = null;

		try {

			template = castorProxy.getCastorCaseTemplate(settings.getTemplateID(), settings.getPartnerID());

		} catch (SOAPFaultException e) {

			String faultString;
			if (e.getFault() != null && (faultString = e.getFault().getFaultString()) != null && faultString.contains("Guid should contain 32 digits with 4 dashes")) {

				throw new InvalidPartnerIDException(faultString);
			}

			throw new IntegrationRequestException("Unknown SOAP fault getting Castor case template", e);

		} catch (Exception e) {

			throw new IntegrationRequestException("Error getting Castor case template", e);
		}

		if (template == null) {

			throw new IntegrationRequestException("Castor template not found (templateID: " + settings.getTemplateID() + ", partnerID: " + settings.getPartnerID() + ")");
		}

		return template;
	}

	private void setPersonDatas(CastorCaseTemplate template, CaseData caseData) {

		setPersonData(template.getPersondata(), caseData.getApplicantData(), CastorPersonTypeEnum.SKANDE);

		List<CastorPersonData> optionalCastorPersonDatas = template.getPersons().getCastorPersonData();

		addOptionalPersonData(optionalCastorPersonDatas, caseData.getRealEstateOwnerData(), CastorPersonTypeEnum.FASTIGHETSGARE);
		addOptionalPersonData(optionalCastorPersonDatas, caseData.getInvoiceRecipientData(), CastorPersonTypeEnum.FAKTURAMOTTAGARE);
		addOptionalPersonData(optionalCastorPersonDatas, caseData.getInspectionManagerData(), CastorPersonTypeEnum.KONTROLLANSVARIG);
		addOptionalPersonData(optionalCastorPersonDatas, caseData.getContactPersonData(), CastorPersonTypeEnum.KONTAKTPERSON);
		addOptionalPersonData(optionalCastorPersonDatas, caseData.getAppellantData(), CastorPersonTypeEnum.KLAGANDE);
		addOptionalPersonData(optionalCastorPersonDatas, caseData.getCoApplicantData(), CastorPersonTypeEnum.MEDSKANDE);
		addOptionalPersonData(optionalCastorPersonDatas, caseData.getConstructionDeveloperData(), CastorPersonTypeEnum.BYGGHERRE);
	}

	private void addOptionalPersonData(List<CastorPersonData> castorPersonDatas, PersonData personData, CastorPersonTypeEnum type) {

		if (personData == null) {

			return;
		}

		CastorPersonData castorPersonData = new CastorPersonData();

		setPersonData(castorPersonData, personData, type);

		castorPersonDatas.add(castorPersonData);
	}

	private void setPersonData(CastorPersonData castorPersonData, PersonData personData, CastorPersonTypeEnum type) {

		castorPersonData.setNumber(personData.getNumber());
		castorPersonData.setName(personData.getName());
		castorPersonData.setNameCo(personData.getCareOf());
		castorPersonData.setAdress(personData.getAddress());
		castorPersonData.setPostalCode(personData.getZIPCode());
		castorPersonData.setCity(personData.getPostalAddress());
		castorPersonData.setPhone(personData.getPhone());
		castorPersonData.setMobilePhone(personData.getMobilePhone());
		castorPersonData.setEmail(personData.getEmail());
		castorPersonData.setRealestatename(personData.getRealEstateName());
		castorPersonData.setRealestatenumber(NumberUtils.toPrimitiveInt(NumberUtils.toInt(personData.getRealEstateNumber())));
		castorPersonData.setRealestateadress(personData.getRealEstateAddress());
		castorPersonData.setRealestatecity(personData.getRealEstatePostalAddress());
		castorPersonData.setRealestatepostnr(personData.getRealEstateZIPCode());

		castorPersonData.setPersontype(type);
	}

	protected void setPDFFile(CastorCaseTemplate template, File pdfFile) throws IntegrationRequestException {

		try {

			CastorDocumentData documentData = new CastorDocumentData();
			documentData.setData(Files.readAllBytes(pdfFile.toPath()));
			documentData.setEventType(CastorEventTypeEnum.CASE_PDF);
			documentData.setFilename(pdfFile.getName());
			documentData.setExtension(".pdf");
			documentData.setFileMappingID(template.getTemplateid() + "_1");

			ArrayOfCastorDocumentData documentDataArray = new ArrayOfCastorDocumentData();
			documentDataArray.getCastorDocumentData().add(documentData);

			template.setFiles(documentDataArray);

		} catch (Exception e) {

			throw new IntegrationRequestException("Error reading and adding pdf file " + pdfFile + "  to template", e);
		}
	}

	protected void setAttachmentFiles(CastorCaseTemplate template, List<PDFByteAttachment> list) {

		for (PDFByteAttachment attachment : list) {

			CastorDocumentData documentData = new CastorDocumentData();
			documentData.setData(attachment.getData());
			documentData.setEventType(CastorEventTypeEnum.DOKUMENT);
			documentData.setExtension("." + FileUtils.getFileExtension(attachment.getFilename()));
			documentData.setFilename(attachment.getFilename());

			Integer attachmentTypeID = getAttachmentTypeID(attachment);

			documentData.setFileMappingID(template.getTemplateid() + "_" + attachmentTypeID);

			template.getFiles().getCastorDocumentData().add(documentData);
		}
	}

	private Integer getAttachmentTypeID(PDFByteAttachment attachment) {

		if (settings.getAttachmentTypeMappings() != null) {

			for (AttachmentTypeMapping mapping : settings.getAttachmentTypeMappings()) {

				String attachmentName = attachment.getAttachmentName();
				String queryName = mapping.getQueryName();

				if (attachmentName != null && attachmentName.toLowerCase().contains(queryName.toLowerCase())) {

					return mapping.getAttachmentTypeID();
				}
			}
		}

		return settings.getDefaultAttachmentTypeID();
	}

	private void logObject(Object object) {

		List<Field> fields = ReflectionUtils.getFields(object.getClass());

		StringBuilder sb = new StringBuilder();

		sb.append(object.getClass().getSimpleName());

		try {

			if (object instanceof Collection<?>) {

				for (Object o : (Collection<?>) object) {

					logObject(o);
				}

			} else {

				for (Field field : fields) {

					ReflectionUtils.fixFieldAccess(field);

					sb.append(", " + field.getName() + ": " + field.get(object));
				}
			}

		} catch (IllegalArgumentException | IllegalAccessException e) {

			throw new RuntimeException(e);
		}

		log.info(sb.toString());
	}

	protected CaseResult createCaseFromTemplate(CastorCaseTemplate template) throws IntegrationRequestException {

		CaseResult result = null;

		try {

			result = castorProxy.createCastorCaseFromTemplate(template, settings.getPartnerID(), true);

		} catch (Exception e) {

			throw new IntegrationRequestException("Error creating Castor case from template", e);
		}

		if (result == null) {

			throw new IntegrationRequestException("No result for template (templateID: " + settings.getTemplateID() + ", partnerID: " + settings.getPartnerID() + ")");
		}

		if (result.getCaseID().equals("00000000-0000-0000-0000-000000000000")) {

			throw new InvalidCaseIDException("Received a CaseResult without valid case id: " + result.getCaseID());
		}

		return result;
	}

}