package com.nordicpeak.flowengine.contactdetailsverification;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.UUID;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.emailutils.framework.SimpleEmail;
import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.SimpleBackgroundModuleDescriptor;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.SimpleSMS;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.enums.PathType;
import se.unlogic.hierarchy.core.exceptions.CachePreconditionException;
import se.unlogic.hierarchy.core.exceptions.ModuleAlreadyBeingCachedException;
import se.unlogic.hierarchy.core.exceptions.ModuleAlreadyCachedException;
import se.unlogic.hierarchy.core.exceptions.ModuleInitializationException;
import se.unlogic.hierarchy.core.exceptions.ModuleInstasiationException;
import se.unlogic.hierarchy.core.exceptions.ModuleNotCachedException;
import se.unlogic.hierarchy.core.exceptions.ModuleUpdateException;
import se.unlogic.hierarchy.core.exceptions.ModuleUpdatedInProgressException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.exceptions.UnableToUpdateUserException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.sms.SMSSender;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.userprofile.events.UserUpdatedEvent;
import se.unlogic.standardutils.annotations.SplitOnLineBreak;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.querys.BooleanQuery;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.random.RandomUtils;
import se.unlogic.standardutils.string.SingleTagSource;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.notifications.StandardFlowNotificationHandler;

public class ContactDetailsVerificationModule extends AnnotatedForegroundModule {
	
	public static final String VERIFICATION_DECLINED = "verificationDeclined";
	public static final String PHONE_CODE = "verificationSMSCode";
	public static final String PHONE_UUID = "verificationSMSUUID";
	public static final String EMAIL_VERIFIED = "verificationEmail";
	public static final String EMAIL_VERIFIED_ADDRESS = "verificationEmailAddress";
	public static final String EMAIL_CODE = "verificationEmailCode";
	public static final String EMAIL_UUID = "verificationEmailUUID";
	public static final String PHONE_VERIFIED = "verificationPhone";
	public static final String PHONE_VERIFIED_NUMBER = "verificationPhoneNumber";

	private static final String ACTIVE_FLOW_INSTANCES_SQL = "SELECT * FROM flowengine_flow_instances fi INNER JOIN flowengine_flow_statuses fs ON (fi.statusID = fs.statusID) WHERE poster = ? AND fs.contentType IN ('SUBMITTED', 'IN_PROGRESS', 'WAITING_FOR_COMPLETION')";
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Verify only if active instances", description = "Verify contact details only if user has active instances")
	private boolean verifyOnlyIfActiveInstances = true;
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Verify phone number", description = "Activate phone number verification")
	private boolean verifyPhone = false;
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Verify email address", description = "Activate email address verification")
	private boolean verifyEmail = false;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Code length", description = "Length of validation code", formatValidator = NonNegativeStringIntegerValidator.class)
	private Integer codeLength = 8;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Phone validity days", description = "Number of days that phone verification is valid, 0 for permanent", formatValidator = NonNegativeStringIntegerValidator.class)
	private Integer phoneValidDays = 0;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Email validity days", description = "Number of days that email verification is valid, 0 for permanent", formatValidator = NonNegativeStringIntegerValidator.class)
	private Integer emailValidDays = 0;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Email subject", description = "Subject for verification email")
	private String emailSubject;
	
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Email message", description = "Message for verfication email, available tags are $verification.url, $verification.directurl, $verification.code")
	private String emailMessage;
	
	@ModuleSetting
	@TextAreaSettingDescriptor(name = "SMS message", description = "Message for verfication SMS, available tags are $verification.url, $verification.directurl, $verification.code")
	private String smsMessage;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Verification warning", description = "Verification warning text used to notify user", required = true)
	private String verificationWarningText;
	
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Verification declined", description = "Text shown when user declines verification")
	private String verificationDeclineText;
	
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Verification info", description = "Text shown before verification start")
	private String verificationText;
	
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Verification phone text", description = "Text shown after verification phone started")
	private String verificationPhoneText;
	
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Verification email text", description = "Text shown after verification email started")
	private String verificationEmailText;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Cellphone attribute", description = "Cellphone attribute")
	private String cellphoneAttribute = "mobilePhone";
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Public module alias", description = "Public module alias", required = true)
	private String publicModuleAlias = "publiccontactdetailsverification";
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Background module priority", description = "Background module priority", formatValidator = NonNegativeStringIntegerValidator.class)
	private Integer priority = 0;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Background module XSL", description = "Background module XSL")
	private String xslPath;
	
	@ModuleSetting
	@SplitOnLineBreak
	@TextAreaSettingDescriptor(name = "Background module aliases", description = "Background module aliases")
	private List<String> aliases = new ArrayList<String>();
	
	@ModuleSetting
	@SplitOnLineBreak
	@TextAreaSettingDescriptor(name = "Background module slots", description = "Background module slots")
	private List<String> slots = new ArrayList<String>();
	
	@InstanceManagerDependency
	private SMSSender smsSender;
	
	@InstanceManagerDependency(required = true)
	private StandardFlowNotificationHandler notificationHandler;
	
	private ContactDetailsVerificationBackgroundModule backgroundModule;
	
	private ContactDetailsVerificationPublicModule foregroundModule;
	
	@Override
	protected void moduleConfigured() throws Exception {

		addOrUpdateBackgroundModule();
		
		addOrUpdatePublicForegroundModule();
	}

	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse verify(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Document doc = createDocument(req, uriParser, user);
		Element verifyElement = doc.createElement("Verify");
		doc.getFirstChild().appendChild(verifyElement);
		
		if (!isPhoneValid(user)) {
			XMLUtils.appendNewElement(doc, verifyElement, "VerifyPhone");
		}
		
		if (!isEmailValid(user)) {
			XMLUtils.appendNewElement(doc, verifyElement, "VerifyEmail");
		}
		
		XMLUtils.appendNewElement(doc, verifyElement, "VerificationText", verificationText);
		
		return new SimpleForegroundModuleResponse(doc);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse verifyEmail(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (isEmailValid(user)) {
			throw new URINotFoundException(uriParser);
		}
		
		Document doc = createDocument(req, uriParser, user);
		Element verifyElement = doc.createElement("VerifyEmail");
		doc.getFirstChild().appendChild(verifyElement);
		
		XMLUtils.appendNewElement(doc, verifyElement, "VerifyEmailText", verificationEmailText);
		
		if (req.getMethod().equals("POST") && user.getAttributeHandler().getString(EMAIL_CODE) != null) {
			if (user.getAttributeHandler().getString(EMAIL_CODE).equals(req.getParameter("verificationCode"))) {
				
				User verificationUser = systemInterface.getUserHandler().getUser(user.getUserID(), false, true);
				
				XMLUtils.appendNewElement(doc, verifyElement, "EmailVerified");
				
				setEmailVerified(user, verificationUser);
			}
			else {
				log.info("Unable to verify email for user " + user + " with code " + req.getParameter("verificationCode"));
				
				XMLUtils.appendNewElement(doc, verifyElement, "EmailNotVerified");
			}
		}
		else {
			
			String uuid;
			String randomValue;
			
			if (!user.getAttributeHandler().isSet(EMAIL_CODE)) {
				User verificationUser = systemInterface.getUserHandler().getUser(user.getUserID(), false, true);
				
				uuid = UUID.randomUUID().toString();
				randomValue = RandomUtils.getRandomString(codeLength, RandomUtils.MIXED_CASE_CHARACTERS_AND_DIGITS);
				
				MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
				attributeHandler.setAttribute(EMAIL_CODE, randomValue);
				attributeHandler.setAttribute(EMAIL_UUID, uuid);
				
				MutableAttributeHandler vAttributeHandler = (MutableAttributeHandler) verificationUser.getAttributeHandler();
				vAttributeHandler.setAttribute(EMAIL_CODE, randomValue);
				vAttributeHandler.setAttribute(EMAIL_UUID, uuid);
				systemInterface.getUserHandler().updateUser(verificationUser, false, false, true);
			}
			else {
				uuid = user.getAttributeHandler().getString(EMAIL_UUID);
				randomValue = user.getAttributeHandler().getString(EMAIL_CODE);
			}
			
			TagReplacer tagReplacer = getTagReplacer(req, "/verifyemail", randomValue, uuid);
			
			SimpleEmail simpleEmail = new SimpleEmail(systemInterface.getEncoding());
			simpleEmail.setSenderAddress(notificationHandler.getEmailSenderAddress());
			simpleEmail.setSenderName(notificationHandler.getEmailSenderName(null));
			simpleEmail.setMessageContentType(SimpleEmail.HTML);
			simpleEmail.setSubject(emailSubject);
			simpleEmail.setMessage(tagReplacer.replace(emailMessage));
			simpleEmail.addRecipient(user.getEmail());
			
			log.info("Sending verification e-mail to " + user.getEmail() + " for user "  + user + " with code " + randomValue + " and uuid " + uuid);
			
			systemInterface.getEmailHandler().send(simpleEmail);
		}
		
		return new SimpleForegroundModuleResponse(doc);
	}

	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse verifyPhone(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (isPhoneValid(user)) {
			throw new URINotFoundException(uriParser);
		}
		
		Document doc = createDocument(req, uriParser, user);
		Element verifyElement = doc.createElement("VerifyPhone");
		doc.getFirstChild().appendChild(verifyElement);
		
		XMLUtils.appendNewElement(doc, verifyElement, "VerifyPhoneText", verificationPhoneText);
		
		if (req.getMethod().equals("POST") && user.getAttributeHandler().getString(PHONE_CODE) != null) {
			if (user.getAttributeHandler().getString(PHONE_CODE).equals(req.getParameter("verificationCode"))) {
				User verificationUser = systemInterface.getUserHandler().getUser(user.getUserID(), false, true);
				
				XMLUtils.appendNewElement(doc, verifyElement, "PhoneVerified");
				
				setPhoneVerified(user, verificationUser);
			}
			else {
				log.info("Unable to verify phone for user " + user + " with code " + req.getParameter("verificationCode"));
				
				XMLUtils.appendNewElement(doc, verifyElement, "PhoneNotVerified");
			}
		}
		else {
			
			String uuid;
			String randomValue;
			
			if (!user.getAttributeHandler().isSet(PHONE_CODE)) {
				User verificationUser = systemInterface.getUserHandler().getUser(user.getUserID(), false, true);
				
				uuid = UUID.randomUUID().toString();
				randomValue = RandomUtils.getRandomString(codeLength, RandomUtils.MIXED_CASE_CHARACTERS_AND_DIGITS);
				
				MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
				attributeHandler.setAttribute(PHONE_CODE, randomValue);
				attributeHandler.setAttribute(PHONE_UUID, uuid);
				
				MutableAttributeHandler vAttributeHandler = (MutableAttributeHandler) verificationUser.getAttributeHandler();
				vAttributeHandler.setAttribute(PHONE_CODE, randomValue);
				vAttributeHandler.setAttribute(PHONE_UUID, uuid);
				systemInterface.getUserHandler().updateUser(verificationUser, false, false, true);
			}
			else {
				uuid = user.getAttributeHandler().getString(PHONE_UUID);
				randomValue = user.getAttributeHandler().getString(PHONE_CODE);
			}
			
			TagReplacer tagReplacer = getTagReplacer(req, "/verifyphone", randomValue, uuid);
			
			String phone = user.getAttributeHandler().getString(cellphoneAttribute);
			
			SimpleSMS sms = new SimpleSMS();
			sms.setMessage(tagReplacer.replace(smsMessage));
			sms.setRecipients(Collections.singletonList(phone));
			sms.setSenderName(notificationHandler.getSMSSenderName(null));
			
			log.info("Sending verification sms to " + phone + " for user " + user + " with code " + randomValue + " and uuid " + uuid);
			
			smsSender.send(sms);
		}
		
		return new SimpleForegroundModuleResponse(doc);
	}
	
	public void setEmailVerified(User user, User verificationUser) throws UnableToUpdateUserException {

		String verifiedDate = DateUtils.DATE_TIME_SECONDS_FORMATTER.format(new Date());
		
		if (user != null) {
			MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
			attributeHandler.removeAttribute(EMAIL_CODE);
			attributeHandler.removeAttribute(EMAIL_UUID);
			attributeHandler.setAttribute(EMAIL_VERIFIED, verifiedDate);
			attributeHandler.setAttribute(EMAIL_VERIFIED_ADDRESS, user.getEmail());
		}
		
		MutableAttributeHandler vAttributeHandler = (MutableAttributeHandler) verificationUser.getAttributeHandler();
		vAttributeHandler.removeAttribute(EMAIL_CODE);
		vAttributeHandler.removeAttribute(EMAIL_UUID);
		vAttributeHandler.setAttribute(EMAIL_VERIFIED, verifiedDate);
		vAttributeHandler.setAttribute(EMAIL_VERIFIED_ADDRESS, verificationUser.getEmail());
		
		log.info("User " + verificationUser + " verified his email address");
		
		systemInterface.getUserHandler().updateUser(verificationUser, false, false, true);
	}

	public void setPhoneVerified(User user, User verificationUser) throws UnableToUpdateUserException {

		String verifiedDate = DateUtils.DATE_TIME_SECONDS_FORMATTER.format(new Date());
		
		if (user != null) {
			MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
			attributeHandler.removeAttribute(PHONE_CODE);
			attributeHandler.removeAttribute(PHONE_UUID);
			attributeHandler.setAttribute(PHONE_VERIFIED, verifiedDate);
			attributeHandler.setAttribute(PHONE_VERIFIED_NUMBER, attributeHandler.getString(cellphoneAttribute));
		}
		
		MutableAttributeHandler vAttributeHandler = (MutableAttributeHandler) verificationUser.getAttributeHandler();
		vAttributeHandler.removeAttribute(PHONE_CODE);
		vAttributeHandler.removeAttribute(PHONE_UUID);
		vAttributeHandler.setAttribute(PHONE_VERIFIED, verifiedDate);
		vAttributeHandler.setAttribute(PHONE_VERIFIED_NUMBER, vAttributeHandler.getString(cellphoneAttribute));
		
		log.info("User " + verificationUser + " verified his phone number");
		
		systemInterface.getUserHandler().updateUser(verificationUser, false, false, true);
	}
	
	public TagReplacer getTagReplacer(HttpServletRequest req, String alias, String randomValue, String uuid) {

		String fullContextPath = RequestUtils.getFullContextPathURL(req);
		
		TagReplacer tagReplacer = new TagReplacer();
		tagReplacer.addTagSource(new SingleTagSource("$verification.code", randomValue));
		tagReplacer.addTagSource(new SingleTagSource("$verification.url", fullContextPath + getFullAlias() + alias));
		tagReplacer.addTagSource(new SingleTagSource("$verification.directurl", fullContextPath + foregroundModule.getFullAlias() + alias + "?uuid=" + uuid));
		
		return tagReplacer;
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse decline(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (user == null) {
			throw new URINotFoundException(uriParser);
		}
		
		String declineDate = DateUtils.DATE_TIME_SECONDS_FORMATTER.format(new Date());
		
		MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
		attributeHandler.setAttribute(VERIFICATION_DECLINED, declineDate);
		
		User verificationUser = systemInterface.getUserHandler().getUser(user.getUserID(), false, true);
		MutableAttributeHandler vAttributeHandler = (MutableAttributeHandler) verificationUser.getAttributeHandler();
		vAttributeHandler.setAttribute(VERIFICATION_DECLINED, declineDate);
		
		systemInterface.getUserHandler().updateUser(verificationUser, false, false, true);
		
		Document doc = createDocument(req, uriParser, user);
		Element declineElement = doc.createElement("Decline");
		
		log.info("User " + verificationUser + " declined verifiying his contact details");
		
		XMLUtils.appendNewElement(doc, declineElement, "DeclineText", verificationDeclineText);
		
		doc.getFirstChild().appendChild(declineElement);
		
		return new SimpleForegroundModuleResponse(doc);
	}
	
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {
		
		Document doc = XMLUtils.createDomDocument();

		Element documentElement = doc.createElement("Document");
		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(documentElement);
		
		req.setAttribute("inVerificationProcess", true);

		return doc;
	}

	public boolean isVerified(User user, HttpServletRequest req) throws SQLException {
		
		if (user == null) {
			return true;
		}
		
		if (user.getAttributeHandler().isSet(VERIFICATION_DECLINED)) {
			return true;
		}
		
		if (verifyOnlyIfActiveInstances) {
			BooleanQuery query = new BooleanQuery(dataSource, ACTIVE_FLOW_INSTANCES_SQL);
			query.setInt(1, user.getUserID());
			
			if (!query.executeQuery()) {
				return true;
			}
		}
		
		return isPhoneValid(user) && isEmailValid(user);
	}
	
	private boolean isEmailValid(User user) {
		
		if (!verifyEmail) {
			return true;
		}
		
		if (StringUtils.isEmpty(user.getEmail())) {
			return true;
		}
		
		String emailVerified = user.getAttributeHandler().getString(EMAIL_VERIFIED);
		
		if (emailVerified != null) {
			if (emailValidDays > 0) {
				long verifiedDate = DateUtils.getDate(DateUtils.DATE_TIME_SECONDS_FORMATTER, emailVerified).getTime();
				long today = new Date().getTime();
				
				if ((long)emailValidDays * MillisecondTimeUnits.DAY < today - verifiedDate) {
					return false;
				}
			}
			
			return true;
		}
		
		return false;
	}

	private boolean isPhoneValid(User user) {
		
		if (!verifyPhone || smsSender == null) {
			return true;
		}
		
		String phone = user.getAttributeHandler().getString(cellphoneAttribute);
		
		if (phone == null) {
			return true;
		}

		String phoneVerified = user.getAttributeHandler().getString(PHONE_VERIFIED);
		
		if (phoneVerified != null) {
			if (phoneValidDays > 0) {
				long verifiedDate = DateUtils.getDate(DateUtils.DATE_TIME_SECONDS_FORMATTER, phoneVerified).getTime();
				long today = new Date().getTime();
				
				if ((long)phoneValidDays * MillisecondTimeUnits.DAY < today - verifiedDate) {
					return false;
				}
			}
			
			return true;
		}
		
		return false;
	}

	public String getVerificationWarningText() {
		
		return verificationWarningText;
	}
	
	public String getModuleURL(HttpServletRequest req) {
		
		return req.getContextPath() + getFullAlias();
	}
	
	public void addOrUpdateBackgroundModule() throws ModuleAlreadyCachedException, ModuleAlreadyBeingCachedException, ModuleInstasiationException, ModuleInitializationException, CachePreconditionException, ModuleNotCachedException, ModuleUpdateException, ModuleUpdatedInProgressException {

		if (backgroundModule == null) {
			log.info("Adding virtual background module to section " + sectionInterface.getSectionDescriptor());
			
			SimpleBackgroundModuleDescriptor descriptor = new SimpleBackgroundModuleDescriptor();
			descriptor.setAdminAccess(moduleDescriptor.allowsAdminAccess());
			descriptor.setAliases(aliases);
			
			if (!CollectionUtils.isEmpty(moduleDescriptor.getAllowedGroupIDs())) {
				descriptor.setAllowedGroupIDs(new ArrayList<Integer>(moduleDescriptor.getAllowedGroupIDs()));
			}
			
			if (!CollectionUtils.isEmpty(moduleDescriptor.getAllowedUserIDs())) {
				descriptor.setAllowedUserIDs(new ArrayList<Integer>(moduleDescriptor.getAllowedUserIDs()));
			}
			
			descriptor.setAnonymousAccess(moduleDescriptor.allowsAnonymousAccess());
			descriptor.setClassname(ContactDetailsVerificationBackgroundModule.class.getName());
			descriptor.setEnabled(true);
			descriptor.setName(moduleDescriptor.getName() + " (background)");
			descriptor.setPriority(priority);
			descriptor.setSectionID(moduleDescriptor.getSectionID());
			descriptor.setSlots(slots);
			descriptor.setStaticContentPackage(moduleDescriptor.getStaticContentPackage());
			descriptor.setUserAccess(moduleDescriptor.allowsUserAccess());
			descriptor.setXslPath(xslPath);
			descriptor.setXslPathType(PathType.Classpath);
			
			backgroundModule = (ContactDetailsVerificationBackgroundModule) sectionInterface.getBackgroundModuleCache().cache(descriptor);
			
			backgroundModule.setParentModule(this);
		}
		else {
			SimpleBackgroundModuleDescriptor descriptor = (SimpleBackgroundModuleDescriptor) backgroundModule.getDescriptor();
			
			log.info("Updating virtual background module " + descriptor);
			
			descriptor.setAdminAccess(moduleDescriptor.allowsAdminAccess());
			descriptor.setAliases(aliases);
			
			if (!CollectionUtils.isEmpty(moduleDescriptor.getAllowedGroupIDs())) {
				descriptor.setAllowedGroupIDs(new ArrayList<Integer>(moduleDescriptor.getAllowedGroupIDs()));
			}
			
			if (!CollectionUtils.isEmpty(moduleDescriptor.getAllowedUserIDs())) {
				descriptor.setAllowedUserIDs(new ArrayList<Integer>(moduleDescriptor.getAllowedUserIDs()));
			}
			
			descriptor.setAnonymousAccess(moduleDescriptor.allowsAnonymousAccess());
			descriptor.setName(moduleDescriptor.getName() + " (background)");
			descriptor.setPriority(priority);
			descriptor.setSectionID(moduleDescriptor.getSectionID());
			descriptor.setSlots(slots);
			descriptor.setStaticContentPackage(moduleDescriptor.getStaticContentPackage());
			descriptor.setUserAccess(moduleDescriptor.allowsUserAccess());
			descriptor.setXslPath(xslPath);
			
			sectionInterface.getBackgroundModuleCache().update(descriptor);
		}
	}
	
	public void addOrUpdatePublicForegroundModule() throws ModuleAlreadyCachedException, ModuleAlreadyBeingCachedException, ModuleInstasiationException, ModuleInitializationException, CachePreconditionException, ModuleNotCachedException, ModuleUpdateException, ModuleUpdatedInProgressException {

		if (foregroundModule == null) {
			log.info("Adding virtual public foreground module to section " + sectionInterface.getSectionDescriptor());
			
			SimpleForegroundModuleDescriptor descriptor = new SimpleForegroundModuleDescriptor();
			descriptor.setAlias(publicModuleAlias);
			descriptor.setDescription(moduleDescriptor.getDescription());
			descriptor.setDataSourceID(moduleDescriptor.getDataSourceID());
			descriptor.setAdminAccess(true);
			descriptor.setUserAccess(true);
			descriptor.setAnonymousAccess(true);
			descriptor.setClassname(ContactDetailsVerificationPublicModule.class.getName());
			descriptor.setEnabled(true);
			descriptor.setName(moduleDescriptor.getName() + " (public)");
			descriptor.setSectionID(moduleDescriptor.getSectionID());
			descriptor.setStaticContentPackage(moduleDescriptor.getStaticContentPackage());
			descriptor.setXslPath(moduleDescriptor.getXslPath());
			descriptor.setXslPathType(moduleDescriptor.getXslPathType());
			descriptor.setMutableSettingHandler(moduleDescriptor.getMutableSettingHandler());
			
			foregroundModule = (ContactDetailsVerificationPublicModule) sectionInterface.getForegroundModuleCache().cache(descriptor);
			
			foregroundModule.setParentModule(this);
		}
		else {
			SimpleForegroundModuleDescriptor descriptor = (SimpleForegroundModuleDescriptor) foregroundModule.getDescriptor();
			
			log.info("Updating virtual public foreground module " + descriptor);
			
			descriptor.setName(moduleDescriptor.getName() + " (public)");
			descriptor.setDescription(moduleDescriptor.getDescription());
			descriptor.setAlias(publicModuleAlias);
			descriptor.setDataSourceID(moduleDescriptor.getDataSourceID());
			descriptor.setSectionID(moduleDescriptor.getSectionID());
			descriptor.setStaticContentPackage(moduleDescriptor.getStaticContentPackage());
			descriptor.setXslPath(moduleDescriptor.getXslPath());
			descriptor.setXslPathType(moduleDescriptor.getXslPathType());
			
			sectionInterface.getForegroundModuleCache().update(descriptor);
		}
	}
	
	@EventListener(channel = User.class)
	public void processEvent(UserUpdatedEvent event, EventSource source) throws UnableToUpdateUserException {
		
		if (event.getUser().getAttributeHandler() != null) {
			boolean phoneOrEmailChanged = false;
			
			User user = event.getUser();
			
			String verifiedEmail = user.getAttributeHandler().getString(EMAIL_VERIFIED_ADDRESS);
			String verifiedPhone = user.getAttributeHandler().getString(PHONE_VERIFIED_NUMBER);
			
			if ((user.getEmail() != null && verifiedEmail != null && !verifiedEmail.equals(user.getEmail())) || (user.getEmail() == null && verifiedEmail != null)) {
				
				MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
				attributeHandler.removeAttribute(EMAIL_VERIFIED);
				attributeHandler.removeAttribute(EMAIL_VERIFIED_ADDRESS);
				
				log.info("Setting email address not verified for user " + user);
				
				phoneOrEmailChanged = true;
			}
			
			String userCellPhone = user.getAttributeHandler().getString(cellphoneAttribute);
			
			if ((userCellPhone != null && verifiedPhone != null && !verifiedPhone.equals(userCellPhone)) || (userCellPhone == null && verifiedPhone != null)) {
				
				MutableAttributeHandler attributeHandler = (MutableAttributeHandler) user.getAttributeHandler();
				attributeHandler.removeAttribute(PHONE_VERIFIED);
				attributeHandler.removeAttribute(PHONE_VERIFIED_NUMBER);
				
				log.info("Setting phone number not verified for user " + user);
				
				phoneOrEmailChanged = true;
			}
			
			if (phoneOrEmailChanged) {
				systemInterface.getUserHandler().updateUser(user, false, false, true);
			}
		}
	}
}