package com.nordicpeak.flowengine.documentsigner;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.MutableUser;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.exceptions.UnableToUpdateUserException;
import se.unlogic.hierarchy.core.interfaces.AccessInterface;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.ViewFragmentUtils;
import se.unlogic.hierarchy.core.utils.crud.FragmentLinkScriptFilter;
import se.unlogic.hierarchy.core.utils.extensionlinks.ExtensionLink;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.standardutils.bool.BooleanUtils;
import se.unlogic.standardutils.dao.AdvancedAnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.LowLevelQuery;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.hash.HashUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.http.enums.ContentDisposition;

import com.nordicpeak.flowengine.UserFlowInstanceMenuModule;
import com.nordicpeak.flowengine.beans.SimpleSigningRequest;
import com.nordicpeak.flowengine.documentsigner.beans.DocumentSigner;
import com.nordicpeak.flowengine.documentsigner.beans.RequestData;
import com.nordicpeak.flowengine.documentsigner.beans.SigningDocument;
import com.nordicpeak.flowengine.documentsigner.enums.SignerType;
import com.nordicpeak.flowengine.documentsigner.enums.SigningDocumentStatus;
import com.nordicpeak.flowengine.interfaces.GenericSigningProvider;
import com.nordicpeak.flowengine.interfaces.SigningResponse;
import com.nordicpeak.flowengine.interfaces.UserMenuProvider;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

public class DocumentSignerUserModule extends AnnotatedRESTModule implements UserMenuProvider {

	// @formatter:off
	private static final String LIST_PENDING_SQL = "SELECT d.* FROM document_signer_documents d"
		                                   +" INNER JOIN document_signer_document_signers s ON d.documentID = s.documentID"
		                                   +" WHERE (s.userID = ? OR (s.userID IS NULL AND s.citizenID = ?)) AND d.status IN ('SIGNING', 'SIGNED') AND s.signedDate IS NULL";
	
	private static final String LIST_SIGNED_SQL = "SELECT d.* FROM document_signer_documents d"
		                                   +" INNER JOIN document_signer_document_signers s ON d.documentID = s.documentID"
		                                   +" WHERE s.userID = ? AND d.status IN ('SIGNING', 'SIGNED') AND s.signedDate IS NOT NULL AND showSignedDocumentInUserModule = 1";
	// @formatter:on

	@XSLVariable(prefix = "java.")
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Signing disclaimer", description = "Disclaimer text displayed on signing page", required = true)
	private String disclaimerText;

	@XSLVariable(prefix = "java.")
	private String userMenuTabTitle = "Documents";

	@XSLVariable(prefix = "java.")
	protected String signingMessage = "You will sign $document.name with application number $document.documentID. The application has the following unique key: $hash ($hashAlgorithm)";

	@XSLVariable(prefix = "java.")
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Sign message", description = "The message displayed above the sign form (available tags: $poster, $document.documentID, $document.name)", required = true)
	protected String signMessage = "Not set";

	@XSLVariable(prefix = "java.")
	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Signed message", description = "The message displayed above the signed form (available tags: $poster, $document.documentID, $document.name, $signed)", required = true)
	protected String signedMessage = "Not set";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "User menu item slot", description = "User menu item slot")
	protected String userMenuExtensionLinkSlot = "30";

	private AnnotatedDAO<SigningDocument> documentDAO;
	private AnnotatedDAO<DocumentSigner> signerDAO;

	private AdvancedAnnotatedDAOWrapper<SigningDocument, Integer> documentDAOWrapper;
	private AdvancedAnnotatedDAOWrapper<DocumentSigner, Integer> signerDAOWrapper;

	@InstanceManagerDependency(required = true)
	protected DocumentSignerAdminModule adminModule;

	protected UserFlowInstanceMenuModule userFlowInstanceMenuModule;

	@InstanceManagerDependency(required = true)
	protected GenericSigningProvider genericSigningProvider;

	protected ExtensionLink userMenuLink;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		if (!systemInterface.getInstanceHandler().addInstance(DocumentSignerUserModule.class, this)) {

			throw new RuntimeException("Unable to register module in global instance handler using key " + DocumentSignerUserModule.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, DocumentSignerAdminModule.class.getName(), new XMLDBScriptProvider(DocumentSignerAdminModule.class.getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler(), true, true, false);

		documentDAO = daoFactory.getDAO(SigningDocument.class);
		signerDAO = daoFactory.getDAO(DocumentSigner.class);

		documentDAOWrapper = documentDAO.getAdvancedWrapper(Integer.class);
		documentDAOWrapper.getGetQuery().addRelations(SigningDocument.SIGNERS_RELATION);

		signerDAOWrapper = signerDAO.getAdvancedWrapper(Integer.class);
	}

	@InstanceManagerDependency
	public void setUserFlowInstanceMenuModule(UserFlowInstanceMenuModule userFlowInstanceMenuModule) {

		if (userFlowInstanceMenuModule == null && this.userFlowInstanceMenuModule != null) {

			this.userFlowInstanceMenuModule.removeUserMenuProvider(this);
		}

		this.userFlowInstanceMenuModule = userFlowInstanceMenuModule;

		if (this.userFlowInstanceMenuModule != null) {

			this.userFlowInstanceMenuModule.addUserMenuProvider(this);
		}
	}

	@Override
	public void unload() throws Exception {

		if (userFlowInstanceMenuModule != null) {

			setUserFlowInstanceMenuModule(null);
		}

		systemInterface.getInstanceHandler().removeInstance(DocumentSignerUserModule.class, this);

		super.unload();
	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		userMenuLink = new ExtensionLink(userMenuTabTitle, getFullAlias(), "b", userMenuExtensionLinkSlot);

		if (userFlowInstanceMenuModule != null) {

			userFlowInstanceMenuModule.sortProviders();
		}
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return listDocuments(req, user, uriParser, null);
	}

	public ForegroundModuleResponse listDocuments(HttpServletRequest req, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Throwable {

		if (user == null) {
			throw new AccessDeniedException("User needs to be logged in");
		}

		log.info("User " + user + " listing documents");

		Document doc = createDocument(req, uriParser, user);

		Element listDocumentsElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "ListDocuments");

		XMLUtils.append(doc, listDocumentsElement, "PendingDocuments", getPendingDocuments(user));
		XMLUtils.append(doc, listDocumentsElement, "SignedDocuments", getSignedDocuments(user));

		XMLUtils.appendNewElement(doc, listDocumentsElement, "Title", userMenuTabTitle);

		if (validationErrors != null) {

			listDocumentsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
			XMLUtils.append(doc, listDocumentsElement, "ValidationErrors", validationErrors);
		}

		return new SimpleForegroundModuleResponse(doc, moduleDescriptor.getName(), getDefaultBreadcrumb());
	}

	public List<SigningDocument> getPendingDocuments(User user) throws SQLException {

		LowLevelQuery<SigningDocument> query = new LowLevelQuery<>(LIST_PENDING_SQL);
		query.addParameter(user.getUserID());
		query.addParameter(CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(user));

		query.addRelation(SigningDocument.SIGNERS_RELATION);

		List<SigningDocument> pendingDocuments = documentDAO.getAll(query);

		if (pendingDocuments != null) {

			Iterator<SigningDocument> it = pendingDocuments.iterator();
			while (it.hasNext()) {

				SigningDocument signingDocument = it.next();

				if (signingDocument.isUseSequentialSigning()) {

					DocumentSigner signer = signingDocument.getSigner(user, null);

					if (isSequentialSigningBlocked(signingDocument, signer)) {
						it.remove();
					}
				}
			}
		}

		return pendingDocuments;
	}

	public List<SigningDocument> getSignedDocuments(User user, Field... relations) throws SQLException {

		LowLevelQuery<SigningDocument> query = new LowLevelQuery<>(LIST_SIGNED_SQL);
		query.addParameter(user.getUserID());

		if (relations != null) {
			query.addRelations(relations);
		}

		return documentDAO.getAll(query);
	}

	public boolean hasDocuments(User user) throws SQLException {

		LowLevelQuery<SigningDocument> query = new LowLevelQuery<>(LIST_SIGNED_SQL + " LIMIT 1");
		query.addParameter(user.getUserID());

		if (documentDAO.getBoolean(query)) {

			return true;
		}

		if (getPendingDocuments(user) != null) {

			return true;
		}

		return false;
	}

	@WebPublic(alias = "unsignedpdf", requireLogin = true)
	public ForegroundModuleResponse getUnsignedPDF(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws URINotFoundException, SQLException, AccessDeniedException {

		RequestData requestData = getRequestData(user, req, uriParser);

		if (requestData.getModuleResponse() != null) {
			return requestData.getModuleResponse();
		}

		SigningDocument signingDocument = requestData.getSigningDocument();

		File pdfFile = adminModule.getOriginalPDF(signingDocument);

		if (pdfFile != null) {

			log.info("Sending unsigned PDF for " + signingDocument + " to user " + user);

			try {
				HTTPUtils.sendFile(pdfFile, signingDocument.getName() + ".pdf", req, res, ContentDisposition.ATTACHMENT);
				return null;

			} catch (IOException e) {
				log.info("Error sending unsigned PDF for  " + signingDocument + " to user " + user + ", " + e);
			}
		}

		log.warn("Unable to find unsigned PDF for " + signingDocument + " requested by user " + user);
		throw new URINotFoundException(uriParser);
	}

	@WebPublic(alias = "signedpdf", requireLogin = true)
	public ForegroundModuleResponse getSignedPDF(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws URINotFoundException, SQLException, AccessDeniedException {

		RequestData requestData = getRequestData(user, req, uriParser);

		if (requestData.getModuleResponse() != null) {

			return requestData.getModuleResponse();
		}

		SigningDocument signingDocument = requestData.getSigningDocument();

		if (signingDocument.isPublishOriginal()) {

			File pdfFile = adminModule.getSignedPDF(signingDocument);

			if (pdfFile != null) {

				log.info("Sending signed PDF for " + signingDocument + " to user " + user);

				try {
					HTTPUtils.sendFile(pdfFile, signingDocument.getName() + ".pdf", req, res, ContentDisposition.ATTACHMENT);
					return null;

				} catch (IOException e) {
					log.info("Error sending signed PDF for " + signingDocument + " to user " + user + ", " + e);
				}
			}

			log.warn("Unable to find signed PDF for " + signingDocument + " requested by user " + user);
			throw new URINotFoundException(uriParser);

		} else {

			File pdfFile = adminModule.getSignReceiptPDF(signingDocument);

			if (pdfFile != null) {

				log.info("Sending sign receipt PDF for " + signingDocument + " to user " + user);

				try {
					HTTPUtils.sendFile(pdfFile, signingDocument.getName() + ".pdf", req, res, ContentDisposition.ATTACHMENT);
					return null;

				} catch (IOException e) {
					log.info("Error sending sign receipt PDF for " + signingDocument + " to user " + user + ", " + e);
				}
			}

			log.warn("Unable to find sign receipt PDF for " + signingDocument + " requested by user " + user);
			throw new URINotFoundException(uriParser);
		}
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.moduleDescriptor.toXML(doc));
		doc.appendChild(document);
		return doc;
	}

	@Override
	public AccessInterface getAccessInterface() {

		return moduleDescriptor;
	}

	@Override
	public String getUserMenuAlias() {

		return getFullAlias();
	}

	@Override
	public ExtensionLink getUserMenuExtensionLink(User user) {

		try {
			if (hasDocuments(user)) {

				return userMenuLink;
			}

		} catch (SQLException e) {

			log.error("Error checking of user " + user + " has signed or pending documents", e);
		}

		return null;
	}

	@WebPublic(alias = "sign", requireLogin = true)
	public ForegroundModuleResponse signDocument(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		RequestData requestData = getRequestData(user, req, uriParser);

		if (requestData.getModuleResponse() != null) {
			return requestData.getModuleResponse();
		}

		SigningDocument signingDocument = requestData.getSigningDocument();
		DocumentSigner signer = requestData.getSigner();

		boolean sequentialSigningBlock = signingDocument.isUseSequentialSigning() && isSequentialSigningBlocked(signingDocument, signer);

		if (signer.getSignedDate() == null) {

			if (signingDocument.getStatus() != SigningDocumentStatus.SIGNING) {

				//User does not have access to sign this flow instance
				log.info("User does not have access to " + signingDocument + ", wrong status : " + signingDocument.getStatus());
				return showMessage("WrongDocumentStatus", req, uriParser);
			}

			if (BooleanUtils.toBoolean(req.getParameter("sign"))) {

				if (sequentialSigningBlock) {

					log.info("User does not have access to sign " + signingDocument + " yet, as sequential signing is active and not all preceeding parties have signed.");
					return showMessage("SequentialSigningBlock", req, uriParser);
				}

				log.info("User " + user + " preparing signing of " + signingDocument);

				return showSignForm(req, res, user, uriParser, signingDocument, signer, null);
			}
		}

		log.info("User " + user + " viewing sign form for " + signingDocument);

		Document doc = createDocument(req, uriParser);

		Element signFlowInstanceElement = doc.createElement("SignDocument");
		doc.getDocumentElement().appendChild(signFlowInstanceElement);

		if (signer.getSignedDate() == null) {

			XMLUtils.appendNewCDATAElement(doc, signFlowInstanceElement, "Message", getMessage(user, signingDocument, signMessage, signer));
			XMLUtils.appendNewCDATAElement(doc, signFlowInstanceElement, "Disclaimer", disclaimerText);

		} else {

			XMLUtils.appendNewCDATAElement(doc, signFlowInstanceElement, "Message", getMessage(user, signingDocument, signedMessage, signer));
		}

		XMLUtils.append(doc, signFlowInstanceElement, signingDocument);
		XMLUtils.append(doc, signFlowInstanceElement, signer);

		if (sequentialSigningBlock) {

			XMLUtils.appendNewElement(doc, signFlowInstanceElement, "SequentialSigningBlock");
		}

		return new SimpleForegroundModuleResponse(doc, getDefaultBreadcrumb());
	}

	private SimpleSigningRequest getSigningRequest(SigningDocument signingDocument, DocumentSigner signer, URIParser uriParser) throws IOException {

		File unsignedPDF = adminModule.getOriginalPDF(signingDocument);

		if (unsignedPDF == null) {
			return null;
		}

		String dataToSign = HashUtils.hash(unsignedPDF, signingDocument.getHashAlgorithm());

		String description = signingMessage.replace("$document.documentID", signingDocument.getDocumentID() + "");
		description = description.replace("$document.name", signingDocument.getName());
		description = description.replace("$hashAlgorithm", signingDocument.getHashAlgorithm());
		description = description.replace("$hash", dataToSign);

		String urlEnd = signingDocument.getDocumentID().toString();

		if (signer.isUnsecure()) {

			urlEnd += "/" + signer.getUnsecureKey();
		}

		String signingFormURL = uriParser.getFullContextPath() + getFullAlias() + "/sign/" + urlEnd;
		String processSigningURL = uriParser.getFullContextPath() + getFullAlias() + "/processsign/" + urlEnd + "?signreq=1";

		return new SimpleSigningRequest(description, unsignedPDF, signingDocument.getHashAlgorithm(), signingFormURL, processSigningURL);
	}

	private ForegroundModuleResponse showSignForm(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, SigningDocument signingDocument, DocumentSigner signer, List<ValidationError> validationErrors) throws Exception {

		Document doc = createDocument(req, uriParser);

		Element signingFormElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "Signing");

		SimpleSigningRequest signingRequest = getSigningRequest(signingDocument, signer, uriParser);

		if (signingRequest == null) {
			throw new RuntimeException("Unable to find signing PDF for " + signingDocument);
		}

		ViewFragment viewFragment = genericSigningProvider.showSignForm(req, res, user, signingRequest, validationErrors);

		signingFormElement.appendChild(signingDocument.toXML(doc));
		signingFormElement.appendChild(viewFragment.toXML(doc));

		XMLUtils.appendNewElement(doc, signingFormElement, "SigningLink", adminModule.getSigningURL(signingDocument, signer, uriParser));

		FragmentLinkScriptFilter.addViewFragment(viewFragment, req);

		SimpleForegroundModuleResponse response = new SimpleForegroundModuleResponse(doc, moduleDescriptor.getName(), getDefaultBreadcrumb());
		ViewFragmentUtils.appendLinksAndScripts(response, viewFragment);

		return response;
	}

	@WebPublic(alias = "processsign", requireLogin = true)
	public ForegroundModuleResponse processSigning(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		RequestData requestData = getRequestData(user, req, uriParser);

		if (requestData.getModuleResponse() != null) {
			return requestData.getModuleResponse();
		}

		SigningDocument signingDocument = requestData.getSigningDocument();
		DocumentSigner signer = requestData.getSigner();

		if (signer.getSignedDate() != null) {

			return signDocument(req, res, user, uriParser);
		}

		if (signingDocument.getStatus() != SigningDocumentStatus.SIGNING) {

			//User does not have access to sign this flow instance
			log.info("User does not have access to  document " + signingDocument + ", wrong status : " + signingDocument.getStatus());
			return showMessage("WrongDocumentStatus", req, uriParser);
		}

		if (signingDocument.isUseSequentialSigning() && isSequentialSigningBlocked(signingDocument, signer)) {

			log.info("User does not have access to sign  document " + signingDocument + " yet, as sequential signing is active and not all preceeding parties have signed.");
			return showMessage("SequentialSigningBlock", req, uriParser);
		}

		log.info("User " + user + " processing signing for " + signingDocument);

		List<ValidationError> validationErrors = null;

		try {
			SimpleSigningRequest signingRequest = getSigningRequest(signingDocument, signer, uriParser);

			if (signingRequest == null) {

				throw new RuntimeException("Unable to find original PDF for document " + signingDocument);
			}

			SigningResponse signingResponse = genericSigningProvider.processSigning(req, res, user, signingRequest);

			if (res.isCommitted()) {
				return null;
			}

			if (signingResponse != null) {

				if (user == null) {

					user = signingResponse.getUser();
				}

				log.info("User " + user + " signed  document " + signingDocument);

				Timestamp now = TimeUtils.getCurrentTimestamp();

				StringBuilder signatureData = new StringBuilder();

				if (signingResponse.getSigningAttributes() != null) {

					for (Entry<String, String> entry : signingResponse.getSigningAttributes().entrySet()) {

						if (entry.getKey().equals("signingChecksum")) {

							signer.setSigningChecksum(entry.getValue());

						} else {

							signatureData.append(entry.getKey() + " = " + entry.getValue() + "\n");
						}
					}
				}

				signer.setDocument(signingDocument);
				signer.setSignedDate(now);
				signer.setSignatureData(signatureData.toString());

				if (user != null) {

					signer.setUser(user);

					if (signer.isUnsecure()) {

						String citizenID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(user);

						if (citizenID != null) {

							signer.setCitizenID(citizenID);
							signer.setUnsecureKey(null);
						}
					}
				}

				signerDAOWrapper.update(signer);
				signer.setDocument(null);

				//TODO Should this really be done here in this hardcoded fashion?
				if (user instanceof MutableUser) {

					MutableUser mutableUser = (MutableUser) user;
					boolean changed = false;

					if (!StringUtils.isEmpty(signer.getEmail()) && StringUtils.isEmpty(mutableUser.getEmail())) {

						changed = true;
						mutableUser.setEmail(signer.getEmail());
					}

					MutableAttributeHandler attributeHandler = mutableUser.getAttributeHandler();

					if (!StringUtils.isEmpty(signer.getMobilePhone()) && StringUtils.isEmpty(attributeHandler.getString("mobilePhone"))) {

						changed = true;
						attributeHandler.setAttribute("mobilePhone", signer.getMobilePhone());
						attributeHandler.setAttribute("contactBySMS", true);
					}

					if (changed) {

						try {
							systemInterface.getUserHandler().updateUser(mutableUser, false, false, true);

						} catch (UnableToUpdateUserException e) {

							log.error("Unable to update user " + user, e);
						}
					}
				}

				if (signingDocument.isFullySigned()) {

					log.info("Signing of document " + signingDocument + " complete");

					signingDocument.setStatus(SigningDocumentStatus.SIGNED);
					signingDocument.setCompletedDate(now);

					TransactionHandler transactionHandler = documentDAO.createTransaction();

					try {
						documentDAO.update(signingDocument, transactionHandler, null);

						adminModule.generateSignedPDF(signingDocument);

						if (!signingDocument.isPublishOriginal()) {

							adminModule.generateSignReceiptPDF(signingDocument);
						}

						transactionHandler.commit();

					} finally {

						TransactionHandler.autoClose(transactionHandler);
					}

					adminModule.sendSigningCompleteNotification(signingDocument, uriParser);

				} else if (signingDocument.isUseSequentialSigning()) {

					if (signingDocument.isReminderSent()) {

						// Clear reminder sent for next signer
						signingDocument.setReminderSent(false);
						documentDAO.update(signingDocument);
					}

					DocumentSigner nextSigner = signingDocument.getSigners().get(signingDocument.getSigners().indexOf(signer) + 1);
					adminModule.sendSigningNotifications(signingDocument, nextSigner, uriParser);
				}

				if (!signingDocument.isFullySigned() && signingDocument.sendEmailWhenSigningPartiallyDone()) {
					//Send message to poster
					adminModule.sendSigningPartiallyCompletedNotification(signingDocument, uriParser, user);
				}

				if (user != null && user.equals(signingDocument.getPoster())) {

					adminModule.redirectToDefaultMethod(req, res);

				} else {

					redirectToMethod(req, res, "/sign/" + signingDocument.getDocumentID());
				}

				return null;

			} else {

				log.warn("Signing of document " + signingDocument + " by user " + user + " failed, got empty signing response.");
				//TODO add validation error
			}

		} catch (ValidationException e) {

			validationErrors = new ArrayList<>();
			validationErrors.addAll(e.getErrors());
		}

		return showSignForm(req, res, user, uriParser, signingDocument, signer, validationErrors);
	}

	@WebPublic(alias = "cancel", requireLogin = true)
	public ForegroundModuleResponse cancelSigning(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		if (!HTTPUtils.isPost(req)) {

			throw new AccessDeniedException("Cancel signing requests using method " + req.getMethod() + " are not allowed.");
		}

		RequestData requestData = getRequestData(user, req, uriParser);

		if (requestData.getModuleResponse() != null) {
			return requestData.getModuleResponse();
		}

		SigningDocument signingDocument = requestData.getSigningDocument();
		DocumentSigner signer = requestData.getSigner();

		if (signingDocument.isFullySigned()) {

			if (signingDocument.isShowSignedDocumentInUserModule()) {

				res.sendRedirect(uriParser.getContextPath() + getFullAlias() + "/sign/" + signingDocument.getDocumentID());
				return null;

			} else {

				redirectToDefaultMethod(req, res);
				return null;
			}
		}

		if (signingDocument.getStatus() != SigningDocumentStatus.SIGNING) {

			//User does not have access to sign this flow instance
			log.info("User does not have access to document " + signingDocument + ", wrong status : " + signingDocument.getStatus());
			return showMessage("WrongDocumentStatus", req, uriParser);
		}

		if (signingDocument.isUseSequentialSigning() && isSequentialSigningBlocked(signingDocument, signer)) {

			log.info("User does not have access to sign document " + signingDocument + " yet, as sequential signing is active and not all preceeding parties have signed.");
			return showMessage("SequentialSigningBlock", req, uriParser);
		}

		log.info("Signer " + signer + " (user " + user + ") cancelling signing of document " + signingDocument);

		signingDocument.setStatus(SigningDocumentStatus.CANCELLED);

		documentDAOWrapper.update(signingDocument);

		if (signingDocument.isUseSequentialSigning()) {

			Iterator<DocumentSigner> iterator = signingDocument.getSigners().iterator();

			while (iterator.hasNext()) { // Remove not yet notified signers, we are the current signer in line

				DocumentSigner party = iterator.next();

				if (party.getSignedDate() == null) {

					iterator.remove();
				}
			}
		}

		adminModule.sendSigningCancelledNotifications(signingDocument, signer, uriParser);

		if (user != null) {

			redirectToDefaultMethod(req, res);
			return null;

		} else {

			res.sendRedirect(uriParser.getContextPath());
			return null;
		}
	}

	private RequestData getRequestData(User user, HttpServletRequest req, URIParser uriParser) throws URINotFoundException, AccessDeniedException, SQLException {

		Integer documentID = uriParser.getInt(2);
		String unsecureKey = uriParser.get(3);

		SigningDocument document = null;

		if (documentID != null) {
			document = documentDAOWrapper.get(documentID);
		}

		if (document == null) {

			throw new URINotFoundException(uriParser);
		}

		DocumentSigner signer = document.getSigner(user, unsecureKey);

		if (signer == null) {

			if (unsecureKey != null) {

				log.warn("User " + user + " is not a signer and unsecure key is invalid for " + document);

			} else {

				log.warn("User " + user + " is not a signer of document " + document);
			}

			return new RequestData(showMessage("SignerNotFound", req, uriParser));
		}

		return new RequestData(document, signer);
	}

	private ForegroundModuleResponse showMessage(String messageKey, HttpServletRequest req, URIParser uriParser) {

		Document doc = createDocument(req, uriParser);

		Element messageElement = doc.createElement("Message");
		messageElement.setAttribute("messageKey", messageKey);

		doc.getDocumentElement().appendChild(messageElement);

		return new SimpleForegroundModuleResponse(doc, getDefaultBreadcrumb());
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser) {

		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");
		doc.appendChild(documentElement);
		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		return doc;
	}

	public static boolean isSequentialSigningBlocked(SigningDocument document, DocumentSigner signer) {

		for (DocumentSigner otherParty : document.getSigners()) {

			if (signer == otherParty) {
				break;
			}

			if (otherParty.getSignedDate() == null) {
				return true;
			}
		}

		return false;
	}

	private String getMessage(User user, SigningDocument signingDocument, String messageTemplate, DocumentSigner signer) {

		StringBuilder userTag = new StringBuilder();

		userTag.append(signingDocument.getPoster().getFirstname() + " " + signingDocument.getPoster().getLastname());

		if (signer.getType() != SignerType.EXTERNAL_USER && signingDocument.getPoster().getUsername() != null) {

			userTag.append(" (");
			userTag.append(signingDocument.getPoster().getUsername());
			userTag.append(")");
		}

		String message = messageTemplate.replace("$poster", userTag.toString());

		message = message.replace("$document.documentID", signingDocument.getDocumentID().toString());
		message = message.replace("$document.name", signingDocument.getName());

		if (signer.getSignedDate() != null) {

			message = message.replace("$signed", DateUtils.DATE_TIME_SECONDS_FORMATTER.format(signer.getSignedDate()));
		}

		return message;
	}

	@Override
	public String getUserMenuPriority() {

		return userMenuExtensionLinkSlot;
	}
}
