package com.nordicpeak.flowengine.queries.dynamictablequery;

import java.lang.reflect.Field;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.beans.QueryDescriptor;
import com.nordicpeak.flowengine.beans.Step;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.enums.QueryState;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableQueryDescriptor;
import com.nordicpeak.flowengine.interfaces.MutableQueryInstanceDescriptor;
import com.nordicpeak.flowengine.queries.basetablequery.TableQueryProviderModule;

public class DynamicTableQueryProviderModule extends TableQueryProviderModule<DropDownAlternative, DynamicTableColumn, DynamicTableQuery, DynamicTableRow, DynamicTableCellValue, DynamicTableQueryInstance, DynamicTableQueryProviderModule> {

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, DynamicTableQueryProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		queryDAO = daoFactory.getDAO(DynamicTableQuery.class);
		queryInstanceDAO = daoFactory.getDAO(DynamicTableQueryInstance.class);
		columnDAO = daoFactory.getDAO(DynamicTableColumn.class);

		AnnotatedDAOWrapper<DynamicTableColumn, Integer> columnDAOWrapper = columnDAO.getWrapper(Integer.class);
		columnDAOWrapper.addRelations(DynamicTableColumn.ROWSUM_ADD_COLUMNS_RELATION, DynamicTableColumn.ROWSUM_SUBTRACT_COLUMNS_RELATION);
		columnDAOWrapper.setUseRelationsOnAdd(true);
		columnDAOWrapper.setUseRelationsOnUpdate(true);

		queryCRUD = new DynamicTableQueryCRUD(queryDAO.getWrapper(Integer.class), new AnnotatedRequestPopulator<>(DynamicTableQuery.class), "DynamicTableQuery", "query", null, this);
		columnCRUD = new DynamicTableColumnCRUD(columnDAOWrapper, new AnnotatedRequestPopulator<>(DynamicTableColumn.class), "DynamicTableColumn", "dynamictablecolumn", null, this);

		queryIDParamFactory = queryDAO.getParamFactory("queryID", Integer.class);
		queryInstanceIDParamFactory = queryInstanceDAO.getParamFactory("queryInstanceID", Integer.class);
		columnQueryParamFactory = columnDAO.getParamFactory("query", DynamicTableQuery.class);
		columnListNameParamFactory = columnDAO.getParamFactory("listName", String.class);

		FlowEngineDAOFactory flowDAOFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		queryDescriptorDAOWrapper = flowDAOFactory.getQueryDescriptorDAO().getWrapper(Integer.class);
		queryDescriptorDAOWrapper.addRelations(QueryDescriptor.STEP_RELATION, Step.FLOW_RELATION, Flow.FLOW_TYPE_RELATION, Flow.CATEGORY_RELATION);
		queryDescriptorDAOWrapper.setUseRelationsOnGet(true);
	}

	@Override
	public DynamicTableQueryInstance getQueryInstance(MutableQueryInstanceDescriptor descriptor, String instanceManagerID, HttpServletRequest req, User user, User poster, InstanceMetadata instanceMetadata) throws SQLException {

		DynamicTableQueryInstance queryInstance = super.getQueryInstance(descriptor, instanceManagerID, req, user, poster, instanceMetadata);

		if (queryInstance != null) {

			calculateQueryInstanceRowSums(queryInstance, queryInstance.getRows());
			calculateQueryInstanceSummary(queryInstance, queryInstance.getRows());
		}

		return queryInstance;
	}

	@Override
	public void populate(DynamicTableQueryInstance queryInstance, HttpServletRequest req, User user, User poster, boolean allowPartialPopulation, MutableAttributeHandler attributeHandler, InstanceRequestMetadata requestMetadata) throws ValidationException {

		List<DynamicTableColumn> columns = queryInstance.getQuery().getColumns();

		if (CollectionUtils.isEmpty(columns)) {

			//If the parent query doesn't have any columns then there is no population to do
			queryInstance.reset(attributeHandler);
			return;
		}

		List<ValidationError> validationErrors = new ArrayList<>(columns.size());

		List<DynamicTableRow> rows = null;
		String[] rowParams = req.getParameterValues("q" + queryInstance.getQuery().getQueryID() + "-row");

		if (rowParams == null) {

			if (allowPartialPopulation || queryInstance.getQueryInstanceDescriptor().getQueryState() != QueryState.VISIBLE_REQUIRED) {

				queryInstance.reset(attributeHandler);
				resetQueryInstanceSummary(queryInstance);
				return;
			}

			validationErrors.add(new ValidationError("NoRows"));

		} else {

			rows = new ArrayList<>();

			for (int i = 0; i < rowParams.length; i++) {

				DynamicTableRow row = new DynamicTableRow();

				row.setSortIndex(i);
				row.setRowIdentifier(rowParams[i]);
				row.setQueryInstance(queryInstance);

				rows.add(row);
			}

			if (queryInstance.getQuery().getMaxRows() != null && rows.size() > queryInstance.getQuery().getMaxRows()) {

				validationErrors.add(new ValidationError("TooManyRows"));
			}
		}

		List<DynamicTableCellValue> cellValues = null;

		if (!CollectionUtils.isEmpty(rows)) {

			cellValues = new ArrayList<>(columns.size() * rows.size());

			for (DynamicTableColumn column : columns) {

				for (DynamicTableRow row : rows) {

					String cellID = "cell_" + row.getRowIdentifier() + "_" + column.getColumnID().toString();

					DynamicTableCellValue cellValue = populateValue(req, cellValues, column, row, queryInstance, cellID, allowPartialPopulation, validationErrors);

					if (cellValue != null) {

						if (row.getValues() == null) {

							row.setValues(new ArrayList<DynamicTableCellValue>());
						}

						row.getValues().add(cellValue);
					}
				}
			}
		}

		if (!validationErrors.isEmpty()) {

			resetQueryInstanceSummary(queryInstance);

			throw new ValidationException(validationErrors);
		}

		if (CollectionUtils.isEmpty(cellValues)) {

			queryInstance.setValues(null);
			queryInstance.setRows(null);
			queryInstance.getQueryInstanceDescriptor().setPopulated(false);
			resetQueryInstanceRowSums(queryInstance);
			resetQueryInstanceSummary(queryInstance);

		} else {

			queryInstance.setValues(cellValues);
			queryInstance.setRows(rows);
			queryInstance.getQueryInstanceDescriptor().setPopulated(true);
			calculateQueryInstanceRowSums(queryInstance, queryInstance.getRows());
			calculateQueryInstanceSummary(queryInstance, rows);
		}
	}

	@Override
	public void copyQuery(MutableQueryDescriptor sourceQueryDescriptor, MutableQueryDescriptor copyQueryDescriptor, TransactionHandler transactionHandler, Map<Integer, ImmutableStatus> statusConversionMap) throws SQLException {

		DynamicTableQuery query = getQuery(sourceQueryDescriptor.getQueryID(), transactionHandler);

		super.copyQuery(query, sourceQueryDescriptor, copyQueryDescriptor, transactionHandler, statusConversionMap);
	}

	@Override
	protected Field[] getQueryInstanceRelations() {

		return new Field[] { DynamicTableQueryInstance.VALUES_RELATION, DynamicTableCellValue.COLUMN_RELATION, DynamicTableCellValue.ROW_RELATION, DynamicTableQueryInstance.QUERY_RELATION, DynamicTableQuery.COLUMNS_RELATION };
	}

	@Override
	protected RelationQuery getSaveQueryInstanceRelationQuery() {

		return new RelationQuery(DynamicTableQueryInstance.ROWS_RELATION, DynamicTableQueryInstance.VALUES_RELATION);
	}

	@Override
	protected Field[] getTableQueryRelations() {

		return new Field[] { DynamicTableQuery.COLUMNS_RELATION, DynamicTableColumn.ROWSUM_ADD_COLUMNS_RELATION, DynamicTableColumn.ROWSUM_SUBTRACT_COLUMNS_RELATION };
	}

	@Override
	protected DynamicTableQueryInstance getNewQueryInstance() {

		return new DynamicTableQueryInstance();
	}

	@Override
	protected DynamicTableQuery getNewQuery() {

		return new DynamicTableQuery();
	}

	@Override
	protected DynamicTableCellValue getNewCellValue(DynamicTableRow row, DynamicTableColumn column, String value) {

		return new DynamicTableCellValue(row, column, value);
	}

	@Override
	protected DynamicTableCellValue getNewCellValue(DynamicTableRow row, DynamicTableColumn column, String value, String displayValue, String xmlValue) {

		return new DynamicTableCellValue(row, column, value, displayValue, xmlValue);
	}

	@Override
	protected DynamicTableColumn getNewTableColumn() {

		return new DynamicTableColumn();
	}

	@Override
	protected DynamicTableQueryInstance getQueryInstance(Integer queryInstanceID) throws SQLException {

		DynamicTableQueryInstance queryInstance = super.getQueryInstance(queryInstanceID);

		if (queryInstance != null) {

			queryInstance.generateRowsFromValues();
		}

		return queryInstance;
	}

	@Override
	protected Class<DynamicTableQueryInstance> getQueryInstanceClass() {

		return DynamicTableQueryInstance.class;
	}

	@Override
	protected Field getColumnQueryRelation() {

		return DynamicTableColumn.QUERY_RELATION;
	}

}
