package com.nordicpeak.flowengine.externaladdflowinstance;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.cron4jutils.CronStringValidator;
import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.GroupMultiListSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.UserMultiListSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.AccessInterface;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.dao.querys.UpdateQuery;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.random.RandomUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.PositiveStringIntegerValidator;
import se.unlogic.standardutils.validation.TooLongContentValidationError;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.BaseFlowModule;
import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.FlowBrowserModule;
import com.nordicpeak.flowengine.addflowinstance.AddFlowInstanceModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;

import it.sauronsoftware.cron4j.Scheduler;

public class ExternalAddFlowInstanceModule extends AnnotatedRESTModule implements Runnable, AccessInterface {

	protected final String ATTRIBUTE_PREFIX = "flowAttr";
	
	private static final AnnotatedRequestPopulator<AddFlowInstanceRequest> POPULATOR = new AnnotatedRequestPopulator<AddFlowInstanceRequest>(AddFlowInstanceRequest.class);
	private static final int KEY_LENGTH = 20;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Flow family ID", description = "The ID of the flow family to use when creating flow instances", formatValidator = NonNegativeStringIntegerValidator.class)
	protected Integer flowFamilyID;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Minutes valid", description = "How many minutes a request is valid", formatValidator = PositiveStringIntegerValidator.class)
	protected Integer minutesValid = 60;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Cleanup interval", description = "How often this module should cleanup old requests (specified in crontab format)", required = true, formatValidator = CronStringValidator.class)
	private String cleanupInterval = "0 0 * * *";

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable limited create session access", description = "Controls whether limited access for creating sessions should be used or not.")
	protected boolean enableLimitedCreateSessionAccess = false;
	
	@ModuleSetting(allowsNull = true)
	@UserMultiListSettingDescriptor(name = "Create session users", description = "Users allowed to create sessions")
	protected List<Integer> createSessionUsers;

	@ModuleSetting(allowsNull = true)
	@GroupMultiListSettingDescriptor(name = "Create session groups", description = "Groups allowed to create sessions")
	protected List<Integer> createSessionGroups;
	
	@ModuleSetting(allowsNull=true)
	@TextAreaSettingDescriptor(name="Attribute whitelist", description="Attributes that will have the flowAttr prefix removed")
	protected List<String> attributePrefixWhitelist;

	@ModuleSetting
	@CheckboxSettingDescriptor(name="Force UTF-8 encoding", description="Forces the request parameters to be parsed using UTF-8 encoding")
	private boolean forceUTF8;
	
	@InstanceManagerDependency
	protected AddFlowInstanceModule addFlowInstanceModule;

	@InstanceManagerDependency
	protected FlowBrowserModule flowBrowserModule;

	@InstanceManagerDependency
	protected FlowAdminModule flowAdminModule;

	private AnnotatedDAO<AddFlowInstanceRequest> requestDAO;
	private QueryParameterFactory<AddFlowInstanceRequest, String> requestKeyParamFactory;

	private Scheduler scheduler;
	private String scheduleID;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		initScheduler();
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ExternalAddFlowInstanceModule.class.getName(), new XMLDBScriptProvider(ExternalAddFlowInstanceModule.class.getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler(), false, true, false);

		requestDAO = daoFactory.getDAO(AddFlowInstanceRequest.class);
		requestKeyParamFactory = requestDAO.getParamFactory("requestKey", String.class);
	}

	@Override
	public void update(ForegroundModuleDescriptor descriptor, DataSource dataSource) throws Exception {

		super.update(descriptor, dataSource);

		scheduler.reschedule(scheduleID, cleanupInterval);
	}

	@Override
	public void unload() throws Exception {

		stopScheduler();

		super.unload();
	}

	@RESTMethod(alias = "createsession", method = { "get", "post" })
	public void createSession(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		if (!AccessUtils.checkAccess(user, this)) {

			throw new AccessDeniedException(sectionInterface.getSectionDescriptor());
		}

		res.setContentType("application/json");
		JsonObject jsonResponse = new JsonObject();

		if (flowFamilyID != null) {

			try {
				List<ValidationError> validationErrors = new ArrayList<ValidationError>();

				AddFlowInstanceRequest addRequest = null;

				try {
					addRequest = POPULATOR.populate(req);

					if(forceUTF8) {
						
						addRequest.setFirstname(StringUtils.parseUTF8(addRequest.getFirstname()));
						addRequest.setLastname(StringUtils.parseUTF8(addRequest.getLastname()));
						addRequest.setEmail(StringUtils.parseUTF8(addRequest.getEmail()));
					}
					
					addRequest.getAttributeHandler().setAttribute("firstname", addRequest.getFirstname());
					addRequest.getAttributeHandler().setAttribute("lastname", addRequest.getLastname());
					addRequest.getAttributeHandler().setAttribute("email", addRequest.getEmail());
					
				} catch (ValidationException e) {
					validationErrors.addAll(e.getErrors());
				}
				
				Set<String> parameterKeySet = req.getParameterMap().keySet();

				for (String name : parameterKeySet) {

					if (name.startsWith(ATTRIBUTE_PREFIX)) {

						if (name.length() > 255) {
							validationErrors.add(new TooLongContentValidationError(name, name.length(), 255));
						}

						String value = ValidationUtils.validateParameter(name, req, false, 0, 65535, validationErrors);

						if (addRequest != null) {

							if(attributePrefixWhitelist != null) {
								
								String nameWithoutPrefix = StringUtils.substringAfter(name, ATTRIBUTE_PREFIX);
								
								if(attributePrefixWhitelist.contains(nameWithoutPrefix)){
								
									name = nameWithoutPrefix;
								}
							}
							
							addRequest.getAttributeHandler().setAttribute(name, value);
						}
					}
				}

				if (validationErrors.isEmpty()) {

					TransactionHandler transactionHandler = requestDAO.createTransaction();

					try {
						String key;

						do {
							key = RandomUtils.getRandomString(KEY_LENGTH, RandomUtils.MIXED_CASE_CHARACTERS_AND_DIGITS);

						} while (getAddRequest(key, transactionHandler) != null);

						addRequest.setRequestKey(key);
						addRequest.setCreated(TimeUtils.getCurrentTimestamp());

						log.info("Creating external add request " + addRequest);

						requestDAO.add(addRequest, transactionHandler, null);
						transactionHandler.commit();

					} finally {
						TransactionHandler.autoClose(transactionHandler);
					}

					jsonResponse.putField("url", uriParser.getFullContextPath() + getFullAlias() + "/add/" + addRequest.getRequestKey());

					Calendar calendar = Calendar.getInstance();
					calendar.setTime(addRequest.getCreated());
					calendar.add(Calendar.MINUTE, minutesValid);

					jsonResponse.putField("validUntil", DateUtils.DATE_TIME_FORMATTER.format(calendar.getTime()));

				} else {

					jsonResponse.putField("error", "input error");

					JsonArray jsonErrors = new JsonArray();

					for (ValidationError validationError : validationErrors) {
						jsonErrors.addNode(validationError.toString());
					}

					jsonResponse.putField("errors", jsonErrors);
				}

			} catch (SQLException e) {

				log.error("Error creating add request", e);

				jsonResponse.putField("error", "internal error");
			}

		} else {

			jsonResponse.putField("error", "internal error");
			jsonResponse.putField("description", "module not configured");
		}

		res.getWriter().write(jsonResponse.toJson());
		res.getWriter().flush();
	}

	public boolean isValid(AddFlowInstanceRequest addRequest) {

		Calendar calendar = Calendar.getInstance();
		calendar.add(Calendar.MINUTE, minutesValid);

		return addRequest.getCreated().compareTo(calendar.getTime()) < 0;
	}

	public AddFlowInstanceRequest getAddRequest(String key) throws SQLException {

		HighLevelQuery<AddFlowInstanceRequest> query = new HighLevelQuery<AddFlowInstanceRequest>();
		query.addParameter(requestKeyParamFactory.getParameter(key));

		return requestDAO.get(query);
	}

	public AddFlowInstanceRequest getAddRequest(String key, TransactionHandler transactionHandler) throws SQLException {

		HighLevelQuery<AddFlowInstanceRequest> query = new HighLevelQuery<AddFlowInstanceRequest>();
		query.addParameter(requestKeyParamFactory.getParameter(key));

		return requestDAO.get(query, transactionHandler);
	}

	public void deleteAddRequest(String key) throws SQLException {

		HighLevelQuery<AddFlowInstanceRequest> query = new HighLevelQuery<AddFlowInstanceRequest>();
		query.addParameter(requestKeyParamFactory.getParameter(key));

		requestDAO.delete(query);
	}

	@WebPublic
	public ForegroundModuleResponse add(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String key;
		AddFlowInstanceRequest addRequest;

		if (uriParser.size() == 3 && (key = uriParser.get(2)) != null && (addRequest = getAddRequest(key)) != null) {

			if (!isValid(addRequest)) {

				deleteAddRequest(addRequest.getRequestKey());

				throw new URINotFoundException(uriParser);
			}

			FlowFamily flowFamily = flowAdminModule.getFlowFamily(flowFamilyID);

			if (flowFamily == null) {

				log.error("Found no flowFamily for flowFamilyID " + flowFamilyID);
				throw new URINotFoundException(uriParser);
			}

			List<Flow> flows = flowAdminModule.getFlowVersions(flowFamily);

			Flow latestEnabledFlow = null;

			if (!CollectionUtils.isEmpty(flows)) {
				for (Flow flow : flows) {

					if (flow.isEnabled()) {

						latestEnabledFlow = flow;
					}
				}
			}

			if (latestEnabledFlow == null) {

				log.error("Found no enabled flow for flowFamilyID " + flowFamilyID);
				throw new URINotFoundException(uriParser);
			}

			Integer flowID = latestEnabledFlow.getFlowID();
			String flowBrowserURI = systemInterface.getContextPath() + flowBrowserModule.getFullAlias();

			return addFlowInstanceModule.createAddFlowInstanceRequest(req, res, flowID, user, null, false, false, null, null, BaseFlowModule.MANAGER_REQUEST_METADATA, new ExternalAddFlowInstanceCallback(addRequest, flowBrowserURI), addRequest.getAttributeHandler());
		}

		throw new URINotFoundException(uriParser);
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 50)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws SQLException {

		ImmutableFlowInstance flowInstance = event.getFlowInstanceManager().getFlowInstance();

		String requestKey = flowInstance.getAttributeHandler().getString("addFlowInstanceRequestKey");

		if (requestKey != null && getAddRequest(requestKey) != null) {

			log.info("Deleting add flowinstance request with key " + requestKey);
			
			deleteAddRequest(requestKey);
		}
	}

	protected synchronized void initScheduler() {

		if (scheduler != null) {

			log.warn("Illegal state, scheduler is not null");
			stopScheduler();
		}

		scheduler = new Scheduler(systemInterface.getApplicationName() + " - " + moduleDescriptor.toString());
		scheduler.setDaemon(true);

		scheduleID = scheduler.schedule(cleanupInterval, this);
		scheduler.start();
	}

	protected synchronized void stopScheduler() {

		try {
			if (scheduler != null) {

				scheduler.stop();
				scheduler = null;
			}

		} catch (IllegalStateException e) {
			log.error("Error stopping scheduler", e);
		}
	}

	@Override
	public void run() {

		try {
			removeOldAddRequests();

		} catch (Throwable t) {
			log.error("Error removing old add requests", t);
		}
	}

	private void removeOldAddRequests() throws SQLException {

		UpdateQuery query = new UpdateQuery(requestDAO.getDataSource(), "DELETE FROM " + requestDAO.getTableName() + " WHERE created < ?");

		Calendar calendar = Calendar.getInstance();
		calendar.add(Calendar.MINUTE, -minutesValid);
		query.setTimestamp(1, new Timestamp(calendar.getTimeInMillis()));

		query.executeUpdate();

		if (query.getAffectedRows() > 0) {

			log.info("Removed " + query.getAffectedRows() + " old add requests");
		}
	}

	@WebPublic(requireLogin = true)
	public ForegroundModuleResponse cleanRequests(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws URINotFoundException, SQLException {

		if (user.isAdmin()) {

			removeOldAddRequests();
			return null;
		}

		throw new URINotFoundException(uriParser);
	}

	@Override
	public boolean allowsAdminAccess() {

		return false;
	}

	@Override
	public boolean allowsUserAccess() {

		return enableLimitedCreateSessionAccess == false;
	}

	@Override
	public boolean allowsAnonymousAccess() {

		return enableLimitedCreateSessionAccess == false;
	}

	@Override
	public Collection<Integer> getAllowedGroupIDs() {

		return createSessionGroups;
	}

	@Override
	public Collection<Integer> getAllowedUserIDs() {

		return createSessionUsers;
	}
}
