package com.nordicpeak.flowengine.externalmessageoffer;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.cron4jutils.CronStringValidator;
import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.enums.SystemStatus;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.CRUDCallback;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.serialization.SerializationUtils;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.UserFlowInstanceModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.events.FlowVersionAdded;
import com.nordicpeak.flowengine.externalmessageoffer.beans.ExternalMessageOffer;
import com.nordicpeak.flowengine.externalmessageoffer.beans.ExternalMessageOfferSettings;
import com.nordicpeak.flowengine.externalmessageoffer.daos.ExternalMessageOfferDAO;
import com.nordicpeak.flowengine.externalmessageoffer.daos.ExternalMessageOfferSettingsDAO;
import com.nordicpeak.flowengine.externalmessageoffer.enums.ExternalMessageOfferStatus;
import com.nordicpeak.flowengine.externalmessageoffer.views.ExternalMessageOfferFormProvider;
import com.nordicpeak.flowengine.externalmessageoffer.views.ExternalMessageOfferSettingsFormProvider;
import com.nordicpeak.flowengine.externalmessageoffer.views.ExternalMessageOfferShowProvider;
import com.nordicpeak.flowengine.flowinstanceadmin2.FlowInstanceAdmin2Module;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;

import it.sauronsoftware.cron4j.Scheduler;

public class ExternalMessageOfferProviderModule extends AnnotatedForegroundModule implements CRUDCallback<User>, ViewFragmentModule<ForegroundModuleDescriptor>, Runnable {

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable fragment XML debug", description = "Enables debugging of fragment XML")
	private boolean debugFragmentXML;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable debug logging", description = "Enables extra logging for debugging")
	private boolean debugLogging;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Check for expired and withdrawn offers", description = "How often this module should check for expired and withdrawn offers (specified in crontab format)", required = true, formatValidator = CronStringValidator.class)
	private String cronExp = "0 * * * *";

	private ExternalMessageOfferSettingsDAO settingsDAO;
	private ExternalMessageOfferDAO offerDAO;

	private FlowEngineDAOFactory flowEngineDAOFactory;
	private AnnotatedDAOWrapper<FlowInstance, Integer> flowInstanceDAOWrapper;
	private AnnotatedDAOWrapper<Status, Integer> flowInstanceStatusDAOWrapper;
	private QueryParameterFactory<Status, Flow> flowInstanceStatusFlowParamFactory;

	private FlowAdminModule flowAdminModule;
	private FlowInstanceAdmin2Module flowInstanceAdminModule;
	private UserFlowInstanceModule userFlowInstanceModule;

	private ExternalMessageOfferSettingsFormProvider offerSettingsFormProvider;
	private ExternalMessageOfferShowProvider offerShowProvider;
	private ExternalMessageOfferFormProvider offerFormProvider;

	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;

	private Scheduler scheduler;
	private boolean running;

	private Set<Integer> messageIDCache;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		viewFragmentTransformer = new ModuleViewFragmentTransformer<>(sectionInterface.getForegroundModuleXSLTCache(), this, sectionInterface.getSystemInterface().getEncoding());

		super.init(moduleDescriptor, sectionInterface, dataSource);

		initScheduler();

		cacheMessageIDs();
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ExternalMessageOfferProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("daos/DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = new ExternalMessageOfferSettingsDAO(dataSource, ExternalMessageOfferSettings.class, daoFactory);
		offerDAO = new ExternalMessageOfferDAO(dataSource, ExternalMessageOffer.class, daoFactory);

		flowEngineDAOFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		flowInstanceDAOWrapper = flowEngineDAOFactory.getFlowInstanceDAO().getWrapper(Integer.class);
		flowInstanceDAOWrapper.addRelations(FlowInstance.STATUS_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);
		flowInstanceDAOWrapper.setUseRelationsOnGet(true);

		flowInstanceStatusDAOWrapper = flowEngineDAOFactory.getStatusDAO().getWrapper(Integer.class);
		flowInstanceStatusFlowParamFactory = flowEngineDAOFactory.getStatusDAO().getParamFactory("flow", Flow.class);
	}

	public List<Status> getStatuses(Flow flow) throws SQLException {

		HighLevelQuery<Status> query = new HighLevelQuery<>();

		query.addParameter(flowInstanceStatusFlowParamFactory.getParameter(flow));

		return flowEngineDAOFactory.getStatusDAO().getAll(query);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		viewFragmentTransformer.setDebugXML(debugFragmentXML);

		initScheduler();
	}

	@Override
	public void unload() throws Exception {

		stopScheduler();

		if (flowAdminModule != null) {

			setFlowAdminModule(null);
		}

		if (flowInstanceAdminModule != null) {

			setFlowInstanceAdminModule(null);
		}

		if (userFlowInstanceModule != null) {

			setUserFlowInstanceModule(null);
		}

		super.unload();
	}

	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse trigger(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (user.isAdmin()) {

			log.info("User " + user + " manually triggered");

			run();
			return new SimpleForegroundModuleResponse("See log", getDefaultBreadcrumb());
		}

		throw new URINotFoundException(uriParser);
	}

	private void cacheMessageIDs() {

		try {

			List<ExternalMessageOffer> offers = offerDAO.getAll();

			if (offers == null) {

				messageIDCache = null;

			} else {

				List<Integer> messageIDs = new ArrayList<>(offers.size());

				for (ExternalMessageOffer offer : offers) {

					messageIDs.add(offer.getMessageID());
				}

				messageIDCache = ConcurrentHashMap.newKeySet(messageIDs.size());

				messageIDCache.addAll(messageIDs);
			}

		} catch (SQLException e) {

			log.error("Error caching message ids", e);
		}
	}

	public void addToMessageIDCache(Integer messageID) {

		if (messageIDCache == null) {

			messageIDCache = ConcurrentHashMap.newKeySet();
		}

		messageIDCache.add(messageID);
	}

	@InstanceManagerDependency(required = true)
	public void setFlowAdminModule(FlowAdminModule flowAdminModule) {

		if (flowAdminModule == null && this.flowAdminModule != null) {

			this.flowAdminModule.removeStatusFormExtensionProvider(offerSettingsFormProvider);

			offerSettingsFormProvider = null;
		}

		this.flowAdminModule = flowAdminModule;

		if (this.flowAdminModule != null) {

			offerSettingsFormProvider = new ExternalMessageOfferSettingsFormProvider(this);

			this.flowAdminModule.addStatusFormExtensionProvider(offerSettingsFormProvider);
		}
	}

	@InstanceManagerDependency(required = true)
	public void setFlowInstanceAdminModule(FlowInstanceAdmin2Module flowInstanceAdminModule) {

		if (flowInstanceAdminModule == null && this.flowInstanceAdminModule != null) {

			this.flowInstanceAdminModule.removeExternalMessageShowExtensionProvider(offerShowProvider);

			offerShowProvider = null;

			this.flowInstanceAdminModule.removeExternalMessageFormExtensionProvider(offerFormProvider);

			offerFormProvider = null;
		}

		this.flowInstanceAdminModule = flowInstanceAdminModule;

		if (this.flowInstanceAdminModule != null) {

			offerShowProvider = getExternalMessageOfferShowProvider();

			this.flowInstanceAdminModule.addExternalMessageShowExtensionProvider(offerShowProvider);

			offerFormProvider = new ExternalMessageOfferFormProvider(this);

			this.flowInstanceAdminModule.addExternalMessageFormExtensionProvider(offerFormProvider);
		}
	}

	@InstanceManagerDependency(required = true)
	public void setUserFlowInstanceModule(UserFlowInstanceModule userFlowInstanceModule) {

		if (userFlowInstanceModule == null && this.userFlowInstanceModule != null) {

			this.userFlowInstanceModule.removeExternalMessageShowExtensionProvider(offerShowProvider);
			this.userFlowInstanceModule.removeFlowInstanceOverviewExtensionProvider(offerShowProvider);

			offerShowProvider = null;
		}

		this.userFlowInstanceModule = userFlowInstanceModule;

		if (this.userFlowInstanceModule != null) {

			offerShowProvider = getExternalMessageOfferShowProvider();

			this.userFlowInstanceModule.addExternalMessageShowExtensionProvider(offerShowProvider);
			this.userFlowInstanceModule.addFlowInstanceOverviewExtensionProvider(offerShowProvider);
		}
	}

	private ExternalMessageOfferShowProvider getExternalMessageOfferShowProvider() {

		if (offerShowProvider != null) {

			return offerShowProvider;
		}

		return new ExternalMessageOfferShowProvider(this);
	}

	protected synchronized void initScheduler() {

		if (scheduler != null) {

			stopScheduler();
		}

		scheduler = new Scheduler(systemInterface.getApplicationName() + " - " + moduleDescriptor.toString());
		scheduler.setDaemon(true);
		scheduler.schedule(cronExp, this);

		scheduler.start();
	}

	protected synchronized void stopScheduler() {

		try {
			if (scheduler != null) {

				scheduler.stop();
				scheduler = null;
			}

		} catch (IllegalStateException e) {
			log.error("Error stopping scheduler", e);
		}
	}

	@Override
	public void run() {

		if (this.systemInterface.getSystemStatus() == SystemStatus.STARTED && !running) {

			try {

				running = true;

				checkForExpiredOffers();

				cacheMessageIDs();

			} finally {

				running = false;
			}
		}
	}

	private void checkForExpiredOffers() {

		try {

			List<ExternalMessageOffer> activeOffers = offerDAO.getByStatus(ExternalMessageOfferStatus.ACTIVE, ExternalMessageOffer.FIELD_SETTINGS);

			if (activeOffers != null) {

				log.info("Checking for withdrawn and expired offers");

				int withdrawnOffers = 0;
				int expiredOffers = 0;

				TransactionHandler transactionHandler = null;

				try {

					transactionHandler = new TransactionHandler(dataSource);

					for (ExternalMessageOffer offer : activeOffers) {

						FlowInstance flowInstance = flowInstanceDAOWrapper.get(offer.getFlowInstanceID());

						logDebug("Checking if " + offer + " (flowInstanceLastStatusChanged = " + flowInstance.getLastStatusChange() + ") has been withdrawn or expired (Time limit = " + offer.getSettings().isSetTimeLimit() + ", Duration days = " + offer.getSettings().getDurationDays() + ")");

						ExternalMessageOfferStatus newStatus = calculateOfferStatus(offer, flowInstance);

						if (newStatus == ExternalMessageOfferStatus.WITHDRAWN) {

							withdrawnOffers++;

							logDebug("Offer " + offer + " has been withdrawn");

						} else if (newStatus == ExternalMessageOfferStatus.EXPIRED) {

							expiredOffers++;

							logDebug("Offer " + offer + " has expired");

						} else {

							logDebug("Offer " + offer + " has not been withdrawn or expired");

							continue;
						}

						offer.setStatus(newStatus);

						offerDAO.update(offer, transactionHandler, null);

						handleOfferTermination(offer, flowInstance, null);
					}

					if (withdrawnOffers > 0) {

						log.info("Found " + withdrawnOffers + " withdrawn offers.");
					}

					if (expiredOffers > 0) {

						log.info("Found " + expiredOffers + " expired offers.");
					}

					transactionHandler.commit();

				} finally {

					TransactionHandler.autoClose(transactionHandler);
				}

			}

		} catch (SQLException e) {

			log.error("Error checking for expired offers for offers", e);
		}
	}

	@Override
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();

		Element documentElement = doc.createElement("Document");

		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(documentElement);

		return doc;
	}

	public ExternalMessageOfferStatus calculateOfferStatus(ExternalMessageOffer offer, FlowInstance flowInstance) {

		LocalDateTime expirationTime = offer.calculateExpirationTime();
		boolean offerHasExpired = expirationTime != null && LocalDateTime.now().isAfter(expirationTime);

		if (offerHasBeenWithdrawn(offer, flowInstance, expirationTime, offerHasExpired)) {

			return ExternalMessageOfferStatus.WITHDRAWN;

		} else if (offerHasExpired) {

			return ExternalMessageOfferStatus.EXPIRED;

		} else {

			return ExternalMessageOfferStatus.ACTIVE;
		}
	}

	private boolean offerHasBeenWithdrawn(ExternalMessageOffer offer, FlowInstance flowInstance, LocalDateTime expirationTime, boolean offerHasExpired) {

		Timestamp lastFlowInstanceStatusChange = flowInstance.getLastStatusChange();
		boolean flowInstanceStatusHasChanged = !lastFlowInstanceStatusChange.equals(offer.getCreated());

		if (flowInstanceStatusHasChanged && offerHasExpired) {

			// Offer was withdrawn before it expired
			return lastFlowInstanceStatusChange.toLocalDateTime().isBefore(expirationTime);
		}

		return flowInstanceStatusHasChanged;
	}

	public void handleOfferTermination(ExternalMessageOffer offer, FlowInstance flowInstance, User user) throws SQLException {

		if (offer.getSettings().isSetAsAttribute()) {

			setAttribute(offer, flowInstance);
		}

		updateStatus(offer, flowInstance, user);
	}

	private void setAttribute(ExternalMessageOffer offer, FlowInstance flowInstance) throws SQLException {

		MutableAttributeHandler flowInstanceAttributeHandler = flowInstance.getAttributeHandler();

		String attributeName = offer.getSettings().getAttributeName();
		String attributeValue = offer.getStatus().toString();

		FlowInstanceAttribute attribute = new FlowInstanceAttribute(attributeName, attributeValue);
		attribute.setFlowInstance(flowInstance);

		flowInstanceAdminModule.getDAOFactory().getFlowInstanceAttributeDAO().addOrUpdate(attribute, null);

		flowInstanceAttributeHandler.setAttribute(attributeName, attributeValue);
	}

	private void updateStatus(ExternalMessageOffer offer, FlowInstance flowInstance, User user) throws SQLException {

		Integer newStatusID = getNewFlowInstanceStatusID(offer);
		Status newStatus;

		if (newStatusID == null || (newStatus = flowInstanceStatusDAOWrapper.get(newStatusID)) == null) {

			log.warn("Error handling offer termination, could not get new flow instance status for offer " + offer + " and status id " + newStatusID);
			return;
		}

		Status previousStatus = flowInstance.getStatus();

		FlowInstanceEvent updateStatusEvent = flowInstanceAdminModule.updateStatus(flowInstance, newStatus, user, null);

		flowInstanceAdminModule.sendStatusChangedEvents(flowInstance, updateStatusEvent, previousStatus, user);
	}

	private Integer getNewFlowInstanceStatusID(ExternalMessageOffer offer) {

		ExternalMessageOfferSettings settings = offer.getSettings();

		switch (offer.getStatus()) {

			case ACCEPTED:
				return settings.getAcceptedStatusID();
			case DECLINED:
				return settings.getDeclinedStatusID();
			case EXPIRED:
				return settings.getExpiredStatusID();
			default:
				return null;
		}
	}

	@EventListener(channel = FlowInstance.class)
	public void processFlowInstanceEvent(CRUDEvent<FlowInstance> event, EventSource eventSource) {

		if (event.getAction() == CRUDAction.DELETE) {

			for (FlowInstance flowInstance : event.getBeans()) {

				log.info("Deleting offers in " + flowInstance);

				try {

					Integer deletedCount = offerDAO.deleteByFlowInstanceID(flowInstance.getFlowInstanceID());

					log.info("Deleted " + deletedCount + " offers in " + flowInstance);

				} catch (SQLException e) {

					log.error("Error deleting offers for flow instance: " + flowInstance.getFlowInstanceID(), e);
				}
			}
		}
	}

	@EventListener(channel = Flow.class)
	public void processFlowEvent(CRUDEvent<Flow> event, EventSource eventSource) throws Exception {

		if (event.getAction() == CRUDAction.DELETE) {

			for (Flow flow : event.getBeans()) {

				if (flow.getStatuses() == null) {

					continue;
				}

				log.info("Deleting offer settings for " + flow);

				try (TransactionHandler transactionHandler = settingsDAO.createTransaction()) {

					Integer deletedCount = 0;

					for (Status status : flow.getStatuses()) {

						deletedCount += settingsDAO.deleteByStatusID(status.getStatusID(), transactionHandler);
					}

					log.info("Deleted " + deletedCount + " offer settings for flow " + flow);

					transactionHandler.commit();

				} catch (SQLException e) {

					log.error("Error deleting offers settings for flow: " + flow, e);
				}
			}
		}
	}

	@EventListener(channel = Flow.class)
	public void processFlowVersionAddedEvent(FlowVersionAdded event, EventSource eventSource) throws Exception {

		Map<Integer, ImmutableStatus> statusConversionMap = event.getStatusConversionMap();

		if (statusConversionMap != null) {

			try (TransactionHandler transactionHandler = new TransactionHandler(dataSource)) {

				for (Entry<Integer, ImmutableStatus> entry : statusConversionMap.entrySet()) {

					Integer statusID = entry.getKey();
					ImmutableStatus newStatus = entry.getValue();

					ExternalMessageOfferSettings settings = settingsDAO.getByStatusID(statusID);

					if (settings != null) {

						ExternalMessageOfferSettings newSettings = SerializationUtils.cloneSerializable(settings);

						newSettings.setSettingsID(null);
						newSettings.setOffers(null);
						newSettings.setStatusID(newStatus.getStatusID());

						ImmutableStatus newExpiredStatus = statusConversionMap.get(settings.getExpiredStatusID());
						ImmutableStatus newAcceptedStatus = statusConversionMap.get(settings.getAcceptedStatusID());
						ImmutableStatus newDeclinedStatus = statusConversionMap.get(settings.getDeclinedStatusID());

						if (newExpiredStatus != null) {

							newSettings.setExpiredStatusID(newExpiredStatus.getStatusID());
						}

						if (newAcceptedStatus != null) {

							newSettings.setAcceptedStatusID(newAcceptedStatus.getStatusID());
						}

						if (newDeclinedStatus != null) {

							newSettings.setDeclinedStatusID(newDeclinedStatus.getStatusID());
						}

						settingsDAO.add(newSettings, transactionHandler, null);
					}
				}

				transactionHandler.commit();

			} catch (SQLException e) {

				log.error("Error adding status settings for event: " + event, e);
			}
		}
	}

	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {

		return moduleDescriptor;
	}

	@Override
	public List<LinkTag> getLinkTags() {

		return links;
	}

	@Override
	public List<ScriptTag> getScriptTags() {

		return scripts;
	}

	@Override
	public String getTitlePrefix() {

		return moduleDescriptor.getName();
	}

	public ModuleViewFragmentTransformer<ForegroundModuleDescriptor> getViewFragmentTransformer() {

		return viewFragmentTransformer;
	}

	public ExternalMessageOfferSettingsDAO getSettingsDAO() {

		return settingsDAO;
	}

	public ExternalMessageOfferDAO getOfferDAO() {

		return offerDAO;
	}

	public AnnotatedDAOWrapper<Status, Integer> getFlowInstanceStatusDAOWrapper() {

		return flowInstanceStatusDAOWrapper;
	}

	public String getProviderID() {

		return this.moduleDescriptor.getModuleID().toString();
	}

	public Set<Integer> getMessageIDCache() {

		return messageIDCache;
	}

	public FlowInstanceAdmin2Module getFlowInstanceAdminModule() {

		return flowInstanceAdminModule;
	}

	public UserFlowInstanceModule getUserFlowInstanceModule() {

		return userFlowInstanceModule;
	}

	private void logDebug(String message) {

		if (debugLogging) {

			log.info(message);
		}
	}

	//TODO Add event?
	//	private FlowInstanceEventGenerator getFlowInstanceEventGenerator() {
	//
	//		return flowInstanceAdminModule.getFlowInstanceEventGenerator();
	//	}

}