package com.nordicpeak.flowengine.queries.treequery;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.apache.log4j.Level;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.annotations.SplitOnLineBreak;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;

import com.nordicpeak.treequerydataprovider.SimpleTreeNode;
import com.nordicpeak.treequerydataprovider.TreeDataHandler;
import com.nordicpeak.treequerydataprovider.TreeDataProvider;
import com.nordicpeak.treequerydataprovider.TreeNode;

public class FileTreeQueryDataProvider extends AnnotatedForegroundModule implements TreeDataProvider {

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Root directory", description = "", required = true)
	private String rootPath;

	@ModuleSetting(allowsNull = true)
	@SplitOnLineBreak
	@TextAreaSettingDescriptor(name = "Excluded subpaths", description = "Exclusion paths are relative from the root. One per row.")
	private List<String> excludedPaths;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Show files", description = "If files should be shown. If false only directories are shown.")
	private boolean showFiles;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Max tree depth", description = "How many directory levels to search. 0 for root only.", required = true, formatValidator = NonNegativeStringIntegerValidator.class)
	private Integer maxDepth = 1;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Provider ID", description = "The unique identifier for this provider.", required = true)
	private String providerID = getClass().getName();

	protected TreeDataHandler treeDataHandler;

	protected Path directoryPath;
	protected DirectoryStream.Filter<Path> filter;
	private boolean needsEncodingFix;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		needsEncodingFix = !"UTF-8".equals(System.getProperty("file.encoding"));
	}

	@Override
	public void update(ForegroundModuleDescriptor descriptor, DataSource dataSource) throws Exception {

		if (treeDataHandler != null) {
			treeDataHandler.removeTreeProvider(this);
		}

		super.update(descriptor, dataSource);

		if (treeDataHandler != null) {
			treeDataHandler.addTreeProvider(this);
		}
	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		if (ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.ERROR)) {

			directoryPath = Paths.get(rootPath);

			if (!Files.exists(directoryPath)) {

				log.error("Path " + directoryPath + " does not exist");

			} else if (!Files.isDirectory(directoryPath)) {

				log.error("Path " + directoryPath + " is not a directory");

			} else if (!Files.isReadable(directoryPath)) {

				log.error("Path " + directoryPath + " is not readable");
			}

			if (!CollectionUtils.isEmpty(excludedPaths)) {

				filter = new DirectoryStream.Filter<Path>() {

					@Override
					public boolean accept(Path entry) throws IOException {

						Path subPath = entry.subpath(directoryPath.getNameCount(), entry.getNameCount());

						for (String excludedPath : excludedPaths) {

							if (subPath.startsWith(excludedPath)) {
								return false;
							}
						}

						return true;
					}
				};

			} else {

				filter = new DirectoryStream.Filter<Path>() {

					@Override
					public boolean accept(Path entry) throws IOException {

						return true;
					}
				};
			}
		} else {

			directoryPath = null;
		}
	}

	@Override
	public void unload() throws Exception {

		if (treeDataHandler != null) {
			treeDataHandler.removeTreeProvider(this);
		}

		super.unload();
	}

	// Fix for path always returning UTF-8 data from filesystem but if java io encoding is something else it gets parsed incorrectly 
	private String fixEncoding(String text) {

		if (needsEncodingFix) {
			return StringUtils.parseUTF8(text);
		}

		return text;
	}

	private void getFiles(int depth, List<SimpleTreeNode> nodes, Map<String, Path> paths) throws IOException {

		if (depth < maxDepth) {

			List<SimpleTreeNode> newNodes = new ArrayList<>();
			Map<String, Path> subPaths = new HashMap<>();

			for (TreeNode node : nodes) {

				DirectoryStream<Path> directoryStream = null;

				try {
					directoryStream = Files.newDirectoryStream(paths.get(node.getKey()), filter);

					for (Path subPath : directoryStream) {

						boolean directory = Files.isDirectory(subPath);

						if (!showFiles && !directory) {
							continue;
						}

						SimpleTreeNode subNode = new SimpleTreeNode(fixEncoding(subPath.getFileName().toString()), fixEncoding(subPath.toString()));
						subNode.setFolder(directory);
						node.put(subNode);

						if (directory) {
							newNodes.add(subNode);
							subPaths.put(subNode.getKey(), subPath);
						}
					}

				} finally {

					if (directoryStream != null) {
						directoryStream.close();
					}
				}

			}

			Collections.sort(newNodes);
			getFiles(depth + 1, newNodes, subPaths);
		}
	}

	@Override
	public TreeNode getTree() {

		log.info("listing files in " + directoryPath);

		try {
			Path path = directoryPath;

			SimpleTreeNode rootNode = new SimpleTreeNode(fixEncoding(path.getFileName().toString()), fixEncoding(path.toString()));

			getFiles(0, Collections.singletonList(rootNode), Collections.singletonMap(rootNode.getKey(), path));

			return rootNode;

		} catch (IOException e) {

			log.error("Error listing files in " + directoryPath, e);
		}

		return null;
	}

	@InstanceManagerDependency
	public void setTreeDataHandler(TreeDataHandler treeDataHandler) {

		if (treeDataHandler != null) {

			treeDataHandler.addTreeProvider(this);

		} else if (this.treeDataHandler != null) {

			this.treeDataHandler.removeTreeProvider(this);
		}

		this.treeDataHandler = treeDataHandler;
	}

	@Override
	public String getTreeName() {

		return moduleDescriptor.getDescription();
	}

	@Override
	public String getTreeID() {

		return providerID;
	}

}
