package com.nordicpeak.flowengine.api;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.beans.Group;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.XMLResponseHandler;
import se.unlogic.standardutils.bool.BooleanUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.populators.BooleanPopulator;
import se.unlogic.standardutils.populators.DatePopulator;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.populators.PrimitiveBooleanPopulator;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowFamilyManager;
import com.nordicpeak.flowengine.beans.FlowFamilyManagerGroup;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.interfaces.APIAccessController;
import com.nordicpeak.flowengine.utils.APIAccessUtils;

public class FlowFamilyAPIModule extends APIModule {

	@InstanceManagerDependency(required = true)
	private FlowAdminModule flowAdminModule;
	
	@InstanceManagerDependency(required = false)
	private APIAccessController apiAccessModule;

	protected AnnotatedDAO<FlowFamily> flowFamilyDAO;

	protected QueryParameterFactory<FlowFamily, Integer> flowFamilyIDParamFactory;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		addResponseHandler(new XMLResponseHandler(sectionInterface.getSystemInterface().getEncoding()));

		super.init(moduleDescriptor, sectionInterface, dataSource);
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		FlowEngineDAOFactory daoFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		flowFamilyDAO = daoFactory.getFlowFamilyDAO();

		flowFamilyIDParamFactory = flowFamilyDAO.getParamFactory("flowFamilyID", Integer.class);
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		throw new URINotFoundException(uriParser);
	}

	protected FlowFamily getFlowFamily(Integer flowFamilyID) throws SQLException {

		HighLevelQuery<FlowFamily> query = new HighLevelQuery<FlowFamily>(FlowFamily.MANAGER_USERS_RELATION, FlowFamily.MANAGER_GROUPS_RELATION);
		query.addParameter(flowFamilyIDParamFactory.getParameter(flowFamilyID));

		return flowFamilyDAO.get(query);
	}

	@RESTMethod(alias = "getflowfamilymanagers", method = { "get" })
	public Document getFlowFamilyManagers(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		List<ValidationError> validationErrors = new ArrayList<ValidationError>();

		Integer flowFamilyID = ValidationUtils.validateParameter("flowfamilyid", req, true, PositiveStringIntegerPopulator.getPopulator(), validationErrors);

		FlowFamily flowFamily = null;

		if (flowFamilyID != null) {

			flowFamily = getFlowFamily(flowFamilyID);

			if (flowFamily == null) {
				validationErrors.add(new ValidationError("FlowFamilyNotFound"));
			}
		}
		
		if (flowFamily != null && !APIAccessUtils.hasAccess(apiAccessModule, flowFamily, user)) {
			
			validationErrors.add(new ValidationError("FlowFamilyAccessDenied"));
		}

		Document doc = XMLUtils.createDomDocument();
		Element responseElement = doc.createElement("FlowFamilyManagers");
		doc.appendChild(responseElement);

		if (validationErrors.isEmpty() && flowFamily != null) {

			log.info("User " + user + " listing managers of flow family " + flowFamily);

			flowFamily.setManagerUsersAndGroups(systemInterface.getUserHandler(), systemInterface.getGroupHandler());

			XMLUtils.append(doc, responseElement, "Users", flowFamily.getManagers());
			XMLUtils.append(doc, responseElement, "Groups", flowFamily.getManagerGroups());

		} else {

			log.warn("User " + user + " got following validation errors when trying to list managers of flow family " + flowFamilyID + ": " + validationErrors);

			XMLUtils.append(doc, responseElement, "ValidationErrors", validationErrors);
		}

		return doc;
	}

	@RESTMethod(alias = "addflowfamilymanagers", method = { "post" })
	public Document addFlowFamilyManagers(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		List<ValidationError> validationErrors = new ArrayList<ValidationError>();

		//TODO replace BooleanPopulators with new StrictBooleanPopulator that only accepts "false" or "true"
		Integer flowFamilyID = ValidationUtils.validateParameter("flowfamilyid", req, true, PositiveStringIntegerPopulator.getPopulator(), validationErrors);
		List<String> userNames = ValidationUtils.validateParameters("username", req, false, 1, 255, StringPopulator.getPopulator(), validationErrors);
		List<String> groupNames = ValidationUtils.validateParameters("groupname", req, false, 1, 255, StringPopulator.getPopulator(), validationErrors);
		boolean restricted = ValidationUtils.validateParameter("restricted", req, true, PrimitiveBooleanPopulator.getPopulator(), validationErrors);
		boolean allowUpdatingManagers = BooleanUtils.valueOf(ValidationUtils.validateParameter("allowupdatingmanagers", req, false, BooleanPopulator.getPopulator(), validationErrors));
		boolean notifyGroupMembersPersonally = BooleanUtils.valueOf(ValidationUtils.validateParameter("notifygroupmemberspersonally", req, false, BooleanPopulator.getPopulator(), validationErrors));

		Date validFromDate = ValidationUtils.validateParameter("fromdate", req, false, DatePopulator.getYearLimitedPopulator(), validationErrors);
		Date validToDate = ValidationUtils.validateParameter("todate", req, false, DatePopulator.getYearLimitedPopulator(), validationErrors);

		userNames = RequestUtils.parseParameters(req, systemInterface.getEncoding(), userNames);
		groupNames = RequestUtils.parseParameters(req, systemInterface.getEncoding(), groupNames);
		
		if (groupNames != null && (validFromDate != null || validToDate != null)) {
			validationErrors.add(new ValidationError("MayNotUseDateRestrictionsWithGroups"));
		}
		
		if (allowUpdatingManagers && !restricted) {
			validationErrors.add(new ValidationError("MayOnlyUseAllowUpdatingManagersWithRestricted"));
		}
		
		if(notifyGroupMembersPersonally && !restricted) {
			validationErrors.add(new ValidationError("MayOnlyUseNotifyGroupMembersPersonallyWithRestricted"));

		}

		if (userNames == null && groupNames == null) {
			validationErrors.add(new ValidationError("UsernameOrGroupRequired"));
		}

		FlowFamily flowFamily = null;

		if (flowFamilyID != null) {

			flowFamily = getFlowFamily(flowFamilyID);

			if (flowFamily == null) {
				validationErrors.add(new ValidationError("FlowFamilyNotFound"));
			}
		}
		
		if (flowFamily != null && !APIAccessUtils.hasAccess(apiAccessModule, flowFamily, user)) {
			
			validationErrors.add(new ValidationError("FlowFamilyAccessDenied"));
		}

		List<User> postedUsers = null;
		List<Group> postedGroups = null;

		List<FlowFamilyManager> addedManagerUsers = null;
		List<FlowFamilyManager> updatedManagerUsers = null;
		List<FlowFamilyManagerGroup> addedManagerGroups = null;
		List<FlowFamilyManagerGroup> updatedManagerGroups = null;

		if (validationErrors.isEmpty() && flowFamily != null) {

			if (userNames != null) {

				postedUsers = systemInterface.getUserHandler().getUsersByUsername(userNames, false, false);

				if (CollectionUtils.getSize(postedUsers) != userNames.size()) {

					findMissingUsers(userNames, postedUsers, validationErrors);

				} else {

					if (flowFamily.getManagers() == null) {
						flowFamily.setManagerUsers(new ArrayList<FlowFamilyManager>());
					}

					addedManagerUsers = new ArrayList<FlowFamilyManager>();
					updatedManagerUsers = new ArrayList<FlowFamilyManager>();

					for (User postedUser : postedUsers) {

						FlowFamilyManager manager = null;

						for (FlowFamilyManager familyManager : flowFamily.getManagers()) {

							if (familyManager.getUserID().equals(postedUser.getUserID())) {
								manager = familyManager;
								updatedManagerUsers.add(manager);
								break;
							}
						}

						if (manager == null) {
							manager = new FlowFamilyManager(postedUser);
							flowFamily.getManagers().add(manager);

							addedManagerUsers.add(manager);
						}

						manager.setRestricted(restricted);
						
						if (restricted) {
							
							manager.setAllowUpdatingManagers(allowUpdatingManagers);
						}

						if (validFromDate != null) {
							manager.setValidFromDate(new java.sql.Date(validFromDate.getTime()));

						} else {
							manager.setValidFromDate(null);
						}

						if (validToDate != null) {
							manager.setValidToDate(new java.sql.Date(validToDate.getTime()));

						} else {
							manager.setValidToDate(null);
						}
					}
				}
			}

			if (groupNames != null) {

				postedGroups = systemInterface.getGroupHandler().getGroupsByName(groupNames, false);

				checkMultipleGroupsWithSameName(postedGroups, validationErrors);

				if (CollectionUtils.getSize(postedGroups) != groupNames.size()) {

					findMissingGroups(groupNames, postedGroups, validationErrors);

				} else {

					if (flowFamily.getManagerGroups() == null) {
						flowFamily.setManagerGroups(new ArrayList<FlowFamilyManagerGroup>());
					}

					addedManagerGroups = new ArrayList<FlowFamilyManagerGroup>();
					updatedManagerGroups = new ArrayList<FlowFamilyManagerGroup>();

					for (Group postedGroup : postedGroups) {

						FlowFamilyManagerGroup managerGroup = null;

						for (FlowFamilyManagerGroup familyManagerGroup : flowFamily.getManagerGroups()) {

							if (familyManagerGroup.getGroupID().equals(postedGroup.getGroupID())) {
								managerGroup = familyManagerGroup;
								updatedManagerGroups.add(managerGroup);
								break;
							}
						}

						if (managerGroup == null) {
							managerGroup = new FlowFamilyManagerGroup(postedGroup);
							flowFamily.getManagerGroups().add(managerGroup);

							addedManagerGroups.add(managerGroup);
						}

						managerGroup.setRestricted(restricted);
						
						if (restricted) {
							
							managerGroup.setAllowUpdatingManagers(allowUpdatingManagers);
							managerGroup.setNotifyGroupMembersPersonally(notifyGroupMembersPersonally);
						}
					}
				}
			}
		}

		Document doc = XMLUtils.createDomDocument();
		Element responseElement = doc.createElement("AddFlowFamilyManagers");
		doc.appendChild(responseElement);

		if (validationErrors.isEmpty() && flowFamily != null && (postedUsers != null || postedGroups != null)) {

			String addedUsersString = addedManagerUsers != null ? StringUtils.toCommaSeparatedString(addedManagerUsers) : "";
			String addedGroupsString = addedManagerGroups != null ? StringUtils.toCommaSeparatedString(addedManagerGroups) : "";
			String updatedUsersString = updatedManagerUsers != null ? StringUtils.toCommaSeparatedString(updatedManagerUsers) : "";
			String updatedGroupsString = updatedManagerGroups != null ? StringUtils.toCommaSeparatedString(updatedManagerGroups) : "";

			log.info("User " + user + " adding managers to flow family " + flowFamily + ": adding " + CollectionUtils.getSize(addedManagerUsers) + " users (" + addedUsersString + ") and " + CollectionUtils.getSize(addedManagerGroups) + " groups (" + addedGroupsString + "), updating " + CollectionUtils.getSize(updatedManagerUsers) + " users (" + updatedUsersString + ") and " + CollectionUtils.getSize(updatedManagerGroups) + " groups (" + updatedGroupsString + ")");

			RelationQuery relationQuery = new RelationQuery(FlowFamily.MANAGER_USERS_RELATION, FlowFamily.MANAGER_GROUPS_RELATION);
			flowFamilyDAO.update(flowFamily, relationQuery);

			systemInterface.getEventHandler().sendEvent(FlowFamily.class, new CRUDEvent<FlowFamily>(CRUDAction.UPDATE, flowFamily), EventTarget.ALL);

			XMLUtils.appendNewElement(doc, responseElement, "AddedManagerUsers", CollectionUtils.getSize(addedManagerUsers));
			XMLUtils.appendNewElement(doc, responseElement, "AddedManagerGroups", CollectionUtils.getSize(addedManagerGroups));
			XMLUtils.appendNewElement(doc, responseElement, "UpdatedManagerUsers", CollectionUtils.getSize(updatedManagerUsers));
			XMLUtils.appendNewElement(doc, responseElement, "UpdatedManagerGroups", CollectionUtils.getSize(updatedManagerGroups));

		} else {

			log.warn("User " + user + " got following validation errors when trying to add " + CollectionUtils.getSize(userNames) + " user- and " + CollectionUtils.getSize(groupNames) + " group-managers to flow family " + flowFamily + ": " + validationErrors);

			XMLUtils.append(doc, responseElement, "ValidationErrors", validationErrors);
		}

		return doc;
	}

	@RESTMethod(alias = "removeflowfamilymanagers", method = { "post" })
	public Document removeFlowFamilyManagers(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		List<ValidationError> validationErrors = new ArrayList<ValidationError>();

		Integer flowFamilyID = ValidationUtils.validateParameter("flowfamilyid", req, true, PositiveStringIntegerPopulator.getPopulator(), validationErrors);
		List<String> userNames = ValidationUtils.validateParameters("username", req, false, 1, 255, StringPopulator.getPopulator(), validationErrors);
		List<String> groupNames = ValidationUtils.validateParameters("groupname", req, false, 1, 255, StringPopulator.getPopulator(), validationErrors);

		userNames = RequestUtils.parseParameters(req, systemInterface.getEncoding(), userNames);
		groupNames = RequestUtils.parseParameters(req, systemInterface.getEncoding(), groupNames);
		
		if (userNames == null && groupNames == null) {
			validationErrors.add(new ValidationError("UsernameOrGroupRequired"));
		}

		FlowFamily flowFamily = null;

		if (flowFamilyID != null) {

			flowFamily = getFlowFamily(flowFamilyID);

			if (flowFamily == null) {
				validationErrors.add(new ValidationError("FlowFamilyNotFound"));
			}
		}
		
		if (flowFamily != null && !APIAccessUtils.hasAccess(apiAccessModule, flowFamily, user)) {
			
			validationErrors.add(new ValidationError("FlowFamilyAccessDenied"));
		}

		List<User> postedUsers = null;
		List<Group> postedGroups = null;

		List<FlowFamilyManager> removedManagerUsers = null;
		List<FlowFamilyManagerGroup> removedManagerGroups = null;

		if (flowFamily != null) {

			if (userNames != null) {

				postedUsers = systemInterface.getUserHandler().getUsersByUsername(userNames, false, false);

				if (CollectionUtils.getSize(postedUsers) != userNames.size()) {

					findMissingUsers(userNames, postedUsers, validationErrors);

				} else if (flowFamily.getManagers() != null) {

					removedManagerUsers = new ArrayList<FlowFamilyManager>();

					for (User postedUser : postedUsers) {

						Iterator<FlowFamilyManager> it = flowFamily.getManagers().iterator();
						while (it.hasNext()) {
							FlowFamilyManager familyManager = it.next();

							if (familyManager.getUserID().equals(postedUser.getUserID())) {
								it.remove();
								removedManagerUsers.add(familyManager);
								break;
							}
						}
					}
				}
			}

			if (groupNames != null) {

				postedGroups = systemInterface.getGroupHandler().getGroupsByName(groupNames, false);

				checkMultipleGroupsWithSameName(postedGroups, validationErrors);

				if (CollectionUtils.getSize(postedGroups) != groupNames.size()) {

					findMissingGroups(groupNames, postedGroups, validationErrors);

				} else if (flowFamily.getManagerGroups() != null) {

					removedManagerGroups = new ArrayList<FlowFamilyManagerGroup>();

					for (Group postedGroup : postedGroups) {

						Iterator<FlowFamilyManagerGroup> it = flowFamily.getManagerGroups().iterator();
						while (it.hasNext()) {
							FlowFamilyManagerGroup familyManagerGroup = it.next();

							if (familyManagerGroup.getGroupID().equals(postedGroup.getGroupID())) {
								it.remove();
								removedManagerGroups.add(familyManagerGroup);
								break;
							}
						}
					}
				}
			}
		}

		if (validationErrors.isEmpty()) {

			// Invalidate userID cache
			flowFamily.setManagerUsers(flowFamily.getManagers());
			flowFamily.setManagerGroups(flowFamily.getManagerGroups());

			flowAdminModule.getFlowFamilyCRUD().checkManagersInUse(flowFamily, flowFamily.getActiveManagerUserIDs(), flowFamily.getManagerGroupIDs(), validationErrors);
		}

		Document doc = XMLUtils.createDomDocument();
		Element responseElement = doc.createElement("RemoveFlowFamilyManagers");
		doc.appendChild(responseElement);

		if (validationErrors.isEmpty() && flowFamily != null && (postedUsers != null || postedGroups != null)) {

			String removedUsersString = removedManagerUsers != null ? "(" + StringUtils.toCommaSeparatedString(removedManagerUsers) + ") " : "";
			String removedGroupsString = removedManagerGroups != null ? "(" + StringUtils.toCommaSeparatedString(removedManagerGroups) + ")" : "";

			log.info("User " + user + " removing managers from flow family " + flowFamily + ": " + CollectionUtils.getSize(removedManagerUsers) + " users " + removedUsersString + "and " + CollectionUtils.getSize(removedManagerGroups) + " groups " + removedGroupsString);

			RelationQuery relationQuery = new RelationQuery(FlowFamily.MANAGER_USERS_RELATION, FlowFamily.MANAGER_GROUPS_RELATION);
			flowFamilyDAO.update(flowFamily, relationQuery);

			systemInterface.getEventHandler().sendEvent(FlowFamily.class, new CRUDEvent<FlowFamily>(CRUDAction.UPDATE, flowFamily), EventTarget.ALL);

			XMLUtils.appendNewElement(doc, responseElement, "RemovedManagerUsers", CollectionUtils.getSize(removedManagerUsers));
			XMLUtils.appendNewElement(doc, responseElement, "RemovedManagerGroups", CollectionUtils.getSize(removedManagerGroups));

		} else {

			log.warn("User " + user + " got following validation errors when trying to remove " + CollectionUtils.getSize(userNames) + " user- and " + CollectionUtils.getSize(groupNames) + " group-managers from flow family " + flowFamily + ": " + validationErrors);

			XMLUtils.append(doc, responseElement, "ValidationErrors", validationErrors);
		}

		return doc;
	}

	private void findMissingUsers(List<String> userNames, List<User> postedUsers, List<ValidationError> validationErrors) {

		HashSet<String> foundUsernames = new HashSet<String>();

		if (postedUsers != null) {
			for (User postedUser : postedUsers) {
				foundUsernames.add(postedUser.getUsername().toLowerCase());
			}
		}

		for (String username : userNames) {
			if (!foundUsernames.contains(username.toLowerCase())) {
				validationErrors.add(new ValidationError("UserNotFound", username));
			}
		}
	}

	private void findMissingGroups(List<String> groupNames, List<Group> postedGroups, List<ValidationError> validationErrors) {

		HashSet<String> foundGroupnames = new HashSet<String>();

		if (postedGroups != null) {
			for (Group postedGroup : postedGroups) {
				foundGroupnames.add(postedGroup.getName().toLowerCase());
			}
		}

		for (String groupname : groupNames) {
			if (!foundGroupnames.contains(groupname.toLowerCase())) {
				validationErrors.add(new ValidationError("GroupNotFound", groupname));
			}
		}
	}

	private void checkMultipleGroupsWithSameName(List<Group> postedGroups, List<ValidationError> validationErrors) {

		if (postedGroups != null) {

			Set<String> foundGroupNames = new HashSet<String>();

			for (Group postedGroup : postedGroups) {
				foundGroupNames.add(postedGroup.getName().toLowerCase());
			}

			if (foundGroupNames.size() < postedGroups.size()) {
				validationErrors.add(new ValidationError("MultipleGroupsWithSameNameFound"));
			}
		}
	}

}
