package com.nordicpeak.flowengine.flowinstanceadminremoteinfo;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.VisibleModuleDescriptor;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.flowinstanceadminremoteinfo.beans.FlowInstanceAdminRemoteInfoEndpoint;
import com.nordicpeak.flowengine.flowinstanceadminremoteinfo.beans.FlowInstanceAdminRemoteInfoSetting;
import com.nordicpeak.flowengine.interfaces.FlowAdminFragmentExtensionViewProvider;

public class FlowInstanceAdminRemoteInfoSettingsFragment implements FlowAdminFragmentExtensionViewProvider {

	protected Logger log = Logger.getLogger(this.getClass());

	private final FlowInstanceAdminRemoteInfoModule callback;
	private final int priority;
	private final String adminExtensionViewTitle;

	public FlowInstanceAdminRemoteInfoSettingsFragment(FlowInstanceAdminRemoteInfoModule callback, int priority, String adminExtensionViewTitle) {

		this.callback = callback;
		this.priority = priority;
		this.adminExtensionViewTitle = adminExtensionViewTitle;
	}

	@Override
	public FlowAdminExtensionShowView getShowView(String extensionRequestURL, Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws TransformerException, SQLException {

		Document doc = callback.createDocument(req, uriParser, user);

		Element showViewElement = doc.createElement("ShowSettings");
		doc.getDocumentElement().appendChild(showViewElement);

		XMLUtils.appendNewElement(doc, showViewElement, "extensionRequestURL", extensionRequestURL);

		FlowInstanceAdminRemoteInfoSetting setting = callback.getSettingDAO().get(flow.getFlowFamily().getFlowFamilyID());

		boolean enabled = false;

		if (setting != null) {

			enabled = true;

			XMLUtils.append(doc, showViewElement, setting);
		}

		return new FlowAdminExtensionShowView(callback.getViewFragmentTransformer().createViewFragment(doc, true), enabled);
	}

	@Override
	public ViewFragment processRequest(String extensionRequestURL, Flow flow, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String method = uriParser.get(4);

		if ("updateflowsettings".equals(method)) {

			return updateFlowSettings(extensionRequestURL, flow, req, user, uriParser);

		} else if ("deleteflowsettings".equals(method)) {

			if(!HTTPUtils.isPost(req)) {
				
				throw new AccessDeniedException("Delete requests using method " + req.getMethod() + " are not allowed.");
			}
			
			return deleteFlowSettings(flow, user);
		}

		throw new URINotFoundException(uriParser);
	}

	private ViewFragment updateFlowSettings(String extensionRequestURL, Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		FlowFamily flowFamily = flow.getFlowFamily();
		FlowInstanceAdminRemoteInfoSetting setting = callback.getSettingDAO().get(flowFamily.getFlowFamilyID());

		List<ValidationError> validationErrors = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			validationErrors = new ArrayList<>();

			Integer endpointID = ValidationUtils.validateParameter("endpoint", req, false, IntegerPopulator.getPopulator(), validationErrors);

			if (endpointID == null) {

				log.info("User " + user + " deleted admin remote info settings for flow " + flow);

				callback.getSettingDAO().deleteByID(flowFamily.getFlowFamilyID());
				return null;
			}

			FlowInstanceAdminRemoteInfoEndpoint endpoint = callback.getEndpointDAO().get(endpointID);
			boolean newSetting = false;

			if (endpoint != null) {

				if (setting == null) {

					setting = new FlowInstanceAdminRemoteInfoSetting();
					setting.setFlowFamilyID(flowFamily.getFlowFamilyID());
					
					newSetting = true;
				}

				setting.setEndpoint(endpoint);

			} else {

				validationErrors.add(new ValidationError("endpoint", ValidationErrorType.InvalidFormat));
			}

			if (CollectionUtils.isEmpty(validationErrors)) {
				
				if (newSetting) {
					
					log.info("User " + user + " added admin remote info settings for flow " + flow);
					
					callback.getSettingDAO().add(setting);
					
				} else {

					log.info("User " + user + " updated admin remote info settings for flow " + flow);
					
					callback.getSettingDAO().update(setting);
				}
				
				return null;
			}
		}

		log.info("User " + user + " requesting admin remote info settings form for flow " + flow);

		Document doc = callback.createDocument(req, uriParser, user);

		Element settingsElement = doc.createElement("UpdateSettings");
		doc.getDocumentElement().appendChild(settingsElement);

		XMLUtils.appendNewElement(doc, settingsElement, "extensionRequestURL", extensionRequestURL);

		XMLUtils.append(doc, settingsElement, flow);
		XMLUtils.append(doc, settingsElement, setting);
		XMLUtils.append(doc, settingsElement, "AvailableEndpoints", callback.getEndpointDAO().getAll());

		if (validationErrors != null) {

			XMLUtils.append(doc, settingsElement, "ValidationErrors", validationErrors);
			settingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return callback.getViewFragmentTransformer().createViewFragment(doc, true);
	}

	private ViewFragment deleteFlowSettings(Flow flow, User user) throws Exception {

		FlowInstanceAdminRemoteInfoSetting setting = callback.getSettingDAO().get(flow.getFlowFamily().getFlowFamilyID());

		if (setting != null) {

			log.info("User " + user + " deleting admin remote info settings for flow " + flow);

			callback.getSettingDAO().delete(setting);

		} else {

			log.warn("User " + user + " trying to delete admin remote info settings for flow " + flow + " which has no settings");
		}

		return null;
	}

	@Override
	public int getPriority() {

		return priority;
	}

	@Override
	public String getExtensionViewTitle() {

		return adminExtensionViewTitle;
	}

	@Override
	public String getExtensionViewLinkName() {

		return "adminremoteinfo";
	}

	@Override
	public int getModuleID() {

		return callback.getModuleDescriptor().getModuleID();
	}

	@Override
	public VisibleModuleDescriptor getModuleDescriptor() {

		return callback.getModuleDescriptor();
	}

}
