package com.nordicpeak.flowengine.flowlistextension;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import javax.naming.OperationNotSupportedException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Level;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.SimpleViewFragment;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.ResponseType;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionDescriptor;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.core.utils.extensionlinks.ExtensionLink;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.FlowBrowserModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowType;
import com.nordicpeak.flowengine.flowlistextension.beans.FlowListExtensionSettings;
import com.nordicpeak.flowengine.interfaces.FlowAdminFragmentExtensionViewProvider;

public class FlowListExtensionProviderModule extends AnnotatedForegroundModule implements ViewFragmentModule<ForegroundModuleDescriptor>, FlowAdminFragmentExtensionViewProvider {

	private static final AnnotatedRequestPopulator<FlowListExtensionSettings> SETTINGS_POPULATOR = new AnnotatedRequestPopulator<FlowListExtensionSettings>(FlowListExtensionSettings.class);

	@XSLVariable(prefix = "java.")
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "TabTitle", description = "The title for settingstab", required = true)
	protected String tabTitle;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable fragment XML debug", description = "Enables debugging of fragment XML")
	private boolean debugFragmentXML;

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "List of your current flows", description = "Below is a list of your current flows, click on the links to be redirected to my pages", required = true)
	@XSLVariable(prefix = "java.")
	private String listCurrentFlows;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "My pages", description = "Sets default value for my pages")
	@XSLVariable(prefix = "java.")
	private String myPages = "minasidor";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Priority", description = "The priority of this extension provider compared to other providers. A low value means a higher priority. Valid values are 0 - " + Integer.MAX_VALUE + ".", required = true, formatValidator = NonNegativeStringIntegerValidator.class)
	protected int priority = 100;

	@InstanceManagerDependency(required = true)
	protected FlowBrowserModule flowBrowserModule;

	@InstanceManagerDependency(required = true)
	private StaticContentModule staticContentModule;

	private AnnotatedDAO<FlowListExtensionSettings> flowListExtensionSettingsDAO;

	private AnnotatedDAOWrapper<FlowListExtensionSettings, Integer> flowListExtensionSettingsDAOWrapper;

	private QueryParameterFactory<FlowListExtensionSettings, Integer> flowListExtensionSettingsIDParameterFactory;

	protected FlowAdminModule flowAdminModule;

	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;
	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentSettingsTransformer;

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.WARN);

		viewFragmentTransformer = new ModuleViewFragmentTransformer<>(sectionInterface.getForegroundModuleXSLTCache(), this, systemInterface.getEncoding());
		viewFragmentTransformer.setDebugXML(debugFragmentXML);

		viewFragmentSettingsTransformer = new ModuleViewFragmentTransformer<>(sectionInterface.getForegroundModuleXSLTCache(), this, systemInterface.getEncoding());
		viewFragmentSettingsTransformer.setDebugXML(debugFragmentXML);

		viewFragmentSettingsTransformer.modifyScriptsAndLinks(true, null);
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, FlowListExtensionProviderModule.class.getName(), new XMLDBScriptProvider(FlowListExtensionProviderModule.class.getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);

		flowListExtensionSettingsDAO = daoFactory.getDAO(FlowListExtensionSettings.class);

		flowListExtensionSettingsDAOWrapper = flowListExtensionSettingsDAO.getWrapper(Integer.class);

		flowListExtensionSettingsIDParameterFactory = flowListExtensionSettingsDAO.getParamFactory("flowFamilyID", Integer.class);
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return list(req, user, uriParser);
	}

	public ForegroundModuleResponse list(HttpServletRequest req, User user, URIParser uriParser) throws SQLException {

		log.info("User " + user + " requested current flow listings");

		Document doc = createDocument(req, uriParser);
		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "listCurrentFlows", listCurrentFlows);

		populateDocumentWithFlowTypeAndFlow(user, doc);

		return new SimpleForegroundModuleResponse(doc, this.moduleDescriptor.getName(), this.getDefaultBreadcrumb());
	}

	private void populateDocumentWithFlowTypeAndFlow(User user, Document doc) throws SQLException {

		//get all current published flow versions
		Collection<Flow> flowList = flowBrowserModule.getLatestPublishedFlowVersions();

		//get all flowsettings that is enabled
		List<FlowListExtensionSettings> settings = getAllEnabledSettings();

		Element myListedFlowsElement = doc.createElement("MyListedFlows");

		if (flowList != null && !settings.isEmpty()) {

			//trim all non appliable flows based on settings
			flowList = flowList.stream().filter(e -> settings.stream().map(FlowListExtensionSettings::getFlowFamilyID).collect(Collectors.toList()).contains(e.getFlowFamily().getFlowFamilyID())).collect(Collectors.toList());

			//group by flowtype
			Map<FlowType, List<Flow>> flowTypeMap = flowList.stream().collect(Collectors.groupingBy(Flow::getFlowType));

			java.util.Iterator<FlowType> flowTypeMapIterator = flowTypeMap.keySet().iterator();
			while (flowTypeMapIterator.hasNext()) {
				addFlowType(doc, user, settings, flowTypeMap, myListedFlowsElement, flowTypeMapIterator);
			}

		}
		doc.getDocumentElement().appendChild(myListedFlowsElement);
	}

	protected void addFlowType(Document doc, User user, List<FlowListExtensionSettings> settings, Map<FlowType, List<Flow>> flowTypeMap, Element myListedFlowsElement, java.util.Iterator<FlowType> flowTypeMapIterator) {

		FlowType flowType = flowTypeMapIterator.next();

		if (AccessUtils.checkAccess(user, flowType.getUserAccessInterface())) {

			Element flowTypeElement = doc.createElement("FlowType");
			flowTypeElement.appendChild(XMLUtils.createElement("FlowTypeID", flowType.getFlowTypeID(), doc));
			flowTypeElement.appendChild(XMLUtils.createElement("FlowTypeName", flowType.getName(), doc));

			Element flowElements = doc.createElement("Flows");
			List<Flow> flowWithType = flowTypeMap.get(flowType);
			for (Flow flow : flowWithType) {

				Element flowElement = doc.createElement("Flow");

				Optional<FlowListExtensionSettings> setting = settings.stream().filter(e -> e.getFlowFamilyID().equals(flow.getFlowFamily().getFlowFamilyID())).findFirst();
				
				String flowName;
				
				if (setting.isPresent() && setting.get().getCustomName() != null) {
					flowName = setting.get().getCustomName();
				} else {
					flowName = flow.getName();
				}
				
				Element flowNameElement = XMLUtils.createElement("FlowName", flowName, doc);
				Element flowIDElement = XMLUtils.createElement("FlowID", flow.getFlowID(), doc);
				Element flowEnabled = XMLUtils.createElement("FlowEnabled", flow.isEnabled(), doc);
				if (flow.getIconFileName() != null) {
					flowElement.appendChild(XMLUtils.createElement("FlowIconFileName", flow.getIconFileName(), doc));
				}
				Element flowHrefElement;
				
				if (setting.isPresent() && setting.get().getExternalLink() != null) {
					flowHrefElement = XMLUtils.createCDATAElement("FlowHref", setting.get().getExternalLink(), doc);
				} else {
					flowHrefElement = XMLUtils.createElement("FlowHref", myPages, doc);
				}

				flowElement.appendChild(flowIDElement);
				flowElement.appendChild(flowNameElement);
				flowElement.appendChild(flowHrefElement);
				flowElement.appendChild(flowEnabled);
				flowElements.appendChild(flowElement);

			}
			flowTypeElement.appendChild(flowElements);
			myListedFlowsElement.appendChild(flowTypeElement);
		}
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.sectionInterface.getSectionDescriptor().toXML(doc));
		document.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(document);

		return doc;
	}

	public String getTitlePrefix() {

		return moduleDescriptor.getName();
	}

	public List<FlowListExtensionSettings> getSettings(Integer flowFamilyID) throws SQLException {

		HighLevelQuery<FlowListExtensionSettings> query = new HighLevelQuery<>();
		query.addParameter(flowListExtensionSettingsIDParameterFactory.getParameter(flowFamilyID));

		return flowListExtensionSettingsDAO.getAll(query);
	}

	public List<FlowListExtensionSettings> getAllEnabledSettings() throws SQLException {

		List<FlowListExtensionSettings> list = flowListExtensionSettingsDAO.getAll();
		if (list == null) {
			return new ArrayList<>();
		}

		return list.stream().filter(Objects::nonNull).filter(FlowListExtensionSettings::isModuleEnabled).collect(Collectors.toList());
	}

	public SectionDescriptor getSectionDescriptor() {

		return sectionInterface.getSectionDescriptor();
	}

	@WebPublic(alias = "flowicon")
	public ForegroundModuleResponse getFlowIcon(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws URINotFoundException, SQLException, IOException {

		return flowAdminModule.getFlowIcon(req, res, user, uriParser);
	}

	public ExtensionLink getOverviewTabHeaderExtensionLink(HttpServletRequest req, URIParser uriParser, User user) throws NullPointerException {

		return new ExtensionLink(tabTitle, "#flowlist", null, null);
	}

	public ViewFragment getOverviewTabContentsViewFragment(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user, boolean manager) throws SQLException, TransformerException {

		return getTabContentsViewFragment(flowInstance, req, uriParser, null, manager);
	}

	public ForegroundModuleResponse processOverviewExtensionRequest(FlowInstance flowInstance, HttpServletRequest req, HttpServletResponse res, URIParser uriParser, User user) throws OperationNotSupportedException {

		throw new OperationNotSupportedException();
	}

	public ViewFragment getTabContentsViewFragment(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, List<ValidationError> validationErrors, boolean manager) throws SQLException, TransformerException {

		FlowListExtensionSettings settings = flowListExtensionSettingsDAOWrapper.get(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		//hide tab if inactivated
		if (settings == null || !settings.isModuleEnabled()) {
			return null;
		}

		Document doc = createDocument(req, uriParser);
		Element documentElement = doc.getDocumentElement();

		Element tabElement = XMLUtils.appendNewElement(doc, documentElement, "TabContents");

		XMLUtils.append(doc, tabElement, flowInstance);

		XMLUtils.appendNewElement(doc, tabElement, "TabTitle", tabTitle);

		XMLUtils.append(doc, tabElement, validationErrors);

		return viewFragmentTransformer.createViewFragment(doc);
	}

	public ViewFragment getValidationErrorsViewFragment(HttpServletRequest req, URIParser uriParser, List<ValidationError> validationErrors) throws TransformerException {

		Document doc = createDocument(req, uriParser);
		Element documentElement = doc.getDocumentElement();

		XMLUtils.append(doc, documentElement, "ValidationErrors", validationErrors);

		return viewFragmentTransformer.createViewFragment(doc);
	}

	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {

		return moduleDescriptor;
	}

	@Override
	public List<LinkTag> getLinkTags() {

		return links;
	}

	@Override
	public List<ScriptTag> getScriptTags() {

		return scripts;
	}

	public String getOverviewExtensionProviderID() {

		return moduleDescriptor.getModuleID().toString();
	}

	@Override
	public int getPriority() {

		return priority;
	}

	@Override
	public FlowAdminExtensionShowView getShowView(String extensionRequestURL, Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws TransformerException, SQLException {

		Document doc = createDocument(req, uriParser);

		Element showViewElement = doc.createElement("ShowSettings");
		doc.getDocumentElement().appendChild(showViewElement);

		XMLUtils.appendNewElement(doc, showViewElement, "FullAlias", getFullAlias());
		showViewElement.appendChild(flow.toXML(doc));

		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "ModuleURI", req.getContextPath() + getFullAlias());
		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "StaticContentURL", getStaticContentURL());

		XMLUtils.appendNewElement(doc, showViewElement, "extensionRequestURL", extensionRequestURL);

		XMLUtils.appendNewElement(doc, showViewElement, "title", tabTitle);

		FlowListExtensionSettings settings = flowListExtensionSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());

		boolean enabled = false;

		if (settings != null) {

			XMLUtils.append(doc, showViewElement, settings);
			enabled = true;
		}

		return new FlowAdminExtensionShowView(viewFragmentSettingsTransformer.createViewFragment(doc, true), enabled);
	}

	public String getStaticContentURL() {

		return staticContentModule.getModuleContentURL(moduleDescriptor);
	}

	@Override
	public String getExtensionViewTitle() {

		//special handling in this case, return module-name
		return moduleDescriptor.getName();
	}

	@Override
	public String getExtensionViewLinkName() {

		return "addflowinstanceasmanagersettings";
	}

	@Override
	public ViewFragment processRequest(String extensionRequestURL, Flow flow, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String method = uriParser.get(4);

		if ("showupdatesettings".equals(method)) {

			return getViewFragmentResponse(showUpdateSettings(extensionRequestURL, req, user, uriParser, null));

		} else if ("updatesettings".equals(method)) {

			return getViewFragmentResponse(updateSettings(extensionRequestURL, req, res, user, uriParser));

		} else if ("deletesettings".equals(method)) {

			return getViewFragmentResponse(deleteSettings(req, res, user, uriParser));
		}

		return null;
	}

	@Override
	public int getModuleID() {

		return moduleDescriptor.getModuleID();
	}

	@InstanceManagerDependency(required = true)
	public void setFlowAdminModule(FlowAdminModule flowAdminModule) {

		if (flowAdminModule == null && this.flowAdminModule != null) {

			this.flowAdminModule.removeFragmentExtensionViewProvider(this);
		}

		this.flowAdminModule = flowAdminModule;

		if (this.flowAdminModule != null) {

			this.flowAdminModule.addFragmentExtensionViewProvider(this);
		}
	}

	private ForegroundModuleResponse showUpdateSettings(String extensionRequestURL, HttpServletRequest req, User user, URIParser uriParser, ValidationException validationException) throws AccessDeniedException, SQLException {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		Document doc = createDocument(req, uriParser);

		Element settingsElement = doc.createElement("ShowUpdateSettings");
		doc.getDocumentElement().appendChild(settingsElement);

		settingsElement.appendChild(flow.toXML(doc));

		XMLUtils.appendNewElement(doc, settingsElement, "extensionRequestURL", extensionRequestURL);

		FlowListExtensionSettings settings = flowListExtensionSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());

		if (settings != null) {

			XMLUtils.append(doc, settingsElement, settings);
		}

		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "ModuleURI", req.getContextPath() + getFullAlias());
		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "StaticContentURL", getStaticContentURL());

		if (validationException != null) {

			settingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
			XMLUtils.append(doc, settingsElement, "ValidationErrors", validationException.getErrors());
		}

		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}

	private ForegroundModuleResponse updateSettings(String extensionRequestURL, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws AccessDeniedException, SQLException, IOException {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		try {

			updateSettings(req, flow);

		} catch (ValidationException e) {

			return showUpdateSettings(extensionRequestURL, req, user, uriParser, e);
			
		}

		flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#" + getExtensionViewLinkName());

		return null;
	}

	private ForegroundModuleResponse deleteSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws AccessDeniedException, SQLException, IOException {

		if (!HTTPUtils.isPost(req)) {

			throw new AccessDeniedException("Delete requests using method " + req.getMethod() + " are not allowed.");
		}

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		FlowListExtensionSettings settings = flowListExtensionSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());

		if (settings != null) {

			log.info("User " + user + " deleting attachmentsettings for flow " + flow);
			flowListExtensionSettingsDAOWrapper.delete(settings);

		} else {

			log.warn("User " + user + " trying to delete attachmentsettings for flow " + flow + " which has no settings");
		}

		flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#" + getExtensionViewLinkName());

		return null;
	}

	private void updateSettings(HttpServletRequest req, Flow flow) throws SQLException, ValidationException {

		FlowListExtensionSettings settings = flowListExtensionSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());

		if (req.getParameter("moduleEnabled") == null && settings != null) {

			flowListExtensionSettingsDAOWrapper.delete(settings);

		} else if (settings == null) {

			settings = SETTINGS_POPULATOR.populate(req);
			settings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
			flowListExtensionSettingsDAOWrapper.add(settings);

		} else {

			settings = SETTINGS_POPULATOR.populate(settings, req);
			flowListExtensionSettingsDAOWrapper.update(settings);

		}
	}

	private ViewFragment getViewFragmentResponse(ForegroundModuleResponse foregroundModuleResponse) throws TransformerException {

		if (foregroundModuleResponse != null) {

			if (foregroundModuleResponse.getResponseType() == ResponseType.XML_FOR_SEPARATE_TRANSFORMATION) {

				return viewFragmentSettingsTransformer.createViewFragment(foregroundModuleResponse.getDocument());

			} else {

				log.warn("Scripts and links have not been modified for FlowAdminFragmentExtensionViewProviderProcessRequest");
				return new SimpleViewFragment(foregroundModuleResponse.getHtml(), debugFragmentXML ? foregroundModuleResponse.getDocument() : null, foregroundModuleResponse.getScripts(), foregroundModuleResponse.getLinks());
			}
		}

		return null;
	}
}
