package com.nordicpeak.flowengine.evaluators.geometrycontainsaddressquerystateevaluator;

import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.lantmateriet.namespace.distribution.produkter.belagenhetsadress.v4.BelagenhetsadressResponseType;
import se.lantmateriet.namespace.distribution.produkter.belagenhetsadress.v4.BelagenhetsadressType;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.object.ObjectUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.evaluators.basequerystateevaluator.BaseQueryStateEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.queries.textareaquery.TextAreaQueryInstance;
import com.nordicpeak.lmsearch.LMSearchProvider;
import com.nordicpeak.lmsearch.beans.ParsedAddress;
import com.nordicpeak.lmsearch.utils.AddressUtils;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.io.WKTReader;

public class GeometryContainsAddressQueryStateEvaluationProviderModule extends BaseQueryStateEvaluationProviderModule<GeometryContainsAddressQueryStateEvaluator> {

	protected AnnotatedDAO<GeometryContainsAddressQueryStateEvaluator> evaluatorDAO;

	protected QueryParameterFactory<GeometryContainsAddressQueryStateEvaluator, Integer> evaluatorIDParamFactory;

	protected GeometryContainsAddressQueryStateEvaluatorCRUD evaluatorCRUD;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "Connection timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer connectionTimeout = 5;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Request timeout", description = "Request timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer requestTimeout = 10;

	@InstanceManagerDependency(required = true)
	private LMSearchProvider lmSearchProvider;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, GeometryContainsAddressQueryStateEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(GeometryContainsAddressQueryStateEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new GeometryContainsAddressQueryStateEvaluatorCRUD(GeometryContainsAddressQueryStateEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, GeometryContainsAddressQueryStateEvaluator evaluator, EvaluationCallback callback, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {

		if (evaluator.getInitialisedState() != null) {

			if (evaluator.getInitialisedState()) {

				return applyEvaluatorQueryStates(queryInstance, evaluator, callback, attributeHandler);

			} else {

				return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
			}
		}

		if (evaluator.getParsedGeometry() == null) {

			log.info("No geometry set for evaluator " + evaluator);
			return null;
		}
		
		TextAreaQueryInstance debugTextAreaQueryInstance = null;
		
		if (queryInstance instanceof TextAreaQueryInstance && "debug".equalsIgnoreCase(queryInstance.getQueryInstanceDescriptor().getQueryDescriptor().getName())) {
			
			debugTextAreaQueryInstance = (TextAreaQueryInstance) queryInstance;
		}

		Point posterAddressPosition = null;

		if (poster != null) {

			AttributeHandler posterAttributeHandler = poster.getAttributeHandler();

			String careOf = posterAttributeHandler.getString("official.careOf");
			String posterAddress = posterAttributeHandler.getString("official.address");
			String posterPostalCode = posterAttributeHandler.getString("official.zipCode");
			String posterPostalAddress = posterAttributeHandler.getString("official.postalAddress");

			if (ObjectUtils.containsNull(posterAddress, posterPostalCode, posterPostalAddress)) {

				log.info("User " + poster + " has no offical address");
				
				if (debugTextAreaQueryInstance != null) {
					debugTextAreaQueryInstance.setValue("Ingen folkbokfringsadress p anvndaren");
				}
				
			} else {
				
				if (careOf != null && posterAddress.startsWith("CO " + careOf)) {
					
					posterAddress = posterAddress.substring(3 + careOf.length());
				}

				posterPostalCode = posterPostalCode.replace(" ", "");
				
				ParsedAddress parsedAddress = AddressUtils.getAddressNumber(posterAddress);
				String address = parsedAddress.getStreet() + " " + parsedAddress.getNumber() + " " + posterPostalAddress + " " + posterPostalCode;

				BelagenhetsadressResponseType addresses = lmSearchProvider.getLmConnection().searchAddressInMunicipality(lmSearchProvider.getMunicipality(), address);

				if (addresses == null || CollectionUtils.isEmpty(addresses.getBelagenhetsadressMember())) {

					log.info("Found no buildings on posters offical address: " + address);
					
					if (debugTextAreaQueryInstance != null) {
						debugTextAreaQueryInstance.setValue("Hittade inga byggnader p folkbokfringsadressen: " + address);
					}
					
				} else {

					BelagenhetsadressType firstBuilding = addresses.getBelagenhetsadressMember().get(0).getBelagenhetsadress();

					List<Double> coordinates = firstBuilding.getAdressplatsattribut().getAdressplatspunkt().getPoint().getPos().getValue();
					double x = coordinates.get(0);
					double y = coordinates.get(1);
					
					log.info("Found " + addresses.getBelagenhetsadressMember().size() + " buildings on posters offical address " + address + ", first building at X " + x + ", Y " + y);
					
					if (debugTextAreaQueryInstance != null) {
						debugTextAreaQueryInstance.setValue("Hittade " + addresses.getBelagenhetsadressMember().size() + " byggnader p folkbokfringsadressen " + address + ", frsta vid X " + x + ", Y " + y);
					}

					posterAddressPosition = lmSearchProvider.getLmConnection().getGeometryFactory().createPoint(new Coordinate(y, x));
				}
			}
		}

		boolean livesInsideGeometry = posterAddressPosition != null && evaluator.getParsedGeometry().covers(posterAddressPosition);
		
		if (debugTextAreaQueryInstance != null) {
			debugTextAreaQueryInstance.setValue(debugTextAreaQueryInstance.getValue() + (livesInsideGeometry ? "\r\nBor innanfr omrdet" : "\r\nBor utanfr omrdet"));
		}
		
		if ((livesInsideGeometry || evaluator.isInverted()) && !(livesInsideGeometry && evaluator.isInverted())) {

			evaluator.setInitialisedState(true);
			return applyEvaluatorQueryStates(queryInstance, evaluator, callback, attributeHandler);

		} else {

			evaluator.setInitialisedState(false);
			return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
		}
	}

	@Override
	public Evaluator createEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		GeometryContainsAddressQueryStateEvaluator evaluator = new GeometryContainsAddressQueryStateEvaluator();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());

		return evaluator;
	}

	@Override
	public Evaluator getEvaluator(MutableEvaluatorDescriptor descriptor, InstanceMetadata instanceMetadata) throws Throwable {

		GeometryContainsAddressQueryStateEvaluator evaluator = (GeometryContainsAddressQueryStateEvaluator) super.getEvaluator(descriptor, instanceMetadata);

		if (evaluator != null && evaluator.getGeometry() != null) {

			try {
				evaluator.setParsedGeometry(getGeometry(evaluator));

				if (evaluator.getParsedGeometry() == null) {
					log.warn("Geometry string set but got no geometry after parsing in evaluator " + evaluator);
				}

			} catch (Exception e) {

				log.warn("Invalid geometry in evaluator " + evaluator, e);
			}
		}

		return evaluator;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return true;
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		GeometryContainsAddressQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<GeometryContainsAddressQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	@Override
	protected Class<GeometryContainsAddressQueryStateEvaluator> getEvaluatorClass() {

		return GeometryContainsAddressQueryStateEvaluator.class;
	}

	@Override
	protected AnnotatedDAO<GeometryContainsAddressQueryStateEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	@Override
	protected QueryParameterFactory<GeometryContainsAddressQueryStateEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return this.evaluatorIDParamFactory;
	}

	@Override
	protected List<Integer> getTargetQueryIDs(GeometryContainsAddressQueryStateEvaluator evaluator, EvaluationCallback callback) {

		return Collections.singletonList(callback.getQueryInstance().getQueryInstanceDescriptor().getQueryDescriptor().getQueryID());
	}

	@Override
	public GeometryContainsAddressQueryStateEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		GeometryContainsAddressQueryStateEvaluator evaluator = importEvaluatorDescriptor(descriptor, transactionHandler, query, statusConversionMap);

		evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}

	public Geometry getGeometry(GeometryContainsAddressQueryStateEvaluator evaluator) throws ValidationException {

		if (evaluator.getGeometry() != null) {

			WKTReader reader = new WKTReader(lmSearchProvider.getLmConnection().getGeometryFactory());

			try {
				return reader.read(evaluator.getGeometry());

			} catch (Exception e) {

				throw new ValidationException(new ValidationError("geometry", ValidationErrorType.InvalidFormat));
			}
		}

		return null;
	}
}
