package com.nordicpeak.flowengine.queries.treequery;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import javax.sql.DataSource;

import org.apache.log4j.Level;

import se.unlogic.cron4jutils.CronStringValidator;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.Group;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.serialization.SerializationUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;

import com.nordicpeak.flowengine.queries.usertreequery.UserTreeQueryProviderModule;
import com.nordicpeak.flowengine.queries.usertreequery.beans.TreeNode;
import com.nordicpeak.flowengine.queries.usertreequery.beans.TreeUser;
import com.nordicpeak.flowengine.queries.usertreequery.interfaces.UserTreeQueryDataProvider;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

import it.sauronsoftware.cron4j.Scheduler;
import it.sauronsoftware.cron4j.Task;
import it.sauronsoftware.cron4j.TaskExecutionContext;

public class GroupUserTreeQueryDataProvider extends AnnotatedForegroundModule implements UserTreeQueryDataProvider {

	private static final String GROUP_LDAP_UID_ATTRIBUTE = "ldap_group_uid";
	private static final String GROUP_PARENT_UID_ATTRIBUTE = "parentGroupUID";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Provider ID", description = "The unique identifier for this provider.", required = true)
	private String providerID = getClass().getName();

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Cache interval", description = "How often organizations should be cached (specified in crontab format)", required = true, formatValidator = CronStringValidator.class)
	private String cacheInterval = "0 0 * * *";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Cache filestore", description = "Optional filestore to improve restart times")
	private String cacheFilestore;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Group id attribute name", description = "The name of the group id attribute", required = true)
	private String groupIDAttribute = GROUP_LDAP_UID_ATTRIBUTE;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Parent group id attribute name", description = "The name of the parent group id attribute", required = true)
	private String parentGroupIDAttribute = GROUP_PARENT_UID_ATTRIBUTE;

	private File treeCacheFile;
	private File usersCacheFile;

	private Scheduler scheduler;

	protected UserTreeQueryProviderModule treeQueryProviderModule;

	protected TreeNode organizationTree;
	protected HashMap<String, List<TreeUser>> usersMap;

	boolean firstStart = true;

	@Override
	public void update(ForegroundModuleDescriptor descriptor, DataSource dataSource) throws Exception {

		if (treeQueryProviderModule != null) {
			treeQueryProviderModule.removeTreeProvider(this);
		}

		super.update(descriptor, dataSource);

		if (treeQueryProviderModule != null) {
			treeQueryProviderModule.addTreeProvider(this);
		}
	}

	@Override
	protected void moduleConfigured() throws Exception {

		stopScheduler();

		super.moduleConfigured();

		if (StringUtils.isEmpty(cacheFilestore)) {

			log.warn("Cache filestore not set");

		} else if (!FileUtils.isReadable(cacheFilestore)) {

			log.error("Cache filestore not readable");

		} else {

			treeCacheFile = new File(cacheFilestore + File.separator + this.getClass().getSimpleName() + "-tree.cache");
			usersCacheFile = new File(cacheFilestore + File.separator + this.getClass().getSimpleName() + "-users.cache");
		}

		if (ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.ERROR)) {

			initScheduler();
		}

	}

	@SuppressWarnings("unchecked")
	protected void initScheduler() {

		scheduler = new Scheduler(systemInterface.getApplicationName() + " - " + moduleDescriptor.toString());
		scheduler.setDaemon(true);

		Task cacheTask = new Task() {

			@Override
			public void execute(TaskExecutionContext context) throws RuntimeException {

				cacheTree(context);
			}

			@Override
			public boolean canBeStopped() {

				return true;
			}

		};

		scheduler.schedule(cacheInterval, cacheTask);
		scheduler.start();

		if (firstStart) {

			if (treeCacheFile != null && treeCacheFile.exists() && usersCacheFile != null && usersCacheFile.exists()) {

				try {
					long startTime = System.currentTimeMillis();

					organizationTree = SerializationUtils.deserializeFromArray(TreeNode.class, FileUtils.getRawBytes(treeCacheFile));
					usersMap = SerializationUtils.deserializeFromArray(HashMap.class, FileUtils.getRawBytes(usersCacheFile));

					log.info("Loaded tree data from disk cache in " + TimeUtils.millisecondsToString(System.currentTimeMillis() - startTime) + " ms");

				} catch (Exception e) {

					organizationTree = null;
					usersMap = null;

					log.error("Error loading tree data from disk cache", e);
				}
			}

			firstStart = false;

			if (organizationTree == null || usersMap == null) {

				scheduler.launch(cacheTask);
			}

		} else {

			scheduler.launch(cacheTask);
		}
	}

	protected void stopScheduler() {

		try {
			if (scheduler != null) {

				scheduler.stop();
				scheduler = null;
			}

		} catch (IllegalStateException e) {
			log.error("Error stopping scheduler", e);
		}
	}

	@Override
	public void unload() throws Exception {

		stopScheduler();

		if (treeQueryProviderModule != null) {
			treeQueryProviderModule.removeTreeProvider(this);
		}

		if (treeCacheFile != null && organizationTree != null) {

			log.info("Saving tree data to disk cache");

			FileUtils.writeFile(treeCacheFile, SerializationUtils.serializeToArray(organizationTree));
			FileUtils.writeFile(usersCacheFile, SerializationUtils.serializeToArray(usersMap));
		}

		organizationTree = null;
		usersMap = null;

		super.unload();
	}

	private synchronized void cacheTree(TaskExecutionContext context) {

		log.info("Caching groups and users..");

		long startTime = System.currentTimeMillis();

		try {

			List<Group> groups = systemInterface.getGroupHandler().getGroupsByAttribute(groupIDAttribute, true);

			if (groups != null) {

				boolean hasMultipleRootNodes = false;

				HashMap<String, TreeNode> treeMap = new HashMap<String, TreeNode>();
				TreeNode rootNode = null;

				for (Group group : groups) {

					if(!group.isEnabled()) {
						
						continue;
					}
					
					String groupID = group.getAttributeHandler().getString(groupIDAttribute);
					
					if (treeMap.get(groupID) != null) {

						log.warn("Duplicated group with id " + groupID + " found, skipping..");
						continue;
					}

					TreeNode treeNode = new TreeNode(group.getName(), groupID);
					
					treeMap.put(groupID, treeNode);

					String parentGroupID = group.getAttributeHandler().getString(parentGroupIDAttribute);
					
					if (parentGroupID == null) {

						if(rootNode == null) {
							
							rootNode = treeMap.get(groupID);
							
						} else {
							
							log.warn("Has multiple root nodes");
							
							hasMultipleRootNodes = true;
							
						}
						
					}
					
				}
				
				if (hasMultipleRootNodes || rootNode == null) {

					rootNode = new TreeNode(getTreeName(), getTreeID());
				}

				int usersCount = 0;
				HashMap<String, List<TreeUser>> usersMap = new HashMap<String, List<TreeUser>>();

				for (Group group : groups) {

					if (context.isStopped()) {

						log.info("Caching aborted");
						return;
					}

					if (!group.isEnabled()) {

						continue;
					}
					
					String groupID = group.getAttributeHandler().getString(groupIDAttribute);
					String parentGroupID = group.getAttributeHandler().getString(parentGroupIDAttribute);

					if (!StringUtils.isEmpty(parentGroupID)) {

						TreeNode parentNode = treeMap.get(parentGroupID);

						if (parentNode == null) {

							log.warn("Parent node " + parentGroupID + " not found specified on " + group + " (" + groupID + ")");
							continue;
						}

						if (parentNode.getKey().equals(groupID)) {

							log.warn("Group " + group + " (" + groupID + ") has itself as parent and will be skipped");
							continue;
						}

						parentNode.put(treeMap.get(groupID));
						
					} else if(!rootNode.getKey().equals(groupID)) {
						
						rootNode.put(treeMap.get(groupID));
						
					}

					List<User> groupUsers = systemInterface.getUserHandler().getUsersByGroup(group.getGroupID(), false, true);

					if (groupUsers != null) {

						List<TreeUser> users = new ArrayList<TreeUser>(groupUsers.size());

						for (User groupUser : groupUsers) {

							TreeUser user = new TreeUser();

							user.setFirstname(groupUser.getFirstname());
							user.setLastname(groupUser.getLastname());
							user.setProviderUserID(groupUser.getUserID() + "");
							user.setCitizenIdentifier(CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(groupUser));
							user.setEmail(groupUser.getEmail());

							users.add(user);

						}

						Collections.sort(users);
						usersMap.put(groupID, users);
						usersCount += users.size();

					}

				}

				organizationTree = rootNode;
				this.usersMap = usersMap;

				log.info("Cached " + groups.size() + " groups and " + usersCount + " users in " + TimeUtils.millisecondsToString(System.currentTimeMillis() - startTime) + " ms");

				return;

			}

		} catch (Exception e) {

			log.error("Unable to cache group and users", e);
		}

	}

	@Override
	public TreeNode getTree() {

		return organizationTree;
	}

	@InstanceManagerDependency
	public void setTreeQueryProviderModule(UserTreeQueryProviderModule treeQueryProviderModule) {

		if (treeQueryProviderModule != null) {

			treeQueryProviderModule.addTreeProvider(this);

		} else if (this.treeQueryProviderModule != null) {

			this.treeQueryProviderModule.removeTreeProvider(this);
		}

		this.treeQueryProviderModule = treeQueryProviderModule;
	}

	@Override
	public String getTreeName() {

		return moduleDescriptor.getDescription();
	}

	@Override
	public String getTreeID() {

		return providerID;
	}

	@Override
	public List<TreeUser> getUsers(String nodeKey) {

		if (usersMap == null) {
			return null;
		}

		List<TreeUser> users = usersMap.get(nodeKey);

		if (users == null) {
			return Collections.emptyList();
		}

		return users;
	}

	@Override
	public User getLocalUser(String providerUserID) {

		if (!StringUtils.isEmpty(providerUserID)) {

			if (!NumberUtils.isInt(providerUserID)) {

				return null;
			}

			return systemInterface.getUserHandler().getUser(Integer.parseInt(providerUserID), false, true);
		}

		return null;
	}

}
