package com.nordicpeak.flowengine.tags.httpsharing;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.flowengine.tags.TagSharingProvider;
import com.nordicpeak.flowengine.tags.TagSharingTarget;
import com.nordicpeak.flowengine.tags.TextTag;

public class HTTPTagSharingProviderModule extends AnnotatedForegroundModule implements TagSharingProvider{

	@ModuleSetting(allowsNull = true)
	@TextAreaSettingDescriptor(name = "Tag sharing targets", description = "https://url:username:password", formatValidator = HTTPTagSharingTargetValidator.class)
	protected List<String> tagSharingConfiguration;	
	
	protected Map<String, HTTPTagSharingTarget> tagSharingTargets;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);
		
		if (!systemInterface.getInstanceHandler().addInstance(TagSharingProvider.class, this)) {
			
			throw new RuntimeException("Unable to register module " + this.moduleDescriptor + " in global instance handler using key " + TagSharingProvider.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}	
	
	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(TagSharingProvider.class, this);
		
		super.unload();
	}
	
	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();
		
		parseTagSharingConfiguration();
	}

	private void parseTagSharingConfiguration() {

		if (!CollectionUtils.isEmpty(tagSharingConfiguration)) {
			
			LinkedHashMap<String, HTTPTagSharingTarget> targets = new LinkedHashMap<String, HTTPTagSharingTarget>(tagSharingConfiguration.size());
			
			for (String configRow : tagSharingConfiguration) {
				
				String[] splits = configRow.split("(?<!https?):");
				
				if (splits.length == 3) {
					
					String url = splits[0];
					String username = splits[1];
					String password = splits[2];
					
					HTTPTagSharingTarget target = new HTTPTagSharingTarget(url.replace("|", ":"), username, password);
					
					targets.put(target.getName(), target);
					
				} else {
					
					log.warn("Incorrect format of config row \"" + configRow + "\"");
				}
			}
			
			if(!targets.isEmpty()) {
				
				this.tagSharingTargets = targets;
				
				return;
			}
		}
		
		this.tagSharingTargets = null;
	}

	@Override
	public void shareTags(List<TextTag> tags, List<TagSharingTarget> targets, boolean overwrite) throws TransformerFactoryConfigurationError, TransformerException, UnsupportedEncodingException {

		Document doc = XMLUtils.createDomDocument();

		Element tagsElement = doc.createElement("Tags");
		doc.appendChild(tagsElement);

		XMLUtils.append(doc, tagsElement, tags);

		String xml = XMLUtils.toString(doc, systemInterface.getEncoding(), true);

		SimpleRequest request = new SimpleRequest("");

		String data = "xml=" + URLEncoder.encode(xml, systemInterface.getEncoding()) + "&overwrite=" + overwrite;

		for (TagSharingTarget target : targets) {

			InputStream inputStream = new ByteArrayInputStream(data.getBytes());

			HTTPTagSharingTarget httpTarget = (HTTPTagSharingTarget) target;

			request.setEndpoint(httpTarget.getUrl());
			request.setUsername(httpTarget.getUsername());
			request.setPassword(httpTarget.getPassword());

			try {
				
				log.info("Sending tags " + tags + " to target " + target + "...");

				log.info("Got response: " + HTTPUtils.sendHTTPPostRequest(request, inputStream, Charset.defaultCharset()));

			} catch (IOException e) {

				log.error("Error sending tags " + tags + " to target " + target, e);
			}
		}
	}

	@Override
	public TagSharingTarget getTarget(String name) {

		if(tagSharingTargets != null) {
			
			return tagSharingTargets.get(name);
		}
		
		return null;
	}

	@Override
	public Collection<? extends TagSharingTarget> getTargets() {

		if(tagSharingTargets != null) {
			
			return tagSharingTargets.values();
		}
		
		return null;
	}
}
