package com.nordicpeak.flowengine.queries.iconmapquery;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;

import com.nordicpeak.flowengine.queries.iconmapquery.beans.Geometry;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQueryInstance;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.SelectedRemoteGeometry;
import com.nordicpeak.flowengine.queries.iconmapquery.utils.GeoJSONUtils;
import com.nordicpeak.flowengine.remotegeometryendpointprovider.RemoteGeometry;
import com.vividsolutions.jts.algorithm.CentroidArea;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;

public class MapImageDataGenerator {

	private final String fullModuleURL;

	private JsonObject styles;
	private JsonArray features;
	private List<Coordinate> allCoordinates;

	public MapImageDataGenerator(String fullModuleURL) {

		this.fullModuleURL = fullModuleURL;
	}

	public MapImageData generateMapImageData(IconMapQueryInstance queryInstance) throws ValidationException {

		styles = new JsonObject();
		features = new JsonArray();
		allCoordinates = new ArrayList<>();

		generateGeometryData(queryInstance);
		generateRemoteGeometryData(queryInstance);

		Envelope extent = generateExtent(queryInstance);
		Coordinate centerCoordinate = calculateCentreCoordinate(extent);

		if (queryInstance.getPudGeometries() != null) {

			JsonObject style = new JsonObject();

			style.putField("strokeWidth", 1.5);
			style.putField("strokeOpacity", 1);
			style.putField("strokeColor", "#0099FF");
			style.putField("fillColor", "#0099FF");
			style.putField("fillOpacity", 0);

			String styleID = "pud-geometry";

			styles.putField(styleID, style);

			for (com.vividsolutions.jts.geom.Geometry geometry : queryInstance.getPudGeometries()) {

				JsonObject properties = new JsonObject();
				properties.putField("_style", styleID);

				JsonNode featureGeometry = GeoJSONUtils.getJSONForGeometry(geometry);

				JsonObject feature = new JsonObject();
				feature.putField("type", "Feature");
				feature.putField("properties", properties);
				feature.putField("geometry", featureGeometry);

				features.addNode(feature);

			}

		}

		return new MapImageData(styles, features, extent, centerCoordinate);
	}

	private void generateGeometryData(IconMapQueryInstance queryInstance) {

		if (!CollectionUtils.isEmpty(queryInstance.getGeometries())) {

			for (Geometry geometry : queryInstance.getGeometries()) {

				com.vividsolutions.jts.geom.Geometry printableGeometry = geometry.getPrintableGeometry();

				if (printableGeometry != null) {

					JsonObject style = new JsonObject();
					style.putField("externalGraphic", fullModuleURL + "/icon/" + geometry.getMapIcon().getMapIconID(), true);

					style.putField("pointRadius", 10);

					if (queryInstance.getQuery().isUseDetailedIconInfo()) {
						style.putField("label", "#" + geometry.getIndex());
						style.putField("labelAlign", "rm");
						style.putField("labelXOffset", "25");
						style.putField("fontSize", "10px");
						style.putField("fontWeight", "normal");
					}

					String styleID = "geometry-" + geometry.getIndex();

					styles.putField(styleID, style);

					JsonObject properties = new JsonObject();
					properties.putField("_style", styleID);

					JsonObject featureGeometry = new JsonObject();

					featureGeometry.putField("type", printableGeometry.getGeometryType());

					Coordinate[] coordinates = printableGeometry.getCoordinates();

					if (coordinates != null) {

						JsonArray coords = new JsonArray();

						String x = NumberUtils.formatNumber(coordinates[0].x, 0, 1, false, true);
						String y = NumberUtils.formatNumber(coordinates[0].y, 0, 1, false, true);

						coords.addNode(x);
						coords.addNode(y);

						featureGeometry.putField("coordinates", coords);

						allCoordinates.addAll(Arrays.asList(coordinates));

					}

					JsonObject feature = new JsonObject();
					feature.putField("type", "Feature");
					feature.putField("properties", properties);
					feature.putField("geometry", featureGeometry);

					features.addNode(feature);
				}
			}
		}
	}

	private void generateRemoteGeometryData(IconMapQueryInstance queryInstance) {

		Set<String> generatedRemoteGeometryIDs = new HashSet<>();

		if (!CollectionUtils.isEmpty(queryInstance.getSelectedGeometries())) {

			for (SelectedRemoteGeometry geometry : queryInstance.getSelectedGeometries()) {

				generateRemoteGeometryData(geometry);

				generatedRemoteGeometryIDs.add(geometry.getObjectID());

			}
		}

		if (queryInstance.getQuery().isPrintAllFetchedGeometries() && !CollectionUtils.isEmpty(queryInstance.getFetchedGeometries())) {

			for (RemoteGeometry geometry : queryInstance.getFetchedGeometries()) {

				if (generatedRemoteGeometryIDs.contains(geometry.getObjectID())) {

					continue;
				}

				generateRemoteGeometryData(geometry);
			}
		}
	}

	private void generateRemoteGeometryData(Object geometry) {

		Integer index;
		String objectID;
		String iconURL;
		com.vividsolutions.jts.geom.Geometry printableGeometry;
		String color;

		if (geometry instanceof RemoteGeometry) {

			RemoteGeometry remoteGeometry = (RemoteGeometry) geometry;

			index = null;
			objectID = remoteGeometry.getObjectID();
			iconURL = remoteGeometry.getIconURL();
			printableGeometry = remoteGeometry.getGeometry();
			color = remoteGeometry.getColor() != null ? remoteGeometry.getColor() : "#0000ff";

		} else if (geometry instanceof SelectedRemoteGeometry) {

			SelectedRemoteGeometry selectedRemoteGeometry = (SelectedRemoteGeometry) geometry;

			index = selectedRemoteGeometry.getIndex();
			objectID = selectedRemoteGeometry.getObjectID();
			iconURL = selectedRemoteGeometry.getIconURL();
			printableGeometry = selectedRemoteGeometry.getPrintableGeometry();
			color = selectedRemoteGeometry.getColor() != null ? selectedRemoteGeometry.getColor() : "#088727";

		} else {

			throw new IllegalArgumentException();
		}

		if (printableGeometry != null) {

			JsonObject style = new JsonObject();

			if (index != null) {
				style.putField("label", "#" + index);
			}
			style.putField("labelAlign", "rm");
			style.putField("labelXOffset", "25");
			style.putField("fontSize", "10px");
			style.putField("fontWeight", "normal");

			String geometryType = printableGeometry.getGeometryType();

			if (geometryType.equalsIgnoreCase("POINT") && iconURL != null) {

				style.putField("externalGraphic", iconURL, true);
				style.putField("pointRadius", 10);

			} else {

				style.putField("pointRadius", 5);
				style.putField("strokeWidth", 2);
				style.putField("strokeColor", color);
				style.putField("strokeOpacity", "0.8");
				style.putField("fillColor", color);
				style.putField("fillOpacity", "0.4");

			}

			String styleID = "remoteGeometry-" + objectID;

			styles.putField(styleID, style);

			JsonObject properties = new JsonObject();
			properties.putField("_style", styleID);

			Coordinate[] coordinates = printableGeometry.getCoordinates();

			if (coordinates != null) {

				JsonNode featureGeometry = GeoJSONUtils.getJSONForGeometry(printableGeometry);

				if (featureGeometry != null) {

					allCoordinates.addAll(Arrays.asList(coordinates));

					JsonObject feature = new JsonObject();
					feature.putField("type", "Feature");
					feature.putField("properties", properties);
					feature.putField("geometry", featureGeometry);

					features.addNode(feature);
				}
			}
		}
	}

	private Envelope generateExtent(IconMapQueryInstance queryInstance) throws ValidationException {

		List<Double> coordinates = NumberUtils.toDouble(Arrays.asList(queryInstance.getExtent().trim().split(",")));

		if (coordinates == null || coordinates.size() != 4) {

			throw new ValidationException(new ValidationError("UnableToGeneratePNG"));

		}

		return new Envelope(coordinates.get(0), coordinates.get(2), coordinates.get(3), coordinates.get(1));
	}

	private Coordinate calculateCentreCoordinate(Envelope extent) {

		return !allCoordinates.isEmpty() ? calculateMapCentroid(allCoordinates) : extent.centre();
	}

	private Coordinate calculateMapCentroid(List<Coordinate> coordinates) {

		if (coordinates.size() == 1) {

			return coordinates.get(0);
		}

		CentroidArea area = new CentroidArea();

		Coordinate[] ring = coordinates.toArray(new Coordinate[coordinates.size()]);

		if (ring.length < 4) {

			return coordinates.get(0);
		}

		area.add(ring);

		return area.getCentroid();
	}

	public class MapImageData {

		private final JsonObject styles;
		private final JsonArray features;
		private final Envelope extent;
		private final Coordinate centerCoordinate;

		private MapImageData(JsonObject styles, JsonArray features, Envelope extent, Coordinate centerCoordinate) {

			this.styles = styles;
			this.features = features;
			this.extent = extent;
			this.centerCoordinate = centerCoordinate;
		}

		public JsonObject getStyles() {

			return styles;
		}

		public JsonArray getFeatures() {

			return features;
		}

		public Envelope getExtent() {

			return extent;
		}

		public Coordinate getCenterCoordinate() {

			return centerCoordinate;
		}

	}

}
