package com.nordicpeak.lmsearch;

import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.apache.log4j.Level;

import se.lantmateriet.namespace.distribution.produkter.belagenhetsadress.v4.BelagenhetsadressType;
import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.interfaces.providers.OAuthCredentialsProvider;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.JsonResponseHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.populators.DoublePopulator;
import se.unlogic.standardutils.string.SimpleStringConverter;
import se.unlogic.standardutils.string.StringConverter;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.lmsearch.beans.Address;
import com.nordicpeak.lmsearch.beans.Addresses;
import com.nordicpeak.lmsearch.beans.BaseLMResponse;
import com.nordicpeak.lmsearch.beans.HeightResponse;
import com.nordicpeak.lmsearch.beans.PlaceName;
import com.nordicpeak.lmsearch.beans.RegisterUnit;
import com.nordicpeak.lmsearch.beans.UnitArea;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

public class LMSearchProvider extends AnnotatedRESTModule implements OAuthCredentialsProvider {

	private static final StringConverter ISO_TO_UTF8_STRING_DECODER = new SimpleStringConverter(Charset.forName("ISO-8859-1"), Charset.forName("UTF-8"));

	@XSLVariable(prefix = "java.")
	private String searchResultTypeRegisterUnit = "not set";

	@XSLVariable(prefix = "java.")
	private String searchResultTypeAddress = "not set";

	@XSLVariable(prefix = "java.")
	private String searchResultTypePlaceName = "not set";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "LM username", description = "", required = true)
	private String username;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "LM password", description = "", required = true)
	private String password;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "LM Consumer Key", description = "The oauth consumer key to LM services", required = false)
	private String consumerKey;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "LM Consumer Secret", description = "The oauth consumer secret to LM services", required = false)
	private String consumerSecret;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "municipality", description = "Search prefix", required = true)
	private String municipality;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "municipalityCode", description = "", required = false)
	private String municipalityCode;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "Connection timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer connectionTimeout = 5;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Request timeout", description = "Request timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer requestTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Max results", description = "Max number of results to return", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer maxResults = 20;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Min search string length", description = "Minimum length of the searchstring required to actually perform a search", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer minimumSearchStringLength = 3;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Incoming request encoding fix", description = "Use when request URI encoding does not match connector URI encoding")
	private boolean useRequestEncodingFix = true;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Use multi register unit area request", description = "Use when LM account has extended authorization to search for multiple unit areas in one request")
	private boolean useMultiRegisterUnitAreaRequest = false;

	private StringConverter URL_STRING_DECODER;

	private LMConnection lmConnection;

	private JsonResponseHandler responseHandler;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		responseHandler = new JsonResponseHandler();

		addResponseHandler(responseHandler);

		super.init(moduleDescriptor, sectionInterface, dataSource);

		systemInterface.getInstanceHandler().addInstance(LMSearchProvider.class, this);

		if (!systemInterface.getInstanceHandler().addInstance(OAuthCredentialsProvider.class, this)) {

			throw new RuntimeException("Unable to register module in global instance handler using key " + OAuthCredentialsProvider.class.getSimpleName() + ", another instance is already registered using this key.");
		}

	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		lmConnection = null;

		if (ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.ERROR)) {

			lmConnection = new LMConnection(username, password, this, connectionTimeout, connectionTimeout, maxResults, minimumSearchStringLength);
		}

		if (useRequestEncodingFix) {

			URL_STRING_DECODER = ISO_TO_UTF8_STRING_DECODER;

			responseHandler.setEncoding("UTF-8");

		} else {

			URL_STRING_DECODER = null;

			responseHandler.setEncoding(systemInterface.getEncoding());
		}
	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(LMSearchProvider.class, this);
		systemInterface.getInstanceHandler().removeInstance(OAuthCredentialsProvider.class, this);

		super.unload();
	}

	// q= municipality[,municipality2 ...] searchtext
	@RESTMethod(alias = "addresses", method = { "get", "options" })
	public JsonNode searchForAddress(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String query = prependMunicipality(fixEncoding(req.getParameter("q"), req));
		boolean returnObject = req.getParameter("json") != null;

		if (!StringUtils.isEmpty(query)) {

			log.info("User " + user + " searching for address with query " + query);

			List<BelagenhetsadressType> addresses = lmConnection.searchAddress(query);

			if (addresses != null) {

				JsonNode jsonNode = new Addresses(addresses).toJson();

				if (returnObject) {

					JsonObject json = new JsonObject();
					json.putField("result", jsonNode);

					return json;
				}

				return jsonNode;
			}
		}

		return new JsonObject();
	}

	@RESTMethod(alias = "enhetsomraden", method = { "get", "options" })
	public JsonNode searchUnitArea(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String xParam = req.getParameter("x");
		String yParam = req.getParameter("y");

		Double x, y;

		if (!StringUtils.isEmpty(xParam) && NumberUtils.isDouble(xParam) && (x = DoublePopulator.getPopulator().getValue(xParam)) != null && !StringUtils.isEmpty(yParam) && NumberUtils.isDouble(yParam) && (y = DoublePopulator.getPopulator().getValue(yParam)) != null) {

			log.info("User " + user + " searching for unit area at coordinates (" + x + ", " + y + "");

			UnitArea unitArea = lmConnection.searchUnitArea(x, y);

			if (unitArea != null) {
				return unitArea.toJson();
			}
		}

		JsonObject response = new JsonObject();
		response.putField("Error", "no results");

		return response;
	}

	@RESTMethod(alias = "placenames", method = { "get", "options" })
	public JsonNode searchForPlace(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String municipalityAndCountyCode = null;

		if (municipalityCode != null) {
			municipalityAndCountyCode = municipalityCode;
		} else {
			municipalityAndCountyCode = req.getParameter("kommunkod");
		}

		String query = fixEncoding(req.getParameter("q"), req);
		boolean returnObject = req.getParameter("json") != null;

		if (!StringUtils.isEmpty(query)) {

			log.info("User " + user + " searching for place name with municipalityAndCountyCode " + municipalityAndCountyCode + " and query " + query);

			List<PlaceName> results = lmConnection.searchForPlaceName(municipalityAndCountyCode, query);

			if (results != null) {
				if (returnObject) {
					JsonObject json = new JsonObject();
					json.putField("result", BaseLMResponse.listToJson(results));

					return json;
				}

				return BaseLMResponse.listToJson(results);
			}
		}

		return new JsonObject();
	}

	// Anvnds troligen ej
	@RESTMethod(alias = "addresses/{id}/enhetsomrade", method = { "get", "options" })
	public JsonNode searchUnitArea(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "id") String objectID) throws Throwable {

		log.info("User " + user + " searching for unit area by objectID " + objectID);

		Point point = lmConnection.getAddressPointByObjectID(objectID);

		if (point != null) {

			UnitArea unitArea = lmConnection.searchUnitArea(point);

			if (unitArea != null) {
				return unitArea.toJson();
			}
		}

		return new JsonObject();
	}

	@RESTMethod(alias = "registerenheter", method = { "get", "options" })
	public JsonNode searchRegisterUnit(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String xParam = req.getParameter("x");
		String yParam = req.getParameter("y");
		String query = prependMunicipality(fixEncoding(req.getParameter("q"), req));
		String estateID = req.getParameter("fnr");
		String estateIDSimple = req.getParameter("fnrsimple");

		if (!StringUtils.isEmpty(xParam) && !StringUtils.isEmpty(yParam)) {

			if (NumberUtils.isDouble(xParam) && NumberUtils.isDouble(yParam)) {

				Double x = DoublePopulator.getPopulator().getValue(xParam);
				Double y = DoublePopulator.getPopulator().getValue(yParam);

				log.info("User " + user + " searching for register unit designation at coordinates (" + x + ", " + y + ")");

				UnitArea unitArea = lmConnection.searchUnitArea(x, y);

				if (unitArea != null) {

					JsonObject response = new JsonObject();
					response.putField("name", unitArea.getDesignation());
					return response;
				}
			}

		} else if (estateID != null) {

			log.info("User " + user + " searching for unit areas with estateID= " + estateID);

			List<UnitArea> unitAreas = lmConnection.getUnitAreas(estateID);

			if (unitAreas != null) {
				return BaseLMResponse.listToFeatureCollectionJson(unitAreas);
			}

		} else if (estateIDSimple != null) {

			log.info("User " + user + " searching for unit area with estateIDSimple= " + estateIDSimple);

			List<UnitArea> unitAreas = lmConnection.getUnitAreas(estateIDSimple);

			if (unitAreas != null) {
				return BaseLMResponse.listToFeatureCollectionJson(unitAreas);
			}

		} else {

			log.info("User " + user + " searching for register unit with query " + query);

			List<RegisterUnit> registerenheter = lmConnection.searchRegisterUnit(query);

			if (registerenheter != null) {
				return BaseLMResponse.listToJson(registerenheter);
			}
		}

		return new JsonObject();
	}

	@RESTMethod(alias = "registerenheter/{id}", method = { "get", "options" })
	public JsonNode getRegisterUnitGeometry(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "id") String objectID) throws Throwable {

		if (!StringUtils.isEmpty(objectID)) {

			log.info("User " + user + " searching for register unit with objectID " + objectID);

			Geometry geometry = lmConnection.getRegisterUnitGeometry(objectID);

			if (geometry != null) {
				return BaseLMResponse.getJSONForGeometry(geometry, log);
			}
		}

		return new JsonObject();
	}

	@RESTMethod(alias = "registerenheter/{id}/enhetsomraden", method = { "get", "options" })
	public JsonNode getUnitAreas(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "id") String objectID) throws Throwable {

		if (!StringUtils.isEmpty(objectID)) {

			log.info("User " + user + " searching for unit areas with objectID " + objectID);

			List<UnitArea> unitAreas = lmConnection.getUnitAreasByRegisterUnitID(objectID);

			if (unitAreas != null) {
				return BaseLMResponse.listToFeatureCollectionJson(unitAreas);
			}
		}

		return new JsonObject();
	}

	@RESTMethod(alias = "registerenheter/{id}/enhetsomraden2", method = { "get", "options" })
	public JsonNode getUnitAreas2(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "id") String objectID) throws Throwable {

		if (!StringUtils.isEmpty(objectID)) {

			log.info("User " + user + " searching for unit areas with objectID " + objectID);

			List<UnitArea> unitAreas = lmConnection.getUnitAreasByRegisterUnitID(objectID, false);

			if (!CollectionUtils.isEmpty(unitAreas)) {

				List<Geometry> geometries = CollectionUtils.map(unitAreas, UnitArea::getGeometry);

				UnitArea unitArea = unitAreas.get(0);

				JsonObject json = new JsonObject(3);

				json.putField("objectID", unitArea.getEstateID());
				json.putField("name", unitArea.getDesignation());
				json.putField("geometry", mergeGeometries(geometries));

				return json;
			}
		}

		return new JsonObject();
	}

	@RESTMethod(alias = "elevation/{srid}/{x}/{y}", method = { "get", "options" })
	public JsonNode getElevation(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "srid") String srID, @URIParam(name = "x") Double x, @URIParam(name = "y") Double y) throws Throwable {

		if (x != null && y != null) {

			log.info("User " + user + " getting elevation for x: " + x + ", y: " + y);

			HeightResponse heightResponse = lmConnection.getElevation(x, y);

			if (heightResponse != null) {
				return heightResponse.toJson();
			}

		}

		return new JsonObject();
	}

	// q= municipality[,municipality2 ...] searchtext
	@RESTMethod(alias = "freetext", method = { "get", "options" })
	public JsonNode searchFreeText(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String query = prependMunicipality(fixEncoding(req.getParameter("q"), req));
		boolean returnObject = req.getParameter("json") != null;

		if (!StringUtils.isEmpty(query)) {

			log.info("User " + user + " free text searching with query " + query);

			JsonArray jsonArray = new JsonArray();
			List<String> objectIdentities = new ArrayList<>();

			List<RegisterUnit> registerUnits = lmConnection.searchRegisterUnit(query);

			List<BelagenhetsadressType> addresses = lmConnection.searchAddress(query);
			List<Address> generatedAddresses = null;

			if (addresses != null) {

				generatedAddresses = new Addresses(addresses).generateAddresses();
			}

			CollectionUtils.add(objectIdentities, CollectionUtils.map(registerUnits, RegisterUnit::getRegisterDesignationID));
			CollectionUtils.add(objectIdentities, CollectionUtils.map(generatedAddresses, Address::getReferenceObjectIdentity));

			Map<String, List<Geometry>> objectIdentityToGeometriesMap = new HashMap<>();
			Map<String, String> objectIdentityToDesignationMap = new HashMap<>();

			if (!CollectionUtils.isEmpty(objectIdentities)) {

				List<UnitArea> unitAreas = null;

				if (useMultiRegisterUnitAreaRequest) {

					unitAreas = lmConnection.getUnitAreasByRegisterUnitIDs(objectIdentities);

				} else {

					unitAreas = new ArrayList<UnitArea>(objectIdentities.size());

					for (String registerUnitID : objectIdentities) {

						CollectionUtils.add(unitAreas, lmConnection.getUnitAreasByRegisterUnitID(registerUnitID));
					}

				}

				if (!CollectionUtils.isEmpty(unitAreas)) {

					for (UnitArea unitArea : unitAreas) {

						if (unitArea.getEstateID() != null) {

							if (unitArea.getGeometry() != null) {

								objectIdentityToGeometriesMap.compute(unitArea.getEstateID(), (key, value) -> CollectionUtils.addAndInstantiateIfNeeded(value, unitArea.getGeometry()));
							}

							if (unitArea.getDesignation() != null) {

								String shortDesignation = StringUtils.substringBeforeLast(StringUtils.substringBeforeLast(unitArea.getDesignation(), " "), " ");
								;

								objectIdentityToDesignationMap.put(unitArea.getEstateID(), shortDesignation);

							}
						}
					}
				}
			}

			if (registerUnits != null) {

				boolean headerIsAdded = false;

				for (RegisterUnit registerUnit : registerUnits) {

					JsonObject registerUnitJson = new JsonObject(3);

					registerUnitJson.putField("objectID", registerUnit.getRegisterDesignationID());
					registerUnitJson.putField("type", searchResultTypeRegisterUnit);
					registerUnitJson.putField("name", registerUnit.getName());
					registerUnitJson.putField("pud", registerUnit.getName());

					Geometry geometry = mergeGeometries(objectIdentityToGeometriesMap.get(registerUnit.getRegisterDesignationID()));
					registerUnitJson.putField("geometry", geometry);

					if (!headerIsAdded) {

						registerUnitJson.putField("header", searchResultTypeRegisterUnit);
						headerIsAdded = true;
					}

					jsonArray.addNode(registerUnitJson);
				}
			}

			if (generatedAddresses != null) {

				boolean headerIsAdded = false;

				for (Address address : generatedAddresses) {

					JsonObject addressJson = new JsonObject(4);

					addressJson.putField("objectID", address.getReferenceObjectIdentity());
					addressJson.putField("type", searchResultTypeAddress);
					addressJson.putField("name", address.getStreetAddress());
					addressJson.putField("pud", objectIdentityToDesignationMap.get(address.getReferenceObjectIdentity()));

					Geometry geometry = mergeGeometries(objectIdentityToGeometriesMap.get(address.getReferenceObjectIdentity()));
					addressJson.putField("geometry", geometry);

					if (!headerIsAdded) {

						addressJson.putField("header", searchResultTypeAddress);
						headerIsAdded = true;
					}

					jsonArray.addNode(addressJson);
				}
			}

			if (returnObject) {

				JsonObject json = new JsonObject();
				json.putField("result", jsonArray);

				return json;
			}

			return jsonArray;
		}

		return new JsonObject();
	}

	private Geometry mergeGeometries(List<Geometry> geometries) {

		if (!CollectionUtils.isEmpty(geometries)) {

			if (geometries.size() == 1) {

				return geometries.get(0);
			}

			List<Polygon> polygonGeometries = null;

			for (Geometry geometry : geometries) {

				if (geometry instanceof Polygon) {

					polygonGeometries = CollectionUtils.addAndInstantiateIfNeeded(polygonGeometries, (Polygon) geometry);
				}
			}

			if (polygonGeometries != null) {

				return lmConnection.getGeometryFactory().createMultiPolygon(polygonGeometries.toArray(new Polygon[] {}));
			}
		}

		return null;
	}

	private String fixEncoding(String input, HttpServletRequest req) {

		if (URL_STRING_DECODER == null) {
			return input;
		}

		return URL_STRING_DECODER.decode(input);
	}

	private String prependMunicipality(String input) {

		if (!StringUtils.isEmpty(input)) {

			String searchPrefix = input.split(" ")[0].trim();

			if (!searchPrefix.equalsIgnoreCase(municipality)) {

				return municipality + " " + input;
			}
		}

		return input;
	}

	public LMConnection getLmConnection() {

		return lmConnection;
	}

	public String getMunicipality() {

		return municipality;
	}

	public String getMunicipalityCode() {

		return municipalityCode;
	}

	@Override
	public String getClientID() {

		return consumerKey;
	}

	@Override
	public String getClientSecret() {

		return consumerSecret;
	}

	public String getUsername() {

		return username;
	}

	public String getPassword() {

		return password;
	}
}
