package com.nordicpeak.flowengine.queries.paragraphinfoquery;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.utils.FCKUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLGenerator;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.url.URLRewriter;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.interfaces.ImmutableQueryDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableQueryInstanceDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableQueryDescriptor;
import com.nordicpeak.flowengine.interfaces.MutableQueryInstanceDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryContentFilter;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryCRUDCallback;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryProviderModule;
import com.nordicpeak.flowengine.utils.JTidyUtils;
import com.nordicpeak.flowengine.utils.TextTagReplacer;

public class ParagraphInfoQueryProviderModule extends BaseQueryProviderModule<ParagraphInfoQueryInstance> implements BaseQueryCRUDCallback {

	private static final RelationQuery QUERY_COPY_IMPORT_RELATIONS = new RelationQuery(ParagraphInfoQuery.PARAGRAPHS_RELATION);
	
	private static final RelationQuery QUERY_INSTANCE_RELATIONS = new RelationQuery(ParagraphInfoQueryInstance.VISIBLE_PARAGRAPHS_RELATION);
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Allow empty title and description", description = "Controls whether one of title or description must be set or not")
	private boolean allowEmptyDescriptionWhenHiddenTitle;
	
	private AnnotatedDAO<ParagraphInfoQuery> queryDAO;
	private AnnotatedDAO<ParagraphInfoQueryInstance> queryInstanceDAO;

	private ParagraphInfoQueryCRUD queryCRUD;
	private ParagraphCRUD paragraphCRUD;

	private QueryParameterFactory<ParagraphInfoQuery, Integer> queryIDParamFactory;
	private QueryParameterFactory<ParagraphInfoQueryInstance, Integer> queryInstanceIDParamFactory;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ParagraphInfoQueryProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		queryDAO = daoFactory.getDAO(ParagraphInfoQuery.class);
		queryInstanceDAO = daoFactory.getDAO(ParagraphInfoQueryInstance.class);

		queryCRUD = new ParagraphInfoQueryCRUD(queryDAO.getWrapper(Integer.class), new AnnotatedRequestPopulator<ParagraphInfoQuery>(ParagraphInfoQuery.class), "ParagraphInfoQuery", "paragraph info query", null, this);
		paragraphCRUD = new ParagraphCRUD(daoFactory.getDAO(Paragraph.class).getWrapper(Integer.class), new AnnotatedRequestPopulator<Paragraph>(Paragraph.class), "Paragraph", "paragraph", null, this);

		queryIDParamFactory = queryDAO.getParamFactory("queryID", Integer.class);
		queryInstanceIDParamFactory = queryInstanceDAO.getParamFactory("queryInstanceID", Integer.class);
	}

	@Override
	public Query createQuery(MutableQueryDescriptor descriptor, TransactionHandler transactionHandler) throws SQLException {

		ParagraphInfoQuery query = new ParagraphInfoQuery();

		query.setQueryID(descriptor.getQueryID());

		this.queryDAO.add(query, transactionHandler, null);

		query.init(descriptor, getFullAlias() + "/config/" + descriptor.getQueryID());

		return query;
	}

	@Override
	public Query importQuery(MutableQueryDescriptor descriptor, TransactionHandler transactionHandler, Map<Integer, ImmutableStatus> statusConversionMap, QueryContentFilter contentFilter) throws Throwable {

		ParagraphInfoQuery query = new ParagraphInfoQuery();

		query.setQueryID(descriptor.getQueryID());

		query.populate(descriptor.getImportParser().getNode(XMLGenerator.getElementName(query.getClass())));
		
		contentFilter.filterHTML(query);

		List<Integer> oldParagraphIDs = new ArrayList<Integer>();

		if (query.getParagraphs() != null) {

			for (Paragraph paragraph : query.getParagraphs()) {

				oldParagraphIDs.add(paragraph.getParagraphID());
				
				paragraph.setParagraphID(null);
				
				paragraph.setText(contentFilter.filterHTML(paragraph.getText()));
			}
		}		
		
		this.queryDAO.add(query, transactionHandler, QUERY_COPY_IMPORT_RELATIONS);

		if (query.getParagraphs() != null) {

			Map<Integer, Integer> paragraphConversionMap = new HashMap<Integer, Integer>(oldParagraphIDs.size());

			int index = 0;

			for (Integer oldParagraphID : oldParagraphIDs) {

				paragraphConversionMap.put(oldParagraphID, query.getParagraphs().get(index).getParagraphID());

				index++;
			}

			query.setParagraphConversionMap(paragraphConversionMap);
		}		
		
		return query;
	}

	@Override
	public Query getQuery(MutableQueryDescriptor descriptor, boolean extraData) throws SQLException {

		ParagraphInfoQuery query = this.getQuery(descriptor.getQueryID());

		if (query == null) {

			return null;
		}

		query.init(descriptor, getFullAlias() + "/config/" + descriptor.getQueryID());

		return query;
	}

	@Override
	public Query getQuery(MutableQueryDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		ParagraphInfoQuery query = this.getQuery(descriptor.getQueryID(), transactionHandler);

		if (query == null) {

			return null;
		}

		query.init(descriptor, getFullAlias() + "/config/" + descriptor.getQueryID());

		return query;
	}

	@Override
	public QueryInstance getQueryInstance(MutableQueryInstanceDescriptor descriptor, String instanceManagerID, HttpServletRequest req, User user, User poster, InstanceMetadata instanceMetadata) throws SQLException {

		ParagraphInfoQueryInstance queryInstance;

		//Check if we should create a new instance or get an existing one
		if (descriptor.getQueryInstanceID() == null) {

			queryInstance = new ParagraphInfoQueryInstance();

		} else {

			queryInstance = getQueryInstance(descriptor.getQueryInstanceID());

			if (queryInstance == null) {

				return null;
			}

			if(queryInstance.getVisibleParagraphs() != null){
				
				for(Paragraph paragraph : queryInstance.getVisibleParagraphs()){
					
					TextTagReplacer.replaceTextTags(paragraph, instanceMetadata.getSiteProfile());
					
					if (req != null) {

						FCKUtils.setAbsoluteFileUrls(paragraph, RequestUtils.getFullContextPathURL(req) + ckConnectorModuleAlias);

						URLRewriter.setAbsoluteLinkUrls(paragraph, req, true);
					}					
				}
			}
		}

		queryInstance.setQuery(getQuery(descriptor.getQueryDescriptor().getQueryID()));

		if (queryInstance.getQuery() == null) {

			return null;
		}

		if(descriptor.getQueryInstanceID() == null){

			//New instance init with default paragraphs
			queryInstance.init();
		}
		
		if (req != null) {

			FCKUtils.setAbsoluteFileUrls(queryInstance.getQuery(), RequestUtils.getFullContextPathURL(req) + ckConnectorModuleAlias);

			URLRewriter.setAbsoluteLinkUrls(queryInstance.getQuery(), req, true);
		}

		queryInstance.getQuery().scanAttributeTags();
		
		TextTagReplacer.replaceTextTags(queryInstance.getQuery(), instanceMetadata.getSiteProfile());

		if(queryInstance.getQuery().getParagraphs() != null){
			
			for(Paragraph paragraph : queryInstance.getQuery().getParagraphs()){
				
				TextTagReplacer.replaceTextTags(paragraph, instanceMetadata.getSiteProfile());
				
				if (req != null) {

					FCKUtils.setAbsoluteFileUrls(paragraph, RequestUtils.getFullContextPathURL(req) + ckConnectorModuleAlias);

					URLRewriter.setAbsoluteLinkUrls(paragraph, req, true);
				}
			}
		}
		
		queryInstance.set(descriptor);
		
		return queryInstance;
	}

	private ParagraphInfoQuery getQuery(Integer queryID) throws SQLException {

		HighLevelQuery<ParagraphInfoQuery> query = new HighLevelQuery<ParagraphInfoQuery>(ParagraphInfoQuery.PARAGRAPHS_RELATION);

		query.addParameter(queryIDParamFactory.getParameter(queryID));

		return queryDAO.get(query);
	}

	private ParagraphInfoQuery getQuery(Integer queryID, TransactionHandler transactionHandler) throws SQLException {

		HighLevelQuery<ParagraphInfoQuery> query = new HighLevelQuery<ParagraphInfoQuery>(ParagraphInfoQuery.PARAGRAPHS_RELATION);

		query.addParameter(queryIDParamFactory.getParameter(queryID));

		return queryDAO.get(query, transactionHandler);
	}

	private ParagraphInfoQueryInstance getQueryInstance(Integer queryInstanceID) throws SQLException {

		HighLevelQuery<ParagraphInfoQueryInstance> query = new HighLevelQuery<ParagraphInfoQueryInstance>(ParagraphInfoQueryInstance.QUERY_RELATION, ParagraphInfoQueryInstance.VISIBLE_PARAGRAPHS_RELATION, ParagraphInfoQuery.PARAGRAPHS_RELATION);

		query.addParameter(queryInstanceIDParamFactory.getParameter(queryInstanceID));

		return queryInstanceDAO.get(query);
	}

	@Override
	public void save(ParagraphInfoQueryInstance queryInstance, TransactionHandler transactionHandler, InstanceRequestMetadata requestMetadata) throws Throwable {

		if (queryInstance.getQueryInstanceID() == null || !queryInstance.getQueryInstanceID().equals(queryInstance.getQueryInstanceDescriptor().getQueryInstanceID())) {

			queryInstance.setQueryInstanceID(queryInstance.getQueryInstanceDescriptor().getQueryInstanceID());

			this.queryInstanceDAO.add(queryInstance, transactionHandler, QUERY_INSTANCE_RELATIONS);

		} else {

			this.queryInstanceDAO.update(queryInstance, transactionHandler, QUERY_INSTANCE_RELATIONS);
		}
	}

	@Override
	public void populate(ParagraphInfoQueryInstance queryInstance, HttpServletRequest req, User user, User poster, boolean allowPartialPopulation, MutableAttributeHandler attributeHandler, InstanceRequestMetadata requestMetadata) throws ValidationException {

		ParagraphInfoQuery query = queryInstance.getQuery();

		if (query.isSetAsAttribute()) {

			attributeHandler.setAttribute(query.getAttributeName(), query.getDescription());
		}

		queryInstance.getQueryInstanceDescriptor().setPopulated(!query.isDontSetPopulated());
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureQuery(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.queryCRUD.show(req, res, user, uriParser);
	}

	@WebPublic(alias = "updatequery")
	public ForegroundModuleResponse updateQuery(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.queryCRUD.update(req, res, user, uriParser);
	}
	
	@WebPublic(alias = "addparagraph")
	public ForegroundModuleResponse addParagraph(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		ParagraphInfoQuery query = this.queryCRUD.getRequestedBean(req, res, user, uriParser, null);

		if (query != null) {

			this.checkUpdateQueryAccess(user, query);

			req.setAttribute("ParagraphInfoQuery", query);

			return paragraphCRUD.add(req, res, user, uriParser);

		}

		throw new URINotFoundException(uriParser);
	}

	@WebPublic(alias = "updateparagraph")
	public ForegroundModuleResponse updateParagraph(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return paragraphCRUD.update(req, res, user, uriParser);
	}

	@WebPublic(alias = "deleteparagraph")
	public ForegroundModuleResponse deleteParagraph(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return paragraphCRUD.delete(req, res, user, uriParser);
	}

	@WebPublic(alias = "sortparagraphs")
	public ForegroundModuleResponse sortParagraphs(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return queryCRUD.sort(req, res, user, uriParser);
	}
	
	@Override
	public boolean deleteQuery(ImmutableQueryDescriptor descriptor, TransactionHandler transactionHandler) throws SQLException {

		ParagraphInfoQuery query = getQuery(descriptor.getQueryID());

		if (query == null) {

			return false;
		}

		this.queryDAO.delete(query, transactionHandler);

		return true;
	}

	@Override
	public boolean deleteQueryInstance(ImmutableQueryInstanceDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		ParagraphInfoQueryInstance queryInstance = this.getQueryInstance(descriptor.getQueryInstanceID());

		if (queryInstance == null) {

			return false;
		}

		this.queryInstanceDAO.delete(queryInstance, transactionHandler);

		return true;
	}

	public void redirectToQueryConfig(ParagraphInfoQuery query, HttpServletRequest req, HttpServletResponse res) throws IOException {

		res.sendRedirect(req.getContextPath() + this.getFullAlias() + "/config/" + query.getQueryID());

	}	
	
	public void checkUpdateQueryAccess(User user, ParagraphInfoQuery query) throws AccessDeniedException, SQLException {

		flowAdminModule.checkFlowStructureManipulationAccess(user, (Flow) query.getQueryDescriptor().getStep().getFlow());
	}
	
	
	public boolean isAllowEmptyDescriptionWhenHiddenTitle() {
	
		return allowEmptyDescriptionWhenHiddenTitle;
	}

	@Override
	public String getTitlePrefix() {

		return this.moduleDescriptor.getName();
	}

	@Override
	public void copyQuery(MutableQueryDescriptor sourceQueryDescriptor, MutableQueryDescriptor copyQueryDescriptor, TransactionHandler transactionHandler, Map<Integer, ImmutableStatus> statusConversionMap) throws SQLException {

		ParagraphInfoQuery query = getQuery(sourceQueryDescriptor.getQueryID(), transactionHandler);

		query.setQueryID(copyQueryDescriptor.getQueryID());

		if (query.getParagraphs() != null) {

			for (Paragraph paragraph : query.getParagraphs()) {

				paragraph.setParagraphID(null);
			}
		}		
		
		queryDAO.add(query, transactionHandler, QUERY_COPY_IMPORT_RELATIONS);
	}

	@Override
	protected void appendPDFData(Document doc, Element showQueryValuesElement, ParagraphInfoQueryInstance queryInstance, AttributeHandler attributeHandler) {

		super.appendPDFData(doc, showQueryValuesElement, queryInstance, attributeHandler);

		if (queryInstance.getQuery().getDescription() != null) {

			XMLUtils.appendNewCDATAElement(doc, showQueryValuesElement, "Description", JTidyUtils.getXHTML(queryInstance.getQuery().getDescription(attributeHandler), systemInterface.getEncoding()));
			XMLUtils.appendNewCDATAElement(doc, showQueryValuesElement, "isHTMLDescription", queryInstance.getQuery().getDescription().contains("<") && queryInstance.getQuery().getDescription().contains(">"));
		}
	}

	@Override
	protected Class<ParagraphInfoQueryInstance> getQueryInstanceClass() {

		return ParagraphInfoQueryInstance.class;
	}
}
