package com.nordicpeak.flowengine.queue;

import java.io.IOException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDate;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.apache.log4j.Level;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.ModuleConfigurationException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.handlers.UserHandler;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionDescriptor;
import se.unlogic.hierarchy.core.interfaces.events.EventHandler;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.fileattachments.FileAttachmentHandler;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.PositiveStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.exceptions.evaluation.EvaluationException;
import com.nordicpeak.flowengine.exceptions.flow.FlowDefaultStatusNotFound;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.MessageCRUDCallback;
import com.nordicpeak.flowengine.queue.beans.ExternalMessage;
import com.nordicpeak.flowengine.queue.beans.ExternalMessageAttachment;
import com.nordicpeak.flowengine.queue.beans.Queue;
import com.nordicpeak.flowengine.queue.beans.QueueSlot;
import com.nordicpeak.flowengine.queue.beans.QueueSlotEvent;
import com.nordicpeak.flowengine.queue.cruds.ExternalMessageCRUD;
import com.nordicpeak.flowengine.queue.enums.ExternalMessageType;
import com.nordicpeak.flowengine.queue.enums.QueueSlotEventType;

public class BaseQueueSlotModule extends AnnotatedRESTModule implements MessageCRUDCallback {

	public static final ValidationError QUEUE_DISABLED_VALIDATION_ERROR = new ValidationError("QueueDisabled");

	@XSLVariable(prefix = "java.")
	protected String eventSortQueueMessage1 = "eventSortQueueMessage1";

	@XSLVariable(prefix = "java.")
	protected String eventSortQueueMessage2 = "eventSortQueueMessage2";

	@XSLVariable(prefix = "java.")
	protected String eventPaymentDirectMessage = "eventPaymentDirectMessage";

	@XSLVariable(prefix = "java.")
	protected String eventPaymentManagerMessage = "eventPaymentManagerMessage";

	@XSLVariable(prefix = "java.")
	protected String eventPaymentRequestAddedMessage = "eventPaymentRequestAddedMessage";

	@XSLVariable(prefix = "java.")
	protected String eventPaymentRequestRemovedMessage = "eventPaymentRequestRemovedMessage";

	@XSLVariable(prefix = "java.")
	protected String eventLeaveMessage = "eventLeaveMessage";

	@XSLVariable(prefix = "java.")
	protected String eventRequestLeaveMessage = "eventRequestLeaveMessage";

	@XSLVariable(prefix = "java.")
	protected String eventArchiveMessage = "eventArchiveByManagerMessage";
	
	@XSLVariable(prefix = "java.")
	protected String eventWaitingPeriodMessage = "eventWaitingPeriodMessage";

	@XSLVariable(prefix = "java.")
	protected String eventAdminSubmittedMessage = "eventAdminSubmittedMessage";

	@XSLVariable(prefix = "java.")
	protected String eventAdminRestoredMessage = "eventAdminRestoredMessage";

	@XSLVariable(prefix = "java.")
	protected String eventSeeExternalMessage = "eventSeeExternalMessage";

	@XSLVariable(prefix = "java.")
	protected String eventDescriptionUpdated = "eventDescriptionUpdated";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Temp dir", description = "Directory for temporary files. Should be on the same filesystem as the file store for best performance. If not set system default temp directory will be used")
	protected String tempDir;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Max request size", description = "Maxmium allowed request size in megabytes", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer maxRequestSize = 1000;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "RAM threshold", description = "Maximum size of files in KB to be buffered in RAM during file uploads. Files exceeding the threshold are written to disk instead.", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer ramThreshold = 500;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Max message file size", description = "Maxmium allowed size in megabytes for external and internal messages", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer maxMessageFileSize = 50;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Reverse external message order", description = "")
	protected boolean reverseOrderMessages = false;

	protected AnnotatedDAO<QueueSlot> queueSlotDAO;
	protected AnnotatedDAO<QueueSlotEvent> queueSlotEventDAO;
	protected AnnotatedDAO<ExternalMessage> externalMessageDAO;
	protected AnnotatedDAO<ExternalMessageAttachment> externalMessageAttachmentDAO;

	protected AnnotatedDAOWrapper<QueueSlotEvent, Integer> queueSlotEventDAOWrapper;

	protected ExternalMessageCRUD externalMessageCRUD;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, QueueAdminModule.class.getName(), new XMLDBScriptProvider(QueueAdminModule.class.getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler(), false, true, false);

		queueSlotDAO = daoFactory.getDAO(QueueSlot.class);
		queueSlotEventDAO = daoFactory.getDAO(QueueSlotEvent.class);
		queueSlotEventDAOWrapper = queueSlotEventDAO.getWrapper(Integer.class);

		queueSlotEventDAOWrapper.addRelations(QueueSlotEvent.ATTRIBUTES_RELATION);

		externalMessageDAO = daoFactory.getDAO(ExternalMessage.class);
		externalMessageAttachmentDAO = daoFactory.getDAO(ExternalMessageAttachment.class);

		externalMessageCRUD = new ExternalMessageCRUD(externalMessageDAO, externalMessageAttachmentDAO, this, true);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.WARN);
	}

	public boolean hasQueueManagerAccess(Queue queue, User user) {

		return queue.isEnabled() && AccessUtils.checkAccess(user, queue);
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");

		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(documentElement);
		return doc;
	}

	public SectionDescriptor getSectionDescriptor() {

		return sectionInterface.getSectionDescriptor();
	}

	public UserHandler getUserHandler() {

		return systemInterface.getUserHandler();
	}

	public boolean isActiveMemberOfQueue(Queue queue, String citizenIdentifier) throws SQLException {

		return QueueUtils.isActiveMemberOfQueue(dataSource, queue, citizenIdentifier);
	}
	
	public List<Timestamp> getWaitingPeriods(Queue queue, String citizenIdentifier) throws SQLException {
		return QueueUtils.getWaitingPeriods(dataSource, queue, citizenIdentifier);
	}
	
	public LocalDate getLatestWaitingDate(Queue queue, String citizenIdentifier) throws SQLException {
		return QueueUtils.getLatestWaitingDate(dataSource, queue, citizenIdentifier);
	}
	
	public List<Integer> getQueueSlotIDs(Queue queue, String citizenIdentifier) throws SQLException {
		return QueueUtils.getQueueSlotIDs(dataSource, queue, citizenIdentifier);
	}

	public EventHandler getEventHandler() {

		return systemInterface.getEventHandler();
	}

	@Override
	public int getRamThreshold() {

		return ramThreshold;
	}

	@Override
	public long getMaxRequestSize() {

		return maxRequestSize;
	}

	@Override
	public long getMaxFileSize() {

		return maxMessageFileSize;
	}

	public Integer getMaxMessageFileSize() {

		return maxMessageFileSize;
	}

	@Override
	public String getTempDir() {

		return tempDir;
	}

	@Override
	public boolean isOperatingStatusDisabled(ImmutableFlowInstance flowInstance, boolean manager) {

		throw new UnsupportedOperationException();
	}

	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> errors) throws SQLException, ModuleConfigurationException {

		throw new UnsupportedOperationException();
	}

	@WebPublic(alias = "externalattachment")
	public ForegroundModuleResponse getExternalMessageAttachment(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws ModuleConfigurationException, SQLException, AccessDeniedException, IOException, FlowDefaultStatusNotFound, EvaluationException, URINotFoundException {

		return externalMessageCRUD.getRequestedMessageAttachment(req, res, user, uriParser);
	}

	public QueueSlotEvent addQueueSlotEvent(QueueSlot queueSlot, QueueSlotEventType eventType, String details, User user, Timestamp timestamp, Map<String, String> attributes) throws SQLException {

		if (timestamp == null) {

			timestamp = TimeUtils.getCurrentTimestamp();
		}

		if (details != null && details.length() > 1024) {
			details = StringUtils.substring(details, 1024);
		}

		QueueSlotEvent queueSlotEvent = new QueueSlotEvent(timestamp, user, eventType, details, attributes);
		queueSlotEvent.setQueueSlot(queueSlot);

		queueSlotEventDAOWrapper.add(queueSlotEvent);

		return queueSlotEvent;
	}

	public ExternalMessage addQueueSlotExternalMessage(QueueSlot queueSlot, String message, boolean postedByManager, User user, ExternalMessageType type, List<ExternalMessageAttachment> attachments) throws SQLException {

		if (message != null && message.length() > 16777215) {
			message = StringUtils.substring(message, 16777215);
		}

		ExternalMessage externalMessage = new ExternalMessage();
		externalMessage.setQueueSlot(queueSlot);
		externalMessage.setPoster(user);
		externalMessage.setMessage(message);
		externalMessage.setAdded(TimeUtils.getCurrentTimestamp());
		externalMessage.setAttachments(attachments);
		externalMessage.setPostedByManager(postedByManager);
		externalMessage.setType(type);

		externalMessageDAO.add(externalMessage);
		return externalMessage;
	}

	public String getEventSortQueueMessage1() {

		return eventSortQueueMessage1;
	}

	public String getEventSortQueueMessage2() {

		return eventSortQueueMessage2;
	}

	public String getEventPaymentManagerMessage() {

		return eventPaymentManagerMessage;
	}

	public String getEventPaymentDirectMessage() {

		return eventPaymentDirectMessage;
	}

	public String getEventPaymentRequestAddedMessage() {

		return eventPaymentRequestAddedMessage;
	}

	public String getEventPaymentRequestRemovedMessage() {

		return eventPaymentRequestRemovedMessage;
	}

	public String getEventRequestLeaveMessage() {

		return eventRequestLeaveMessage;
	}

	public String getEventAdminRestoredMessage() {

		return eventAdminRestoredMessage;
	}

	public String getEventLeaveMessage() {

		return eventLeaveMessage;
	}

	public String getEventArchiveMessage() {

		return eventArchiveMessage;
	}
	
	public String getEventWaitingPeriodMessage() {
		return eventWaitingPeriodMessage;
	}

	public String getEventAdminSubmittedMessage() {

		return eventAdminSubmittedMessage;
	}

	public String getEventSeeExternalMessage() {

		return eventSeeExternalMessage;
	}

	public String getEventDescriptionUpdated() {

		return eventDescriptionUpdated;
	}

	public boolean isReverseOrderMessages() {

		return reverseOrderMessages;
	}

	@Override
	public FileAttachmentHandler getFileAttachmentHandler() {

		// TODO Not implemented in this module yet
		return null;
	}
}
