package com.nordicpeak.flowengine.queue;

import java.sql.SQLException;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.XMLResponseHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.LowLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.xml.XMLGeneratorDocument;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.queue.beans.Queue;
import com.nordicpeak.flowengine.queue.beans.QueueSlot;

public class QueueAPIModule extends AnnotatedRESTModule {

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Force UTF-8 encoding", description = "Forces UTF-8 encoding to be used instead of the system default encoding.")
	protected boolean forceUTF8;

	protected AnnotatedDAO<Queue> queueDAO;
	protected AnnotatedDAO<QueueSlot> queueSlotDAO;

	protected QueryParameterFactory<Queue, Integer> queueIDParamFactory;
	protected QueryParameterFactory<QueueSlot, Queue> queueSlotQueueParamFactory;
	protected QueryParameterFactory<QueueSlot, Integer> queueSlotIDParamFactory;
	protected QueryParameterFactory<QueueSlot, Boolean> queueSlotArchivedParamFactory;

	@Override
	protected void moduleConfigured() throws Exception {

		String encoding;

		if (forceUTF8) {

			encoding = "UTF-8";

		} else {

			encoding = sectionInterface.getSystemInterface().getEncoding();
		}

		setResponseHandlers(CollectionUtils.getGenericSingletonList(new XMLResponseHandler(encoding)));
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, QueueAdminModule.class.getName(), new XMLDBScriptProvider(QueueAdminModule.class.getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler(), false, true, false);

		queueDAO = daoFactory.getDAO(Queue.class);
		queueSlotDAO = daoFactory.getDAO(QueueSlot.class);

		queueIDParamFactory = queueDAO.getParamFactory("queueID", Integer.class);
		queueSlotQueueParamFactory = queueSlotDAO.getParamFactory("queue", Queue.class);
		queueSlotIDParamFactory = queueSlotDAO.getParamFactory("queueSlotID", Integer.class);
		queueSlotArchivedParamFactory = queueSlotDAO.getParamFactory("archived", boolean.class);
	}

	@RESTMethod(alias = "getqueues", method = "get")
	public Document getQueues(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		log.info("User " + user + " requesting queues");

		List<Queue> queues = queueDAO.getAll();

		Document doc = XMLUtils.createDomDocument();

		Element queuesElement = doc.createElement("Queues");
		doc.appendChild(queuesElement);

		XMLUtils.append(doc, queuesElement, queues);

		return doc;
	}

	@RESTMethod(alias = "getactivequeueslots/{queueID}", method = "get")
	public Document getActiveQueueSlots(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "queueID") Integer queueID) throws Throwable {

		return getQueueSlots(queueID, false, user, uriParser);
	}

	@RESTMethod(alias = "getarchivedqueueslots/{queueID}", method = "get")
	public Document getArchivedQueueSlots(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "queueID") Integer queueID) throws Throwable {

		return getQueueSlots(queueID, true, user, uriParser);
	}

	@RESTMethod(alias = "getqueueslot/{queueSlotID}", method = "get")
	public Document getQueueSlotByID(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "queueSlotID") Integer queueSlotID) throws Throwable {

		HighLevelQuery<QueueSlot> query = new HighLevelQuery<>(QueueSlot.CONTACT_RELATION, QueueSlot.MANUAL_CONTACT_RELATION, QueueSlot.EXTERNAL_MESSAGES_RELATION, QueueSlot.QUEUE_RELATION, Queue.INTERNAL_GROUPS_RELATION);

		query.addParameter(queueSlotIDParamFactory.getParameter(queueSlotID));

		QueueSlot queueSlot = queueSlotDAO.get(query);

		if (queueSlot == null) {

			throw new URINotFoundException(uriParser);
		}

		log.info("User " + user + " requesting queue slot with queueSlotID " + queueSlotID);

		queueSlot.getPoster(systemInterface.getUserHandler());
		
		Document doc = XMLUtils.createDomDocument();
		
		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(QueueSlot.SKIP_USER_ATTRIBUTES);

		doc.appendChild(queueSlot.toXML(generatorDocument));

		return doc;
	}

	@RESTMethod(alias = "getqueueslot/{queueID}/{citizenIdentifier}", method = "get")
	public Document getQueueSlotByCitizenIdentifier(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "queueID") Integer queueID, @URIParam(name = "citizenIdentifier") Long citizenIdentifier) throws Throwable {

		LowLevelQuery<QueueSlot> query = new LowLevelQuery<>("SELECT queue_slots.* FROM queue_slots INNER JOIN queue_slot_contacts ON queue_slots.queueSlotID = queue_slot_contacts.queueSlotID INNER JOIN queues ON queue_slots.queueID = queues.queueID WHERE queues.queueID = ? AND queue_slot_contacts.citizenIdentifier = ?");
		
		query.addRelations(QueueSlot.CONTACT_RELATION, QueueSlot.MANUAL_CONTACT_RELATION, QueueSlot.EXTERNAL_MESSAGES_RELATION, QueueSlot.QUEUE_RELATION, Queue.INTERNAL_GROUPS_RELATION);
		
		query.addParameter(queueID);
		query.addParameter(citizenIdentifier);

		QueueSlot queueSlot = queueSlotDAO.get(query);

		if (queueSlot == null) {

			throw new URINotFoundException(uriParser);
		}

		log.info("User " + user + " requesting queue slot for citizenIdentifier " + citizenIdentifier + " and queueID " + queueID);

		queueSlot.getPoster(systemInterface.getUserHandler());
		
		Document doc = XMLUtils.createDomDocument();
		
		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(QueueSlot.SKIP_USER_ATTRIBUTES);

		doc.appendChild(queueSlot.toXML(generatorDocument));

		return doc;
	}

	private Document getQueueSlots(Integer queueID, boolean archived, User user, URIParser uriParser) throws SQLException, URINotFoundException {

		HighLevelQuery<Queue> query = new HighLevelQuery<>(Queue.QUEUE_SLOTS_RELATION, QueueSlot.CONTACT_RELATION, QueueSlot.MANUAL_CONTACT_RELATION);

		query.addParameter(queueIDParamFactory.getParameter(queueID));
		query.addRelationParameter(QueueSlot.class, queueSlotArchivedParamFactory.getParameter(archived));

		Queue queue = queueDAO.get(query);

		if (queue == null) {

			throw new URINotFoundException(uriParser);
		}

		if (archived) {

			log.info("User " + user + " requesting archived queue slots for queue " + queue);

		} else {

			log.info("User " + user + " requesting active queue slots for queue " + queue);
		}

		Document doc = XMLUtils.createDomDocument();

		Element queueSlotsElement = doc.createElement("QueueSlots");
		doc.appendChild(queueSlotsElement);

		XMLUtils.append(doc, queueSlotsElement, queue.getQueueSlots());

		return doc;
	}

}
