package com.nordicpeak.flowengine.remotegeometryendpointprovider;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.menu.MenuItemDescriptor;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AttributeTagUtils;
import se.unlogic.hierarchy.core.utils.CRUDCallback;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.annotations.SplitOnLineBreak;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AdvancedAnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.hash.HashAlgorithms;
import se.unlogic.standardutils.hash.HashUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.AnnotatedBeanTagSourceFactory;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.string.URLEncodingTagReplacer;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.QueryDescriptor;
import com.nordicpeak.flowengine.beans.Step;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.interfaces.APISource;
import com.nordicpeak.flowengine.interfaces.APISourceHandler;
import com.nordicpeak.flowengine.interfaces.APISourceProvider;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.utils.UserAttributeTagUtils;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.io.WKTReader;

public class RemoteGeometryEndpointProviderModule extends AnnotatedForegroundModule implements CRUDCallback<User>, RemoteGeometryEndpointProvider, APISourceProvider {

	public static final AnnotatedBeanTagSourceFactory<User> USER_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(User.class, "$user.");

	@XSLVariable(prefix = "java.")
	private String apiSourceShortTypeName;
	
	@XSLVariable(prefix = "java.")
	private String apiSourceFullTypeName;
	
	@ModuleSetting
	@SplitOnLineBreak
	@TextAreaSettingDescriptor(name = "Allowed encodings", description = "Encodings allowed to use")
	private List<String> allowedEncodings = new ArrayList<>(Arrays.asList("ISO-8859-1", "UTF-8"));

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection")
	private Integer connectionTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The maximum time in seconds to wait for a response")
	private Integer readTimeout = 10;

	@InstanceManagerDependency
	private QueryHandler queryHandler;

	private AnnotatedDAO<RemoteGeometryEndpoint> endpointDAO;
	private AnnotatedDAOWrapper<QueryDescriptor, Integer> queryDescriptorDAOWrapper;

	private RemoteGeometryEndpointCRUD endpointCRUD;
	private QueryParameterFactory<RemoteGeometryEndpoint, String> endpointNameParameterFactory;

	private APISourceHandler apiSourceHandler;
	
	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		if (!systemInterface.getInstanceHandler().addInstance(RemoteGeometryEndpointProvider.class, this)) {

			throw new RuntimeException("Unable to register module in global instance handler using key " + RemoteGeometryEndpointProvider.class.getSimpleName() + ", another instance is already registered using this key.");
		}

	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(RemoteGeometryEndpointProvider.class, this);

		if (this.apiSourceHandler != null) {
			
			this.apiSourceHandler.removeAPISourceProvider(this);
		}
		
		super.unload();
	}

	@InstanceManagerDependency
	public void setAPISourceHandler(APISourceHandler apiSourceHandler) throws SQLException {
		
		if (apiSourceHandler != null) {
			
			apiSourceHandler.addAPISourceProvider(this);
			
		} else if (this.apiSourceHandler != null) {
			
			this.apiSourceHandler.removeAPISourceProvider(this);
		}
		
		this.apiSourceHandler = apiSourceHandler;
		
		//Check if this module is fully cached, then trigger an update of the menu items
		if(sectionInterface.getForegroundModuleCache().isCached(moduleDescriptor)) {
			
			this.sectionInterface.getMenuCache().moduleUpdated(moduleDescriptor, this);
		}
	}	
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, RemoteGeometryEndpointProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		endpointDAO = daoFactory.getDAO(RemoteGeometryEndpoint.class);

		AdvancedAnnotatedDAOWrapper<RemoteGeometryEndpoint, Integer> endpointDAOWrapper = endpointDAO.getAdvancedWrapper(Integer.class);
		endpointDAOWrapper.getGetAllQuery().disableAutoRelations(true);
		endpointDAOWrapper.getGetQuery().disableAutoRelations(true);

		endpointCRUD = new RemoteGeometryEndpointCRUD(endpointDAOWrapper, this);
		endpointNameParameterFactory = endpointDAOWrapper.getAnnotatedDAO().getParamFactory("name", String.class);

		FlowEngineDAOFactory flowDAOFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		queryDescriptorDAOWrapper = flowDAOFactory.getQueryDescriptorDAO().getWrapper(Integer.class);
		queryDescriptorDAOWrapper.addRelations(QueryDescriptor.STEP_RELATION, Step.FLOW_RELATION, Flow.FLOW_TYPE_RELATION, Flow.CATEGORY_RELATION);
		queryDescriptorDAOWrapper.setUseRelationsOnGet(true);
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return endpointCRUD.list(req, res, user, uriParser, null);
	}

	@WebPublic
	public ForegroundModuleResponse show(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return endpointCRUD.show(req, res, user, uriParser);
	}

	@WebPublic
	public ForegroundModuleResponse add(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return endpointCRUD.add(req, res, user, uriParser);
	}

	@WebPublic
	public ForegroundModuleResponse update(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return endpointCRUD.update(req, res, user, uriParser);
	}

	@WebPublic
	public ForegroundModuleResponse delete(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return endpointCRUD.delete(req, res, user, uriParser);
	}

	public List<RemoteGeometryQuery> getQueriesUsingEndpoint(RemoteGeometryEndpoint endpoint) {

		if (queryHandler != null) {

			List<RemoteGeometryQueryProvider> queryProviders = queryHandler.getAssignableQueryProviders(RemoteGeometryQueryProvider.class);

			if (!CollectionUtils.isEmpty(queryProviders)) {

				List<RemoteGeometryQuery> queries = new ArrayList<>(queryProviders.size() * 100);

				for (RemoteGeometryQueryProvider queryProvider : queryProviders) {

					List<RemoteGeometryQuery> queryProviderQueries = null;

					try {

						queryProviderQueries = queryProvider.getQueriesUsingEndpoint(endpoint);

					} catch (SQLException e) {

						log.error("Error getting queries using endpoint " + endpoint + " for query provider " + queryProvider);
					}

					if (queryProviderQueries != null) {

						queries.addAll(queryProviderQueries);
					}
				}

				return queries;
			}
		}

		return null;
	}

	@Override
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();

		Element documentElement = doc.createElement("Document");
		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(documentElement);

		return doc;
	}

	@Override
	public String getTitlePrefix() {

		return moduleDescriptor.getName();
	}

	public List<String> getAllowedEncodings() {

		return allowedEncodings;
	}

	@Override
	public List<? extends RemoteGeometryEndpointDescriptor> getEndpoints() {

		try {

			return endpointDAO.getAll();

		} catch (SQLException e) {

			log.error("Unable to get remote geometry enpoints", e);

		}

		return null;
	}

	@Override
	public RemoteGeometryEndpointDescriptor getEndpoint(String name) {

		try {

			return getRemoteGeometryEndpoint(name);

		} catch (SQLException e) {

			log.error("Unable to get remote geometry enpoint" + name, e);

		}

		return null;
	}

	@Override
	public List<RemoteGeometry> getGeometries(String endPointName, AttributeHandler attributeHandler, User user) {

		RemoteGeometryEndpoint endpoint = null;

		try {

			endpoint = getRemoteGeometryEndpoint(endPointName);

			String url = getEndpointURL(endPointName, attributeHandler, user, endpoint);

			log.info("User " + user + " requesting geometries for endpoint " + endPointName + " using URL " + url);

			SimpleRequest geometryRequest = new SimpleRequest(url);

			geometryRequest.setFollowRedirects(false);
			geometryRequest.setConnectionTimeout(connectionTimeout * 1000);
			geometryRequest.setReadTimeout(readTimeout * 1000);

			if (endpoint.getUsername() != null && endpoint.getPassword() != null) {

				geometryRequest.setUsername(endpoint.getUsername());
				geometryRequest.setPassword(endpoint.getPassword());
			}

			String response = HTTPUtils.sendHTTPGetRequest(geometryRequest, endpoint.getCharset()).getValue();

			if (!StringUtils.isEmpty(response)) {

				GeometryFactory geometryFactory = new GeometryFactory();

				WKTReader reader = new WKTReader(geometryFactory);

				try {

					XMLParser parser = new XMLParser(XMLUtils.parseXML(response, endpoint.getCharset().toString(), false, false));

					List<XMLParser> items = parser.getNodes("/Geometries/Geometry", true);

					List<RemoteGeometry> geometries = new ArrayList<>(items.size());

					if (items.size() > 0) {

						for (XMLParser item : items) {

							String objectID = null;

							try {

								objectID = item.getString("ObjectID");
								String name = item.getString("Name");
								String wktGeometry = item.getString("WKTGeometry");

								if (StringUtils.isEmpty(objectID) || StringUtils.isEmpty(name) || StringUtils.isEmpty(wktGeometry)) {

									continue;
								}

								Geometry parsedGeometry = reader.read(wktGeometry);

								RemoteGeometry geometry = new RemoteGeometry(objectID, name, wktGeometry, parsedGeometry);

								geometry.setDescription(StringUtils.trim(item.getString("Description")));
								geometry.setHideDescription(item.getPrimitiveBoolean("HideDescription"));
								geometry.setIconURL(StringUtils.trim(item.getString("IconURL")));
								geometry.setSelectedIconURL(StringUtils.trim(item.getString("SelectedIconURL")));
								geometry.setSelectable(item.getPrimitiveBoolean("Selectable"));
								geometry.setColor(StringUtils.trim(item.getString("Color")));
								geometry.setSelectedColor(StringUtils.trim(item.getString("SelectedColor")));
								geometry.setPrice(NumberUtils.toInt(item.getString("Price")));

								geometries.add(geometry);

							} catch (Exception e) {

								log.warn("Error parsing remote geometry with objectID " + objectID + " for endpoint " + endpoint, e);

							}
						}
					}

					return geometries;

				} catch (Exception e) {

					log.warn("Error parsing remote geometry response (XML) for " + endpoint, e);

				}
			}

		} catch (Exception e) {

			log.warn("Error getting remote geometries from endpoint " + endpoint, e);

		}

		return null;
	}

	@Override
	public String getEndpointHash(String endPointName, AttributeHandler attributeHandler, User user) {

		RemoteGeometryEndpoint endpoint = null;

		try {

			endpoint = getRemoteGeometryEndpoint(endPointName);

			String url = getEndpointURL(endPointName, attributeHandler, user, endpoint);

			if (url == null) {

				return null;
			}

			return HashUtils.hash(url, HashAlgorithms.MD5);

		} catch (Exception e) {

			log.warn("Error getting hash for endpoint " + endpoint, e);

		}

		return null;
	}

	private String getEndpointURL(String endPointName, AttributeHandler attributeHandler, User user, RemoteGeometryEndpoint endpoint) {

		if (endpoint == null) {

			log.warn("Cannot find requested endpoint " + endPointName);

			return null;
		}

		String url = endpoint.getUrl();

		url = AttributeTagUtils.replaceTags(url, attributeHandler, false, true);

		if (user != null) {

			url = UserAttributeTagUtils.replaceTags(url, user, true);

			TagReplacer tagReplacer = new URLEncodingTagReplacer();

			tagReplacer.addTagSource(USER_TAG_SOURCE_FACTORY.getTagSource(user));

			url = tagReplacer.replace(url);
		}

		return url;
	}

	private RemoteGeometryEndpoint getRemoteGeometryEndpoint(String name) throws SQLException {

		HighLevelQuery<RemoteGeometryEndpoint> query = new HighLevelQuery<>();

		query.addParameter(endpointNameParameterFactory.getParameter(name));

		return endpointDAO.get(query);
	}

	@Override
	public List<? extends APISource> getAPISources() throws Exception {

		return endpointDAO.getAll();
	}
	
	@Override
	public String getShortTypeName() {

		return apiSourceShortTypeName;
	}
	
	@Override
	public String getFullTypeName() {

		return apiSourceFullTypeName;
	}

	@Override
	public String getAddURL() {

		return this.getFullAlias() + "/add";
	}
	
	@Override
	public String getBaseShowURL() {

		return this.getFullAlias() + "/show/";
	}
	
	@Override
	public String getBaseUpdateURL() {

		return this.getFullAlias() + "/update/";
	}

	@Override
	public String getBaseDeleteURL() {

		return this.getFullAlias() + "/delete/";
	}

	@Override
	public boolean isInUse(APISource apiSource) throws SQLException {

		RemoteGeometryEndpoint endpoint = (RemoteGeometryEndpoint)apiSource;
		
		if (queryHandler != null) {

			List<RemoteGeometryQueryProvider> queryProviders = queryHandler.getAssignableQueryProviders(RemoteGeometryQueryProvider.class);

			if (!CollectionUtils.isEmpty(queryProviders)) {

				for (RemoteGeometryQueryProvider queryProvider : queryProviders) {

					if(queryProvider.hasQueriesUsingEndpoint(endpoint)) {
						
						return true;
					}
				}
			}
		}
		
		return false;
	}

	public APISourceHandler getAPISourceHandler() {

		return apiSourceHandler;
	}
	
	@Override
	public List<? extends MenuItemDescriptor> getVisibleMenuItems() {

		if(this.apiSourceHandler != null) {
			
			return null;
		
		}else {
		
			return super.getVisibleMenuItems();
		}
	}
}