package com.nordicpeak.flowengine.remoteflowinstance;

import java.nio.charset.Charset;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;

import se.unlogic.hierarchy.core.annotations.GroupMultiListSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.datatypes.SimpleEntry;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.validation.CharsetValidator;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.UserFlowInstanceModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.enums.ContentType;
import com.nordicpeak.flowengine.interfaces.UserFlowInstanceProvider;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;


public class RemoteXMLFlowInstanceClientModule extends AnnotatedForegroundModule implements UserFlowInstanceProvider{

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API URL", description = "URL to the invoice list API, $citizenIdentifier will be replaced with the users citizen identifier.")
	private String apiURL;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API username", description = "")
	private String apiUsername;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API password", description = "")
	private String apiPassword;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Encoding", description = "Encoding to use when parsing the API response", required = true, formatValidator = CharsetValidator.class)
	private String encoding = "UTF-8";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection", required = true)
	private Integer connectionTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The maximum time in seconds to wait for a response", required = true)
	private Integer readTimeout = 10;
	
	@ModuleSetting(allowsNull = true)
	@GroupMultiListSettingDescriptor(name = "Group access filter", description = "If set then only then this function will only be enabled for the users who are members of the specified groups.")
	private List<Integer> allowedGroups;
	
	private UserFlowInstanceModule userFlowInstanceModule;
	
	@InstanceManagerDependency
	public void setUserFlowInstanceModule(UserFlowInstanceModule userFlowInstanceModule) {

		if (userFlowInstanceModule != null) {

			userFlowInstanceModule.addFlowInstanceProvider(this);

		} else if (this.userFlowInstanceModule != null) {

			this.userFlowInstanceModule.removeFlowInstanceProvider(this);
		}

		this.userFlowInstanceModule = userFlowInstanceModule;
	}	

	@Override
	public List<FlowInstance> getUserFlowInstances(User user) {

		if(allowedGroups != null && !AccessUtils.checkGroupAccess(user, allowedGroups)) {
			
			return null;
		}
		
		String citizenIdentifierID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(user);
		String username = user.getUsername();

		if (apiURL != null && (citizenIdentifierID != null || username != null)) {

			log.info("Getting invoice list for user " + user + "...");

			String url = apiURL;
			
			if(citizenIdentifierID != null) {
				
				url = url.replace("$citizenIdentifier", citizenIdentifierID);
				
			}else {
				
				url = url.replace("$citizenIdentifier", "");
			}
			
			if(username != null) {
				
				url = url.replace("$username", username);
			
			}else {
			
				url = url.replace("$username", "");
			}
			
			SimpleRequest simpleRequest = new SimpleRequest(url);

			simpleRequest.setFollowRedirects(false);
			simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
			simpleRequest.setReadTimeout(readTimeout * 1000);

			if (apiUsername != null && apiPassword != null) {

				simpleRequest.setUsername(apiUsername);
				simpleRequest.setPassword(apiPassword);
			}

			List<Entry<String, String>> headerEntries = new ArrayList<Entry<String, String>>(2);

			headerEntries.add(new SimpleEntry<String, String>("NoLoginRedirect", "NoLoginRedirect"));
			simpleRequest.setHeaders(headerEntries);

			try {
				HTTPResponse response = HTTPUtils.sendHTTPGetRequest(simpleRequest, Charset.forName(encoding));

				try {
					XMLParser parser = new XMLParser(XMLUtils.parseXML(response.getValue(), encoding, false, false));

					List<XMLParser> items = parser.getNodes("/RemoteFlowInstances/RemoteFlowInstance", true);

					if (items.size() > 0) {

						List<FlowInstance> flowInstances = new ArrayList<FlowInstance>(items.size());

						for (XMLParser item : items) {

							String externalID = item.getString("ID");
							String flowName = item.getString("Type");
							String description = item.getString("Description");
							Timestamp added = new Timestamp(DateUtils.DATE_TIME_FORMATTER.parse(item.getString("Added")).getTime());
							
							String statusName = item.getString("Status/Name");
							String statusType = item.getString("Status/Type");
							
							String link = item.getString("Link");
							
							FlowInstance flowInstance = new FlowInstance();
							flowInstance.setRemote(true);
							flowInstance.setPoster(user);

							flowInstance.setAdded(added);
							flowInstance.setFirstSubmitted(added);

							Flow flow = new Flow();
							flow.setName(flowName);
							flow.setEnabled(true);

							flowInstance.setFlow(flow);

							Status flowInstanceStatus = new Status();

							if (statusType != null && statusType.equals("SUBMITTED")) {
								
								flowInstanceStatus.setContentType(ContentType.SUBMITTED);
								
							} else {
								
								flowInstanceStatus.setContentType(ContentType.ARCHIVED);
							}

							flowInstanceStatus.setName(statusName);

							flowInstance.setStatus(flowInstanceStatus);

							MutableAttributeHandler attributeHandler = flowInstance.getAttributeHandler();
							attributeHandler.setAttribute("RemoteFlowInstanceURL", link);
							
							flowInstance.setUserDescription(description);
							attributeHandler.setAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, externalID);

							flowInstances.add(flowInstance);
						}

						log.info("Got " + flowInstances.size() + " remote flow instances for user " + user);
						
						return flowInstances;
					}

					log.info("Got no remote flow instances for user " + user);
					
				} catch (Exception e) {

					log.warn("Error parsing remote flow instance API response for user " + user, e);
				}

			} catch (Throwable e) {

				log.warn("Error connecting to remote flow instance API for user " + user, e);
			}
		}
		
		return null;
	}
	
}
