package com.nordicpeak.flowengine.integration.treservaddecision;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.sql.SQLException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.transform.TransformerConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.GroupMultiListSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.core.utils.extensionlinks.ExtensionLink;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOFactory;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.fileattachments.FileAttachmentHandler;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.http.enums.ContentDisposition;

import com.nordicpeak.flowengine.FlowInstanceAdminModule;
import com.nordicpeak.flowengine.MessageHandler;
import com.nordicpeak.flowengine.MessageHandler.ExternalMessageHandlerInstance;
import com.nordicpeak.flowengine.beans.ExternalMessage;
import com.nordicpeak.flowengine.beans.ExternalMessageAttachment;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.cruds.ExternalMessageCRUD;
import com.nordicpeak.flowengine.integration.treserva.TreservaIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.treserva.beans.CaseData;
import com.nordicpeak.flowengine.integration.treserva.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.treserva.clients.TreservaMinaIFOArendeClient;
import com.nordicpeak.flowengine.integration.treservaddecision.beans.Decision;
import com.nordicpeak.flowengine.integration.treservaddecision.beans.DecisionDocument;
import com.nordicpeak.flowengine.integration.treservaddecision.beans.SentDecision;
import com.nordicpeak.flowengine.interfaces.FlowInstanceOverviewExtensionProvider;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;
import com.nordicpeak.flowengine.utils.FlowEngineFileAttachmentUtils;

import flexjson.JSONDeserializer;

public class TreservaDecisionProviderModule extends AnnotatedForegroundModule implements FlowInstanceOverviewExtensionProvider, ViewFragmentModule<ForegroundModuleDescriptor> {

	private static final DateTimeFormatter YEAR_MONTH_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM");

	@XSLVariable(prefix = "java.")
	protected String tabTitle;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable fragment XML debug", description = "Enables debugging of fragment XML")
	protected boolean debugFragmentXML;

	@ModuleSetting(allowsNull = true)
	@GroupMultiListSettingDescriptor(name = "Manager groups restriction", description = "Restrict access to this feature, if no groups are selected it's available to all")
	protected List<Integer> managerGroupIDs;

	@InstanceManagerDependency(required = true)
	protected QueryHandler queryHandler;

	@InstanceManagerDependency(required = true)
	private TreservaIntegrationProviderModule treservaIntegrationProviderModule;

	@InstanceManagerDependency
	private FileAttachmentHandler fileAttachmentHandler;
	
	@InstanceManagerDependency(required = true)
	protected MessageHandler messageHandler;

	private FlowInstanceAdminModule flowInstanceAdminModule;

	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;

	private AnnotatedDAOWrapper<SentDecision, String> sentDecisionDAOWrapper;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, this.getClass().getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		AnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		sentDecisionDAOWrapper = daoFactory.getDAO(SentDecision.class).getWrapper(String.class);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		this.viewFragmentTransformer = new ModuleViewFragmentTransformer<>(sectionInterface.getForegroundModuleXSLTCache(), this, systemInterface.getEncoding());

		this.viewFragmentTransformer.setDebugXML(debugFragmentXML);
	}

	@Override
	public void unload() throws Exception {

		if (flowInstanceAdminModule != null) {

			setFlowInstanceAdminModule(null);
		}

		super.unload();
	}

	@InstanceManagerDependency(required = true)
	public void setFlowInstanceAdminModule(FlowInstanceAdminModule flowInstanceAdminModule) {

		if (flowInstanceAdminModule == null && this.flowInstanceAdminModule != null) {

			this.flowInstanceAdminModule.removeFlowInstanceOverviewExtensionProvider(this);
		}

		this.flowInstanceAdminModule = flowInstanceAdminModule;

		if (this.flowInstanceAdminModule != null) {

			this.flowInstanceAdminModule.addFlowInstanceOverviewExtensionProvider(this);
		}
	}

	@Override
	public ExtensionLink getOverviewTabHeaderExtensionLink(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user) throws Exception {

		return new ExtensionLink(tabTitle, "#treserva-decision", null, null);
	}

	@Override
	public ViewFragment getOverviewTabContentsViewFragment(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user) throws Exception {

		IntegrationSettings settings = getSettings(flowInstance, user);

		if (settings == null) {

			return null;
		}

		Document doc = createDocument(req, uriParser);
		Element tabContainerElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "TabContentContainer");

		XMLUtils.appendNewElement(doc, tabContainerElement, "ExtensionURL", RequestUtils.getFullContextPathURL(req) + flowInstanceAdminModule.getOverviewExtensionRequestMethodAlias(flowInstance, getOverviewExtensionProviderID()));

		return viewFragmentTransformer.createViewFragment(doc);
	}

	private List<Decision> getDecisions(FlowInstance flowInstance, IntegrationSettings settings) {

		TreservaMinaIFOArendeClient client = getClient(settings);

		Collection<String> signingPartyCitizenIdentifiers = treservaIntegrationProviderModule.getSigningPartiesCitizenIdentifiers(flowInstance);

		String messageTemplate = settings.getDecisionMessageTemplate();

		return getDecisions(flowInstance, signingPartyCitizenIdentifiers, client, messageTemplate);
	}

	public List<Decision> getDecisions(FlowInstance flowInstance, Collection<String> signingPartyCitizenIdentifiers, TreservaMinaIFOArendeClient client, String messageTemplate) {

		String citizenIdentifier = getRegistryOwnerCitizenIdentifier(signingPartyCitizenIdentifiers, flowInstance, client);

		if (citizenIdentifier == null) {

			log.warn("No citizenIdentifier, can not get decisions for flow instance " + flowInstance);
			return Collections.emptyList();
		}

		String response;
		try {

			response = client.GetBeslutDataByPerson(TreservaIntegrationProviderModule.generateCitizenIdentifierJSON(citizenIdentifier));

		} catch (InvalidKeyException | NoSuchAlgorithmException | IOException e) {

			log.warn("Error getting beslut data for person for flow instance " + flowInstance, e);
			return Collections.emptyList();
		}

		if (response == null) {

			log.warn("Empty response getting beslut data for person for flow instance " + flowInstance);
			return Collections.emptyList();
		}

		List<Decision> decisions = new JSONDeserializer<List<Decision>>().use("values", Decision.class).deserialize(response, ArrayList.class);

		if (decisions == null) {

			log.warn("Found no decisions for flow instance " + flowInstance);
			return Collections.emptyList();
		}

		List<DecisionDocument> decisionDocuments = getDecisionDocuments(flowInstance, client, citizenIdentifier);

		for (Decision decision : decisions) {

			DecisionDocument decisionDocument = DecisionDocument.getValidDecisionDocument(decisionDocuments, decision.getNumber());

			decision.setDocument(decisionDocument);

			decision.formatDate();

			decision.generateMessage(messageTemplate);
		}

		return decisions;
	}

	private String getRegistryOwnerCitizenIdentifier(Collection<String> signingPartyCitizenIdentifiers, FlowInstance flowInstance, TreservaMinaIFOArendeClient client) {

		String posterCitizenIdentifier = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(flowInstance.getPoster());

		if (CollectionUtils.isEmpty(signingPartyCitizenIdentifiers)) {

			return posterCitizenIdentifier;
		}

		CaseData posterCaseData = getLatestValidCaseData(posterCitizenIdentifier, flowInstance, client);

		for (String citizenIdentifier : signingPartyCitizenIdentifiers) {

			if (posterCaseData != null) {

				return posterCitizenIdentifier;
			}

			CaseData signingPartyCaseData = getLatestValidCaseData(citizenIdentifier, flowInstance, client);

			if (signingPartyCaseData != null) {

				return citizenIdentifier;
			}
		}

		return null;
	}

	private CaseData getLatestValidCaseData(String citizenIdentifier, FlowInstance flowInstance, TreservaMinaIFOArendeClient client) {

		CaseData caseData = null;

		try {

			caseData = client.getLatestValidCaseData(citizenIdentifier);

		} catch (InvalidKeyException | NoSuchAlgorithmException | IOException e) {

			log.warn("Error getting arende data by personnr for flow instance " + flowInstance, e);
			return null;
		}

		if (caseData == null) {

			log.warn("Empty response getting arende data by personnr for flow instance " + flowInstance);
			return null;
		}

		return caseData;
	}

	private DecisionDocument getDecisionDocument(FlowInstance flowInstance, String decisionNumber, User user) throws SQLException {

		IntegrationSettings settings = getSettings(flowInstance, user);

		if (settings == null) {

			return null;
		}

		TreservaMinaIFOArendeClient client = getClient(settings);

		Collection<String> signingPartyCitizenIdentifiers = treservaIntegrationProviderModule.getSigningPartiesCitizenIdentifiers(flowInstance);

		String citizenIdentifier = getRegistryOwnerCitizenIdentifier(signingPartyCitizenIdentifiers, flowInstance, client);

		return getDecisionDocument(flowInstance, client, citizenIdentifier, decisionNumber);
	}

	private DecisionDocument getDecisionDocument(FlowInstance flowInstance, TreservaMinaIFOArendeClient client, String citizenIdentifier, String decisionNumber) {

		List<DecisionDocument> decisionDocuments = getDecisionDocuments(flowInstance, client, citizenIdentifier);

		return DecisionDocument.getValidDecisionDocument(decisionDocuments, decisionNumber);
	}

	private List<DecisionDocument> getDecisionDocuments(FlowInstance flowInstance, TreservaMinaIFOArendeClient client, String citizenIdentifier) {

		String response = null;
		try {

			response = client.getNormberakningDetails(TreservaIntegrationProviderModule.generateCitizenIdentifierJSON(citizenIdentifier));

		} catch (InvalidKeyException | NoSuchAlgorithmException | IOException e) {

			log.warn("Error getting case documents for person for flow instance " + flowInstance, e);
		}

		if (response == null) {

			log.warn("Empty response getting case documents for person for flow instance " + flowInstance);
			return Collections.emptyList();
		}

		List<DecisionDocument> decisionDocuments = new JSONDeserializer<List<DecisionDocument>>().use("values", DecisionDocument.class).deserialize(response, ArrayList.class);

		if (decisionDocuments == null) {

			return Collections.emptyList();
		}

		return decisionDocuments;
	}

	private void appendFilterMonths(Document doc, Element targetElement, List<Decision> decisions) {

		Set<String> filterMonths = new HashSet<>();

		Element monthsElement = XMLUtils.appendNewElement(doc, targetElement, "FilterMonths");

		for (Decision decision : decisions) {

			if (decision.getIntendedMonth() != null) {

				LocalDateTime parsedIntendedMonth = decision.getParsedIntendedMonth();

				if (parsedIntendedMonth != null) {

					String yearAndMonth = parsedIntendedMonth.format(YEAR_MONTH_FORMATTER);

					if (filterMonths.add(yearAndMonth)) {

						Element monthElement = XMLUtils.appendNewElement(doc, monthsElement, "FilterMonth");

						XMLUtils.appendNewElement(doc, monthElement, "index", parsedIntendedMonth.getMonthValue());
						XMLUtils.appendNewElement(doc, monthElement, "value", yearAndMonth);
					}
				}
			}
		}
	}

	@Override
	public ForegroundModuleResponse processOverviewExtensionRequest(FlowInstance flowInstance, HttpServletRequest req, HttpServletResponse res, URIParser uriParser, User user) throws Exception {

		if (!CollectionUtils.isEmpty(managerGroupIDs) && !AccessUtils.checkGroupAccess(user, managerGroupIDs)) {

			return null;
		}

		String method = uriParser.get(4);

		try {

			if ("tabcontent".equalsIgnoreCase(method)) {

				return getTabContent(flowInstance, req, uriParser, user);
			}
			
			if ("decisiondocument".equalsIgnoreCase(method)) {

				getDecisionDocument(req, res, uriParser, user, flowInstance);

			} else if ("senddecisions".equalsIgnoreCase(method)) {

				sendDecisions(req, res, user, flowInstance);
			}

		} catch (Exception e) {

			log.error("Error processing overview extension request for method " + method, e);

			res.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);

			flowInstanceAdminModule.redirectToMethod(req, res, "/overview/" + flowInstance.getFlowInstanceID() + "#treserva-decision");
		}

		return null;
	}

	private ForegroundModuleResponse getTabContent(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user) throws SQLException, TransformerConfigurationException {

		IntegrationSettings settings = getSettings(flowInstance, user);

		if (settings == null) {

			return null;
		}

		Document doc = createDocument(req, uriParser);
		Element tabContentElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "TabContent");

		log.info("User " + user + " getting decisions for flowInstance " + flowInstance);

		List<Decision> decisions = getDecisions(flowInstance, settings);

		if (decisions != null) {

			List<SentDecision> sentDecisions = sentDecisionDAOWrapper.getAll();

			if (sentDecisions != null) {

				for (Decision decision : decisions) {

					SentDecision sentDecision = SentDecision.getSentDecision(sentDecisions, decision.getDecisionID());

					if (sentDecision != null) {

						decision.setSent(true);

						if (sentDecision.getDate() != null) {

							decision.formatAndSetSentDate(sentDecision.getDate().toLocalDateTime());
						}
					}
				}
			}

			XMLUtils.append(doc, tabContentElement, "Decisions", decisions);
			XMLUtils.append(doc, tabContentElement, flowInstance);

			appendFilterMonths(doc, tabContentElement, decisions);

			XMLUtils.appendNewElement(doc, tabContentElement, "ExtensionURL", RequestUtils.getFullContextPathURL(req) + flowInstanceAdminModule.getOverviewExtensionRequestMethodAlias(flowInstance, getOverviewExtensionProviderID()));

		}

		SimpleForegroundModuleResponse response = new SimpleForegroundModuleResponse(doc);

		response.excludeSystemTransformation(true);

		response.setTransformer(sectionInterface.getForegroundModuleXSLTCache().getModuleTranformer(moduleDescriptor));

		return response;
	}

	private void getDecisionDocument(HttpServletRequest req, HttpServletResponse res, URIParser uriParser, User user, FlowInstance flowInstance) throws SQLException {

		String decisionNumber = uriParser.get(5);

		if (decisionNumber != null) {

			log.info("User " + user + " getting decision document for decison number " + decisionNumber + " and flowInstance " + flowInstance);

			DecisionDocument document = getDecisionDocument(flowInstance, decisionNumber, user);

			if (document != null) {

				try {

					byte[] bytes = document.getDecodedContent();
					InputStream inputStream = new ByteArrayInputStream(bytes);

					String filename = document.getDocumentFileName();

					HTTPUtils.sendFile(inputStream, filename, null, null, req, res, ContentDisposition.ATTACHMENT, (long) bytes.length);

				} catch (IllegalArgumentException e) {

					log.warn("Unable to parse base64  file " + document + " to user " + user + " for decisionNumber " + decisionNumber, e);

				} catch (IOException e) {

					log.warn("Error sending file " + document + " to user " + user + " for decisionNumber " + decisionNumber, e);
				}
			}
		}
	}

	private synchronized void sendDecisions(HttpServletRequest req, HttpServletResponse res, User user, FlowInstance flowInstance) throws Exception {

		String[] decisionIDs = req.getParameterValues("decisionID");

		if (decisionIDs != null) {

			log.info("User " + user + " sending decisions for flowInstance " + flowInstance);

			List<SentDecision> sentDecisions = sentDecisionDAOWrapper.getAll();

			List<ValidationError> validationErrors = new ArrayList<>();
			List<ExternalMessage> externalMessages = null;
			List<SentDecision> newSentDecisions = null;

			for (String decisionID : decisionIDs) {

				if (sentDecisions != null && SentDecision.contains(sentDecisions, decisionID)) {

					continue;
				}

				String message = req.getParameter("decision-message-" + decisionID);
				String number = req.getParameter("decision-number-" + decisionID);

				if (message != null && number != null) {

					List<ExternalMessageAttachment> attachments = null;

					DecisionDocument document = getDecisionDocument(flowInstance, number, user);

					if (document != null) {

						String fileName = document.getDocumentFileName();
						byte[] data = document.getDecodedContent();

						attachments = Collections.singletonList(FlowEngineFileAttachmentUtils.createAttachment(fileAttachmentHandler, fileName, data.length, data, ExternalMessageAttachment.class));
					}

					externalMessages = CollectionUtils.addAndInstantiateIfNeeded(externalMessages, getExternalMessageCRUD().create(message, attachments, false, user, flowInstance, true, validationErrors));

					newSentDecisions = CollectionUtils.addAndInstantiateIfNeeded(newSentDecisions, new SentDecision(decisionID));
				}
			}

			if (!validationErrors.isEmpty()) {

				throw new ValidationException(validationErrors);
			}
			
			List<ExternalMessageHandlerInstance> externalMessageHandlerInstances = CollectionUtils.map(externalMessages, message -> messageHandler.getInstance(message));
			
			try (TransactionHandler transactionHandler = new TransactionHandler(dataSource)) {

				if (externalMessageHandlerInstances != null) {
					
					for (ExternalMessageHandlerInstance externalMessageHandlerInstance : externalMessageHandlerInstances) {
						
						externalMessageHandlerInstance.add(transactionHandler);
					}
				}

				if (newSentDecisions != null) {

					for (SentDecision sentDecision : newSentDecisions) {

						sentDecisionDAOWrapper.add(sentDecision, transactionHandler);
					}
				}

				transactionHandler.commit();

				if (externalMessageHandlerInstances != null) {
					
					messageHandler.commited(externalMessageHandlerInstances.get(0));
				}

			} catch (Throwable t) {

				if (externalMessageHandlerInstances != null) {

					externalMessageHandlerInstances.forEach(ExternalMessageHandlerInstance::cleanup);
				}

				throw t;
			}
		}

		flowInstanceAdminModule.redirectToMethod(req, res, "/messages/" + flowInstance.getFlowInstanceID());
	}

	private ExternalMessageCRUD getExternalMessageCRUD() {

		return flowInstanceAdminModule.getExternalMessageCRUD();
	}

	private Document createDocument(HttpServletRequest req, URIParser uriParser) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(document);

		return doc;
	}

	@Override
	public String getOverviewExtensionProviderID() {

		return moduleDescriptor.getModuleID().toString();
	}

	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {

		return moduleDescriptor;
	}

	@Override
	public List<LinkTag> getLinkTags() {

		return links;
	}

	@Override
	public List<ScriptTag> getScriptTags() {

		return scripts;
	}

	private TreservaMinaIFOArendeClient getClient(IntegrationSettings settings) {

		return treservaIntegrationProviderModule.getTreservaMinaIFOArendeClient(settings);
		//		return new TreservaDecisionDummyClient();
	}

	private IntegrationSettings getSettings(FlowInstance flowInstance, User user) throws SQLException {

		if (!CollectionUtils.isEmpty(managerGroupIDs) && !AccessUtils.checkGroupAccess(user, managerGroupIDs)) {

			return null;
		}

		IntegrationSettings settings = treservaIntegrationProviderModule.getIntegrationSettings(flowInstance);

		if (settings != null && settings.isDecisionModuleEnabled()) {

			return settings;
		}

		return null;
	}

}
