package com.nordicpeak.flowengine.evaluators.dynamicweightquerystateevaluator;

import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.annotations.WebPopulate;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.OneToMany;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLValidationUtils;

import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.beans.QueryModification;
import com.nordicpeak.flowengine.evaluators.basequerystateevaluator.BaseQueryStateEvaluator;
import com.nordicpeak.flowengine.evaluators.setweightevaluator.SetWeightEvaluationEvent;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.EvaluationEventListener;
import com.nordicpeak.flowengine.interfaces.EvaluationHandler;
import com.nordicpeak.flowengine.interfaces.QueryInstance;

@Table(name = "dynamic_weight_query_state_evaluators")
@XMLElement
public class DynamicWeightQueryStateEvaluator extends BaseQueryStateEvaluator implements EvaluationEventListener {

	private static final long serialVersionUID = -1383195490043894435L;

	private ValidationError validationErrorMissingValues = new ValidationError("MissingIntervalValues");
	private ValidationError validationErrorFromBiggerThanTo = new ValidationError("FromValueBiggerThanToValue");
	private ValidationError validationErrorOverlapping = new ValidationError("IntevalOverlapping");
	private ValidationError validationErrorWeightNameMissing = new ValidationError("WeightNameMissing");

	public static final Field VALUE_RELATION = ReflectionUtils.getField(DynamicWeightQueryStateEvaluator.class, "dynamicWeightQueryStateEvaluatorValueList");

	@DAOManaged
	@WebPopulate(required = true, maxLength = 255)
	@XMLElement
	private String weightType;

	@DAOManaged
	@OneToMany(autoAdd = true, autoGet = true, autoUpdate = true)
	@XMLElement
	private List<DynamicWeightQueryStateEvaluatorValues> dynamicWeightQueryStateEvaluatorValueList;

	private HashMap<Integer, BigDecimal> weightSourceMap;

	public String getWeightType() {

		return weightType;
	}

	public void setWeightType(String weightType) {

		this.weightType = weightType;
	}

	public List<DynamicWeightQueryStateEvaluatorValues> getDynamicWeightQueryStateEvaluatorValueList() {

		return dynamicWeightQueryStateEvaluatorValueList;
	}

	public void setDynamicWeightQueryStateEvaluatorValueList(List<DynamicWeightQueryStateEvaluatorValues> dynamicWeightQueryStateEvaluatorValueList) {

		this.dynamicWeightQueryStateEvaluatorValueList = dynamicWeightQueryStateEvaluatorValueList;
	}

	@Override
	public void populate(XMLParser xmlParser) throws ValidationException {

		super.populate(xmlParser);

		List<ValidationError> errors = new ArrayList<>(1);

		weightType = XMLValidationUtils.validateParameter("weightType", xmlParser, false, 1, 255, StringPopulator.getPopulator(), errors);

		try {

			List<XMLParser> dynamicWeightQueryValueNodes = xmlParser.getNodes("dynamicWeightQueryStateEvaluatorValueList/DynamicWeightQueryStateEvaluatorValues");

			if (!CollectionUtils.isEmpty(dynamicWeightQueryValueNodes)) {

				dynamicWeightQueryStateEvaluatorValueList = new ArrayList<>(dynamicWeightQueryValueNodes.size());

				for (XMLParser parser : dynamicWeightQueryValueNodes) {

					DynamicWeightQueryStateEvaluatorValues weightQueryStateEvaluatorValues = new DynamicWeightQueryStateEvaluatorValues();

					weightQueryStateEvaluatorValues.populate(parser);

					weightQueryStateEvaluatorValues.setWeightQueryStateEvaluator(this);

					dynamicWeightQueryStateEvaluatorValueList.add(weightQueryStateEvaluatorValues);
				}
			}
		} catch (ValidationException ve) {
			errors.addAll(ve.getErrors());
		}

		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}

	}

	@Override
	public boolean supportsEvent(Object event) {

		return event instanceof SetWeightEvaluationEvent;
	}

	@Override
	public List<QueryModification> processEvent(Object event, QueryInstance queryInstance, User user, User poster, EvaluationCallback callback, EvaluationHandler evaluationHandler) {

		SetWeightEvaluationEvent weightEvent = (SetWeightEvaluationEvent) event;

		if (this.weightType != null && weightEvent.getWeightType().equalsIgnoreCase(weightType)) {

			if (weightSourceMap == null) {

				weightSourceMap = new HashMap<>();
			}

			this.weightSourceMap.put(weightEvent.getSourceID(), weightEvent.getWeight());

			EvaluationResponse response = super.evaluate(queryInstance, user, poster, callback, evaluationHandler, false, weightEvent.getAttributeHandler());

			if (response != null) {

				return response.getModifications();
			}
		}

		return null;
	}

	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, EvaluationCallback callback, EvaluationHandler evaluationHandler, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {

		return null;
	}

	public BigDecimal getCurrentWeight() {

		if (weightSourceMap != null) {

			BigDecimal totalWeight = BigDecimal.ZERO;

			for (BigDecimal weight : weightSourceMap.values()) {

				totalWeight = totalWeight.add(weight);
			}

			return totalWeight;
		}

		return null;
	}

	public void validate() throws ValidationException {

		List<ValidationError> validationErrors = new ArrayList<>();

		if (StringUtils.isEmpty(weightType)) {
			validationErrors.add(validationErrorWeightNameMissing);
		}

		if (CollectionUtils.isEmpty(dynamicWeightQueryStateEvaluatorValueList)) {
			validationErrors.add(validationErrorMissingValues);
			throw new ValidationException(validationErrors);
		}

		sort(dynamicWeightQueryStateEvaluatorValueList);

		for (int i = 0; i < dynamicWeightQueryStateEvaluatorValueList.size(); i++) {
			DynamicWeightQueryStateEvaluatorValues evaluatorvalue = dynamicWeightQueryStateEvaluatorValueList.get(i);
			if (evaluatorvalue.getFromValue() == null || evaluatorvalue.getToValue() == null) {
				addValidationError(validationErrors, validationErrorMissingValues);
			}

			if (evaluatorvalue.getFromValue() != null && evaluatorvalue.getToValue() != null && evaluatorvalue.getFromValue() > evaluatorvalue.getToValue()) {
				addValidationError(validationErrors, validationErrorFromBiggerThanTo);

			}
			for (int j = i + 1; j < dynamicWeightQueryStateEvaluatorValueList.size(); j++) {
				DynamicWeightQueryStateEvaluatorValues nextEvaluatorvalue = dynamicWeightQueryStateEvaluatorValueList.get(j);
				compareToNextValue(validationErrors, evaluatorvalue, nextEvaluatorvalue);

			}

		}

		if (!validationErrors.isEmpty()) {
			throw new ValidationException(validationErrors);
		}
	}

	private void compareToNextValue(List<ValidationError> validationErrors, DynamicWeightQueryStateEvaluatorValues evaluatorvalue, DynamicWeightQueryStateEvaluatorValues nextEvaluatorvalue) {

		if (evaluatorvalue.getFromValue() != null && nextEvaluatorvalue.getFromValue() != null && (evaluatorvalue.getFromValue() > nextEvaluatorvalue.getFromValue())) {
			addValidationError(validationErrors, validationErrorOverlapping);

		}

		if (evaluatorvalue.getFromValue() != null && nextEvaluatorvalue.getToValue() != null && (evaluatorvalue.getFromValue() > nextEvaluatorvalue.getToValue())) {
			addValidationError(validationErrors, validationErrorOverlapping);

		}

		if (evaluatorvalue.getToValue() != null && nextEvaluatorvalue.getFromValue() != null && (evaluatorvalue.getToValue() > nextEvaluatorvalue.getFromValue())) {
			addValidationError(validationErrors, validationErrorOverlapping);

		}

		if (evaluatorvalue.getToValue() != null && nextEvaluatorvalue.getToValue() != null && (evaluatorvalue.getToValue() > nextEvaluatorvalue.getToValue())) {
			addValidationError(validationErrors, validationErrorOverlapping);

		}

	}

	private void addValidationError(List<ValidationError> validationErrors, ValidationError validationError) {

		if (!validationErrors.contains(validationError)) {
			validationErrors.add(validationError);
		}
	}

	public static void sort(List<DynamicWeightQueryStateEvaluatorValues> values) {

		Collections.sort(values, (o1, o2) -> {

			if (o1.getFromValue() == null && o2.getFromValue() != null) {
				return -1;
			} else if (o1.getFromValue() != null && o2.getFromValue() == null) {
				return 1;
			} else if (o1.getFromValue() == null && o2.getFromValue() == null || o1.getFromValue().equals(o2.getFromValue())) {
				return 0;
			}

			return o1.getFromValue() < o2.getFromValue() ? -1 : 1;

		});
	}

	@Override
	public boolean requiresInitialization() {

		return true;
	}
}
