package com.nordicpeak.flowengine.evaluators.weightapi;

import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.Breadcrumb;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.menu.MenuItemDescriptor;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AttributeTagUtils;
import se.unlogic.hierarchy.core.utils.CRUDCallback;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.userproviders.SimpleUser;
import se.unlogic.standardutils.annotations.SplitOnLineBreak;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOFactory;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.MySQLRowLimiter;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.string.AnnotatedBeanTagSourceFactory;
import se.unlogic.standardutils.string.SingleTagSource;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.string.URLEncodingTagReplacer;
import se.unlogic.standardutils.validation.PositiveStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.evaluators.weightapi.beans.WeightAPISource;
import com.nordicpeak.flowengine.evaluators.weightapi.beans.WeightAPISourceEvent;
import com.nordicpeak.flowengine.evaluators.weightapi.exceptions.UnableToDeleteWeightAPISourceInUseException;
import com.nordicpeak.flowengine.evaluators.weightapi.interfaces.WeightAPIEvaluator;
import com.nordicpeak.flowengine.evaluators.weightapi.interfaces.WeightAPISourceEvaluatorProvider;
import com.nordicpeak.flowengine.interfaces.APISource;
import com.nordicpeak.flowengine.interfaces.APISourceHandler;
import com.nordicpeak.flowengine.interfaces.APISourceProvider;
import com.nordicpeak.flowengine.interfaces.EvaluationHandler;
import com.nordicpeak.flowengine.interfaces.ImmutableFlow;
import com.nordicpeak.flowengine.interfaces.ImmutableStep;
import com.nordicpeak.flowengine.utils.UserAttributeTagUtils;
import com.nordicpeak.flowengine.utils.flowfamilylist.bean.FlowFamilyListItem;
import com.nordicpeak.flowengine.utils.flowfamilylist.connector.FlowFamilyListConnector;
import com.nordicpeak.flowengine.validationerrors.FlowFamilyInUseValidationError;

public class WeightAPIAdminModule extends AnnotatedForegroundModule implements CRUDCallback<User>, APISourceProvider {

	private static final AnnotatedBeanTagSourceFactory<User> USER_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(User.class, "$user.");

	public static final ValidationError API_SOURCE_IN_USE_VALIDATION_ERROR = new ValidationError("UnableToDeleteAPISourceInUse");

	@XSLVariable(prefix = "java.")
	private String apiSourceShortTypeName;
	
	@XSLVariable(prefix = "java.")
	private String apiSourceFullTypeName;
	
	@XSLVariable(prefix = "java.")
	private String eventSourceAddedMessage = "List added";

	@XSLVariable(prefix = "java.")
	private String eventSourceUpdatedMessage = "List updated";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection")
	private Integer connectionTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The maximum time in seconds to wait for a response")
	private Integer readTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Recent events count", description = "Amount of list events to get for showList", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer recentListEventsCount = 4;

	@ModuleSetting
	@SplitOnLineBreak
	@TextAreaSettingDescriptor(name = "Allowed encodings", description = "Encodings allowed to use")
	private List<String> allowedEncodings = new ArrayList<>(Arrays.asList(StandardCharsets.UTF_8.name(), StandardCharsets.ISO_8859_1.name()));

	protected AnnotatedDAO<WeightAPISource> apiSourceDAO;
	protected AnnotatedDAOWrapper<WeightAPISource, Integer> apiSourceDAOWrapper;
	protected AnnotatedDAO<WeightAPISourceEvent> apiSourceEventDAO;

	protected QueryParameterFactory<WeightAPISource, String> apiSourceNameParamFactory;
	protected QueryParameterFactory<WeightAPISourceEvent, WeightAPISource> apiSourceEventAPISourceParamFactory;

	protected WeightAPISourceCRUD apiSourceCRUD;

	@InstanceManagerDependency
	private EvaluationHandler evaluationHandler;
	protected FlowFamilyListConnector flowFamilyListConnector;

	@InstanceManagerDependency
	private FlowAdminModule flowAdminModule;
	
	private APISourceHandler apiSourceHandler;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		if (!systemInterface.getInstanceHandler().addInstance(WeightAPIAdminModule.class, this)) {

			throw new RuntimeException("Unable to register module in global instance handler using key " + WeightAPIAdminModule.class.getSimpleName() + ", another instance is already registered using this key.");
		}
		flowFamilyListConnector = new FlowFamilyListConnector(systemInterface.getEncoding(), flowAdminModule);

	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, WeightAPIAdminModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		AnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);

		apiSourceDAO = daoFactory.getDAO(WeightAPISource.class);
		apiSourceDAOWrapper = apiSourceDAO.getWrapper(Integer.class);
		apiSourceEventDAO = daoFactory.getDAO(WeightAPISourceEvent.class);

		apiSourceNameParamFactory = apiSourceDAO.getParamFactory("name", String.class);
		apiSourceEventAPISourceParamFactory = apiSourceEventDAO.getParamFactory(WeightAPISourceEvent.SOURCE_RELATION, WeightAPISource.class);

		apiSourceCRUD = new WeightAPISourceCRUD(apiSourceDAOWrapper, new AnnotatedRequestPopulator<>(WeightAPISource.class), this);
	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(WeightAPIAdminModule.class, this);

		super.unload();
	}

	@InstanceManagerDependency
	public void setAPISourceHandler(APISourceHandler apiSourceHandler) throws SQLException {
		
		if (apiSourceHandler != null) {
			
			apiSourceHandler.addAPISourceProvider(this);
			
		} else if (this.apiSourceHandler != null) {
			
			this.apiSourceHandler.removeAPISourceProvider(this);
		}
		
		this.apiSourceHandler = apiSourceHandler;
		
		//Check if this module is fully cached, then trigger an update of the menu items
		if(sectionInterface.getForegroundModuleCache().isCached(moduleDescriptor)) {
			
			this.sectionInterface.getMenuCache().moduleUpdated(moduleDescriptor, this);
		}
	}
	
	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return apiSourceCRUD.list(req, res, user, uriParser, null);
	}

	@WebPublic
	public ForegroundModuleResponse add(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return apiSourceCRUD.add(req, res, user, uriParser);
	}

	@WebPublic
	public ForegroundModuleResponse show(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return apiSourceCRUD.show(req, res, user, uriParser);
	}

	@WebPublic
	public ForegroundModuleResponse update(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {
		
		return apiSourceCRUD.update(req, res, user, uriParser);
	}

	@WebPublic
	public ForegroundModuleResponse delete(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		try {
			return apiSourceCRUD.delete(req, res, user, uriParser);

		} catch (UnableToDeleteWeightAPISourceInUseException e) {

			return apiSourceCRUD.show(req, res, user, uriParser, Collections.singletonList(API_SOURCE_IN_USE_VALIDATION_ERROR));
		}
	}

	@WebPublic(alias = "flowfamily")
	public ForegroundModuleResponse getFlowFamily(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return flowFamilyListConnector.getFlowWithFamily(req, res, user);
	}

	@Override
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();

		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.moduleDescriptor.toXML(doc));
		document.appendChild(this.sectionInterface.getSectionDescriptor().toXML(doc));

		if (user != null) {
			document.appendChild(user.toXML(doc));
		}

		doc.appendChild(document);

		return doc;
	}

	@Override
	public String getTitlePrefix() {

		return moduleDescriptor.getName();
	}

	public void addListEvent(String message, WeightAPISource source, User user) {

		message = StringUtils.substring(message, 255, "...");

		WeightAPISourceEvent event = new WeightAPISourceEvent(source, new Timestamp(System.currentTimeMillis()), user, message);

		try {
			apiSourceEventDAO.add(event);

		} catch (SQLException e) {
			log.error("Error adding new list event " + event, e);
		}
	}

	public String getEventSourceAddedMessage() {

		return eventSourceAddedMessage;
	}

	public String getEventSourceUpdatedMessage() {

		return eventSourceUpdatedMessage;
	}

	public List<WeightAPIEvaluator> getEvaluatorsUsingAPISource(WeightAPISource source) {

		List<WeightAPISourceEvaluatorProvider> evaluationProviders = evaluationHandler.getAssignableEvaluationProviders(WeightAPISourceEvaluatorProvider.class);

		if (!CollectionUtils.isEmpty(evaluationProviders)) {

			List<WeightAPIEvaluator> evaluators = new ArrayList<>(evaluationProviders.size() * 100);

			for (WeightAPISourceEvaluatorProvider evaluationProvider : evaluationProviders) {

				Collection<? extends WeightAPIEvaluator> evaluatorInstances = evaluationProvider.getEvaluatorsUsingAPISource(source);

				CollectionUtils.add(evaluators, evaluatorInstances);
			}

			return evaluators;
		}
		
		return null;
	}

	public List<ValidationError> checkMissingFlowFamilies(Integer sourceID, List<Integer> flowFamilyIDs) throws Exception {

		List<ValidationError> validationErrors = null;

		WeightAPISource source = apiSourceDAOWrapper.get(sourceID);

		List<WeightAPIEvaluator> evaluators = getEvaluatorsUsingAPISource(source);

		if (evaluators != null) {

			for (WeightAPIEvaluator evaluator : evaluators) {

				ImmutableFlow flow = evaluator.getEvaluatorDescriptor().getQueryDescriptor().getStep().getFlow();
				Integer flowFamilyID = getFlowFamilyID(flow);

				if(!flowFamilyIDs.contains(flowFamilyID)) {
					
					validationErrors = CollectionUtils.addAndInstantiateIfNeeded(validationErrors, new FlowFamilyInUseValidationError(new FlowFamilyListItem(flowFamilyID, flow.getName())));
				}
			}
		}

		return validationErrors;
	}

	@WebPublic(alias = "usage")
	public ForegroundModuleResponse showListUsage(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Integer sourceID;
		WeightAPISource source;

		if (uriParser.size() >= 3 && PositiveStringIntegerPopulator.getPopulator().validateFormat(uriParser.get(2)) && (sourceID = PositiveStringIntegerPopulator.getPopulator().getValue(uriParser.get(2))) != null && (source = apiSourceDAOWrapper.get(sourceID)) != null) {

			Document doc = createDocument(req, uriParser, user);
			Element showAlternativeListUsageElement = doc.createElement("ShowSourceFlowUsage");
			doc.getDocumentElement().appendChild(showAlternativeListUsageElement);

			showAlternativeListUsageElement.appendChild(source.toXML(doc));

			List<WeightAPIEvaluator> evaluators = getEvaluatorsUsingAPISource(source);

			if (!CollectionUtils.isEmpty(evaluators)) {

				Element evaluatorsElement = XMLUtils.appendNewElement(doc, showAlternativeListUsageElement, "Evaluators");

				Map<Integer, Element> flowElements = new HashMap<>();
				Map<Integer, Element> stepElements = new HashMap<>();

				for (WeightAPIEvaluator evaluator : evaluators) {

					ImmutableFlow flow = evaluator.getEvaluatorDescriptor().getQueryDescriptor().getStep().getFlow();
					Element flowElement = flowElements.get(flow.getFlowID());

					if (flowElement == null) {

						flowElement = (Element) evaluatorsElement.appendChild(flow.toXML(doc));
						flowElements.put(flow.getFlowID(), flowElement);
					}

					ImmutableStep step = evaluator.getEvaluatorDescriptor().getQueryDescriptor().getStep();
					Element stepElement = stepElements.get(step.getStepID());

					if (stepElement == null) {

						stepElement = (Element) flowElement.appendChild(step.toXML(doc));
						stepElements.put(step.getStepID(), stepElement);
					}

					Element queryElement = XMLUtils.appendNewElement(doc, stepElement, "Query");

					queryElement.appendChild(evaluator.getEvaluatorDescriptor().getQueryDescriptor().toXML(doc));
					queryElement.appendChild(evaluator.getEvaluatorDescriptor().toXML(doc));
				}
			}

			return new SimpleForegroundModuleResponse(doc, getTitlePrefix(), getDefaultBreadcrumb(), new Breadcrumb(this, source.getName(), "/show/" + source.getSourceID()));
		}

		throw new URINotFoundException(uriParser);
	}

	public List<WeightAPISourceEvent> getListEvents(WeightAPISource list) throws SQLException {

		HighLevelQuery<WeightAPISourceEvent> query = new HighLevelQuery<>();

		query.addParameter(apiSourceEventAPISourceParamFactory.getParameter(list));

		return apiSourceEventDAO.getAll(query);
	}

	public List<WeightAPISourceEvent> getRecentListEvents(WeightAPISource list) throws SQLException {

		HighLevelQuery<WeightAPISourceEvent> query = new HighLevelQuery<>();

		query.addParameter(apiSourceEventAPISourceParamFactory.getParameter(list));
		query.setRowLimiter(new MySQLRowLimiter(recentListEventsCount));

		return apiSourceEventDAO.getAll(query);
	}

	public WeightAPISource getSource(String name) throws SQLException {

		HighLevelQuery<WeightAPISource> query = new HighLevelQuery<>();

		query.addParameter(apiSourceNameParamFactory.getParameter(name));
		query.setRowLimiter(new MySQLRowLimiter(1));
		
		return apiSourceDAO.get(query);
	}

	public WeightAPISource getSource(String name, ImmutableFlow flow) throws SQLException {

		HighLevelQuery<WeightAPISource> query = new HighLevelQuery<>();

		query.addParameter(apiSourceNameParamFactory.getParameter(name));
		query.setRowLimiter(new MySQLRowLimiter(1));

		WeightAPISource source = apiSourceDAO.get(query);

		Integer flowFamilyID = getFlowFamilyID(flow);
		
		if (source != null && flow != null && source.getFlowFamilies() != null && !source.getFlowFamilies().contains(flowFamilyID)) {

			return null;
		}

		return source;
	}

	@WebPublic(alias = "events")
	public ForegroundModuleResponse showListEvents(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Integer sourceID;
		WeightAPISource source;

		if (uriParser.size() >= 3 && PositiveStringIntegerPopulator.getPopulator().validateFormat(uriParser.get(2)) && (sourceID = PositiveStringIntegerPopulator.getPopulator().getValue(uriParser.get(2))) != null && (source = apiSourceDAOWrapper.get(sourceID)) != null) {

			Document doc = createDocument(req, uriParser, user);
			Element showAlternativeListEventsElement = doc.createElement("ShowAPISourceEvents");
			doc.getDocumentElement().appendChild(showAlternativeListEventsElement);

			showAlternativeListEventsElement.appendChild(source.toXML(doc));
			XMLUtils.append(doc, showAlternativeListEventsElement, "Events", getListEvents(source));

			return new SimpleForegroundModuleResponse(doc, getTitlePrefix(), getDefaultBreadcrumb(), new Breadcrumb(this, source.getName(), "/show/" + source.getSourceID()));
		}

		throw new URINotFoundException(uriParser);
	}

	public BigDecimal getWeight(WeightAPISource source, String fieldValue, User poster, MutableAttributeHandler attributeHandler) {

		String attributedURL = source.getUrl();

		User usedPoster = poster;

		if (usedPoster == null) {

			usedPoster = new SimpleUser();
		}

		if (fieldValue != null) {
			try {
				fieldValue = URLEncoder.encode(fieldValue, StandardCharsets.UTF_8.name());

			} catch (UnsupportedEncodingException ignore) {}
		}

		TagReplacer tagReplacer = new URLEncodingTagReplacer();
		tagReplacer.addTagSource(USER_TAG_SOURCE_FACTORY.getTagSource(usedPoster));
		tagReplacer.addTagSource(new SingleTagSource("$fieldValue", fieldValue));

		attributedURL = tagReplacer.replace(attributedURL);

		attributedURL = UserAttributeTagUtils.replaceTags(attributedURL, usedPoster, true);

		attributedURL = AttributeTagUtils.replaceTags(attributedURL, attributeHandler, false, true);

		log.info("Getting weight for api source " + source + " from: " + attributedURL);

		try {

			SimpleRequest request = new SimpleRequest(attributedURL);

			request.setConnectionTimeout(connectionTimeout * 1000);
			request.setReadTimeout(readTimeout * 1000);
			request.setUsername(source.getUsername());
			request.setPassword(source.getPassword());

			String response = HTTPUtils.sendHTTPGetRequest(request, source.getCharset()).getValue();

			if (!StringUtils.isEmpty(response)) {

				try {
					XMLParser parser = new XMLParser(XMLUtils.parseXML(response, false, false));

					String weight = parser.getString("/Weight");

					if (!StringUtils.isEmpty(weight)) {

						BigDecimal parsedWeight =  new BigDecimal(weight);
						
						log.info("Got weight " + parsedWeight + " for api source " + source + " from: " + attributedURL);
						
						return parsedWeight;
						
					}else {
						
						log.info("Got empty weight for api source " + source + " from: " + attributedURL);
					}

				} catch (Exception e) {

					log.warn("Error parsing weight XML for " + source + ": " + StringUtils.toLogFormat(response, 255), e);
				}
			}

		} catch (Exception e) {
			log.warn("Error getting weight for " + source, e);
		}

		return BigDecimal.ZERO;
	}

	public List<String> getAllowedEncodings() {

		return allowedEncodings;
	}

	public List<WeightAPISource> getSources(ImmutableFlow flow) throws SQLException {

		//TODO replace with cache or low lewel filtering query
		
		List<WeightAPISource> apiSources = apiSourceDAO.getAll();

		if (apiSources != null) {

			Iterator<WeightAPISource> iterator = apiSources.iterator();

			Integer flowFamilyID = getFlowFamilyID(flow);
			
			WeightAPISource apiSource;

			while (iterator.hasNext()) {

				apiSource = iterator.next();
				
				if (apiSource.isUseAccessRestriction() && !apiSource.getFlowFamilies().contains(flowFamilyID)) {

					iterator.remove();
				}
			}

			if (apiSources.isEmpty()) {

				apiSources = null;
			}
		}

		return apiSources;
	}

	//TODO this method can be removed when the base classes for the evaluators has been updated to include the family relation
	private Integer getFlowFamilyID(ImmutableFlow flow) {

		if(flow.getFlowFamily() != null) {
			
			return flow.getFlowFamily().getFlowFamilyID();
		}
		
		Flow cachedFlow = flowAdminModule.getFlowCache().getFlowCacheMap().get(flow.getFlowID());
		
		if(cachedFlow == null) {
			
			throw new RuntimeException("Unable to flow " + flow + " in cache");
		}
		
		return cachedFlow.getFlowFamily().getFlowFamilyID();
	}

	public FlowFamilyListConnector getFlowFamilyListConnector() {
	
		return flowFamilyListConnector;
	}

	@Override
	public List<? extends APISource> getAPISources() throws Exception {

		return apiSourceDAO.getAll();
	}
	
	@Override
	public String getShortTypeName() {

		return apiSourceShortTypeName;
	}
	
	@Override
	public String getFullTypeName() {

		return apiSourceFullTypeName;
	}
	
	@Override
	public String getAddURL() {

		return this.getFullAlias() + "/add";
	}

	@Override
	public String getBaseShowURL() {

		return this.getFullAlias() + "/show/";
	}
	
	@Override
	public String getBaseUpdateURL() {

		return this.getFullAlias() + "/update/";
	}

	@Override
	public String getBaseDeleteURL() {

		return this.getFullAlias() + "/delete/";
	}

	@Override
	public boolean isInUse(APISource apiSource) throws Exception {

		WeightAPISource source = (WeightAPISource)apiSource;
		
		List<WeightAPISourceEvaluatorProvider> evaluationProviders = evaluationHandler.getAssignableEvaluationProviders(WeightAPISourceEvaluatorProvider.class);

		if (!CollectionUtils.isEmpty(evaluationProviders)) {

			for (WeightAPISourceEvaluatorProvider evaluationProvider : evaluationProviders) {

				if(evaluationProvider.hasEvaluatorsUsingAPISource(source)) {
					
					return true;
				}
			}
		}
		
		return false;
	}

	public APISourceHandler getAPISourceHandler() {

		return apiSourceHandler;
	}
	
	@Override
	public List<? extends MenuItemDescriptor> getVisibleMenuItems() {

		if(this.apiSourceHandler != null) {
			
			return null;
		
		}else {
		
			return super.getVisibleMenuItems();
		}
	}
}
